//
// PACL 
//
// Dumps persisted ACL information for files, registry keys, and services.
//
// mmiller@hick.org
// 03/2008
//
#define UNICODE
#include <windows.h>
#include <accctrl.h>
#include <aclapi.h>
#include <strsafe.h>
#include <intsafe.h>

#define USER_MODE
#include "../../driver/metadata.h"

typedef struct _KEY_NAME_STACK
{
	LPTSTR KeyName;
	ULONG KeyNameSize;
	struct _KEY_NAME_STACK *Parent;
} KEY_NAME_STACK, *PKEY_NAME_STACK;

ULONG64 ObjectKeyPool = 1;

//
// Saves the security descriptor of an object
//
VOID
SaveObjectSecurityDescriptor(
	__in HANDLE LogFileHandle,
	__in ULONG64 ObjectKey,
	__in PSECURITY_DESCRIPTOR SecurityDescriptor)
{
	POBJECT_SECURITY_DESCRIPTOR_METADATA SdMetadata = NULL;
	DWORD SecurityDescriptorLength;
	ULONG SdMetadataSize;
	ULONG BytesWritten;

	//
	// We only support self-relative descriptors
	//
	//assert(SecurityDescriptor->Control & SE_SELF_RELATIVE);

	//
	// Build the security descriptor metadata and write it to the file
	//
	SecurityDescriptorLength = GetSecurityDescriptorLength(
			SecurityDescriptor);

	SdMetadataSize = 
		sizeof(OBJECT_SECURITY_DESCRIPTOR_METADATA) + SecurityDescriptorLength;

	SdMetadata = (POBJECT_SECURITY_DESCRIPTOR_METADATA)malloc(
			SdMetadataSize);

	if (SdMetadata)
	{
		SdMetadata->Header.Tag    = OBJECT_METADATA_TAG_SECURITY_DESC;
		SdMetadata->Header.Length = SdMetadataSize;
		SdMetadata->Header.Object = ObjectKey;
		SdMetadata->SdLength      = SecurityDescriptorLength;

		CopyMemory(
				SdMetadata + 1,
				SecurityDescriptor,
				SecurityDescriptorLength);

		WriteFile(
				LogFileHandle,
				SdMetadata,
				SdMetadataSize,
				&BytesWritten,
				NULL);

		free(SdMetadata);
	}
}

//
// Saves the name of an object
//
VOID
SaveObjectName(
	__in HANDLE LogFileHandle,
	__in ULONG64 ObjectKey,
	__in LPTSTR ObjectName)
{
	POBJECT_NAME_METADATA NameMetadata = NULL;
	ULONG ObjectNameCch = 0;
	ULONG ObjectNameCb = 0;
	ULONG NameMetadataSize;
	ULONG BytesWritten;

	ObjectNameCch = wcslen(ObjectName);
	ObjectNameCb  = ObjectNameCch * sizeof(TCHAR);

	//
	// Initialize and write the name metadata to the log file
	//
	NameMetadataSize = 
		sizeof(OBJECT_NAME_METADATA) + ObjectNameCb;

	NameMetadata = (POBJECT_NAME_METADATA)malloc(
			NameMetadataSize);

	if (NameMetadata)
	{
		NameMetadata->Header.Tag    = OBJECT_METADATA_TAG_NAME;
		NameMetadata->Header.Length = NameMetadataSize;
		NameMetadata->Header.Object = ObjectKey;
		NameMetadata->NameLength    = ObjectNameCb;

		CopyMemory(
				NameMetadata + 1,
				ObjectName,
				ObjectNameCb);

		WriteFile(
				LogFileHandle,
				NameMetadata,
				NameMetadataSize,
				&BytesWritten,
				NULL);

		free(NameMetadata);
	}
}

VOID
SaveFileSecurityDescriptor(
	__in HANDLE LogFileHandle,
	__in ULONG64 ObjectKey,
	__in LPTSTR Path)
{
	PISECURITY_DESCRIPTOR SecurityDescriptor = NULL;
	DWORD Status;

	//
	// Query ACL information for the supplied path
	//
	Status = GetNamedSecurityInfo(
			Path,
			SE_FILE_OBJECT,
			DACL_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | OWNER_SECURITY_INFORMATION,
			NULL,
			NULL,
			NULL,
			NULL,
			&SecurityDescriptor);

	if (Status != ERROR_SUCCESS)
	{
		wprintf(L"GetNamedSecurityInfo failed(%s), %lu\n",
				Path,
				GetLastError());
		return;
	}

	SaveObjectSecurityDescriptor(
			LogFileHandle,
			ObjectKey,
			SecurityDescriptor);

	LocalFree(
			SecurityDescriptor);
}

VOID
DumpSpecificDirectoryAcls(
	__in HANDLE LogFileHandle,
	__in LPTSTR Directory)
{
	WIN32_FIND_DATA FindData;
	ULONG64 ObjectKey = ObjectKeyPool++;
	LPTSTR FullPath;
	HANDLE FindHandle;
	ULONG FullPathSize;

	//
	// Save the name & ACLs that have been applied to the directory
	//
	SaveFileSecurityDescriptor(
			LogFileHandle,
			ObjectKey,
			Directory);

	SaveObjectName(
			LogFileHandle,
			ObjectKey,
			Directory);

	//
	// Allocate storage for the directory with wildcard characters
	//
	FullPathSize = (wcslen(Directory) * sizeof(TCHAR)) + sizeof(TCHAR) * 3;

	FullPath = (LPTSTR)malloc(
			FullPathSize);

	if (!FullPath)
	{
		wprintf(L"malloc failed\n");
		return;
	}

	StringCbPrintf(
			FullPath,
			FullPathSize,
			TEXT("%s\\*"),
			Directory);

	//
	// Start our enumeration of all files within this directory
	//
	FindHandle = FindFirstFile(
			FullPath,
			&FindData);

	free(FullPath);

	if (FindHandle == INVALID_HANDLE_VALUE)
		return;

	do
	{
		//
		// Skip . and .., not entirely safe but fine for our purposes
		//
		if (FindData.cFileName[0] == TEXT('.') ||
		    FindData.cFileName[1] == TEXT('.'))
			continue;

		FullPathSize = 
			(wcslen(Directory) * sizeof(TCHAR)) + 
			sizeof(TCHAR) + 
			(wcslen(FindData.cFileName) * sizeof(TCHAR)) +
			sizeof(TCHAR);

		FullPath = malloc(
				FullPathSize);

		if (!FullPath)
			continue;

		StringCbPrintf(
				FullPath,
				FullPathSize,
				TEXT("%s\\%s"),
				Directory,
				FindData.cFileName);

		//
		// If this is a directory, recurse to it
		//
		if ((FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
		{
			DumpSpecificDirectoryAcls(
					LogFileHandle,
					FullPath);
		}
		//
		// Otherwise, if it's a file, query the file's name and security
		// information
		//
		else
		{
			ULONG FileObjectKey = ObjectKeyPool++;

			SaveFileSecurityDescriptor(
					LogFileHandle,
					FileObjectKey,
					FullPath);
			
			SaveObjectName(
					LogFileHandle,
					FileObjectKey,
					FullPath);
		}
		
		free(FullPath);

	} while (FindNextFile(FindHandle, &FindData) != 0);

	FindClose(
			FindHandle);
}

//
// Dump file ACL information
//
VOID 
DumpFileAcls()
{
	HANDLE LogFileHandle;
	DWORD ValidDriveCount;
	DWORD ValidDriveIndex;
	TCHAR ValidDrives[26];

	LogFileHandle = CreateFile(
				L"PersistedFile",
				GENERIC_WRITE,
				FILE_SHARE_READ,
				NULL,
				CREATE_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL);

	if (LogFileHandle == INVALID_HANDLE_VALUE)
	{
		wprintf(L"CreateFile('PersistedFile') failed, %lu\n", 
				GetLastError());
		return;
	}

	//
	// Get the valid drive letters in the system
	//
	ValidDriveCount = GetLogicalDriveStrings(
			sizeof(ValidDrives) / sizeof(TCHAR),
			ValidDrives);

	if (!ValidDriveCount)
	{
		wprintf(L"GetLogicalDriveStrings failed, %lu\n",
				GetLastError());
		return;
	}

	//
	// Walk through each of the valid drives.  When a fixed disk is encountered,
	// collect ACL information.
	//
	for (ValidDriveIndex = 0; 
	     ValidDriveIndex < ValidDriveCount; 
	     ValidDriveIndex++)
	{
		TCHAR CurrentDrive[3] = 
		{ 
			ValidDrives[ValidDriveIndex], 
			TEXT(':'), 
			0
		};

		//
		// Skip all non-fixed disks
		//
		if (GetDriveType(CurrentDrive) != DRIVE_FIXED)
			continue;

		//
		// Dump ACL information starting from the root of the drive
		//
		DumpSpecificDirectoryAcls(
				LogFileHandle,
				CurrentDrive);
	}

	//
	// GetFinalPathNameByHandle
	//

	if (LogFileHandle)
		CloseHandle(LogFileHandle);
}

//
// Dump registry key ACL information for a specific registry key
//
VOID 
DumpSpecificRegistryKeyAcls(
	__in HANDLE LogFileHandle,
	__in HKEY Key,
	__in PKEY_NAME_STACK CurrentStackEntry)
{
	PISECURITY_DESCRIPTOR SecurityDescriptor = NULL;
	PKEY_NAME_STACK IteratorEntry = NULL;
	KEY_NAME_STACK ChildStackEntry = { NULL, 0x80, CurrentStackEntry };
	ULONG64 ObjectKey = ObjectKeyPool++;
	LPTSTR FullKeyName = NULL;
	ULONG FullKeyNameCch;
	ULONG FullKeyNameCb;
	ULONG Index = 0;
	DWORD Status;

	//
	// Query ACL information for the supplied key
	//
	Status = GetSecurityInfo(
			Key,
			SE_REGISTRY_KEY,
			DACL_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | OWNER_SECURITY_INFORMATION,
			NULL,
			NULL,
			NULL,
			NULL,
			&SecurityDescriptor);

	if (Status != ERROR_SUCCESS)
		return;

	//
	// Saves a security descriptor as being associated with the supplied object
	// key
	//
	SaveObjectSecurityDescriptor(
			LogFileHandle,
			ObjectKey,
			SecurityDescriptor);

	LocalFree(
			SecurityDescriptor);

	//
	// Build the name metadata and write it to the file.  The first calculates
	// the size and the second pass populates the buffer.
	//
	FullKeyNameCch = 0;
	
	for (IteratorEntry = CurrentStackEntry;
	     IteratorEntry;
	     IteratorEntry = IteratorEntry->Parent)
	{ 
		// account for additional \ separator
		FullKeyNameCch += IteratorEntry->KeyNameSize + 1;
	}

	FullKeyNameCb = (FullKeyNameCch + 1) * sizeof(TCHAR);
	FullKeyName   = (LPTSTR)malloc(FullKeyNameCb);

	if (FullKeyName)
	{
		ZeroMemory(
				FullKeyName,
				FullKeyNameCb);

		for (IteratorEntry = CurrentStackEntry;
			  IteratorEntry;
			  IteratorEntry = IteratorEntry->Parent)
		{
			ULONG MoveOffset = wcslen(IteratorEntry->KeyName);
			ULONG SlashOffset = 0;

			if (IteratorEntry->Parent)
				SlashOffset = 1;

			MoveMemory(
					FullKeyName + MoveOffset + SlashOffset,
					FullKeyName,
					wcslen(FullKeyName) * sizeof(TCHAR));

			CopyMemory(
					FullKeyName + SlashOffset,
					IteratorEntry->KeyName,
					MoveOffset * sizeof(TCHAR));

			if (IteratorEntry->Parent)
				FullKeyName[0] = L'\\';
		}

		//
		// Save the current key's object name
		//
		SaveObjectName(
				LogFileHandle,
				ObjectKey,
				FullKeyName);

		free(FullKeyName);
	}

	//
	// Now that we've gotten this key's security information, let's query all of
	// it sub-keys
	//
	while (TRUE)
	{
		//
		// Grow the key name buffer as needed
		//
		if (ChildStackEntry.KeyNameSize)
		{
			ULONG AllocationSize;

			if (ChildStackEntry.KeyName)
				free(ChildStackEntry.KeyName);

			if (FAILED(ULongMult(
					ChildStackEntry.KeyNameSize + 1,
					sizeof(TCHAR),
					&AllocationSize)))
				break;

			ChildStackEntry.KeyName = malloc(
					AllocationSize);

			if (!ChildStackEntry.KeyName)
				break;
		}

		//
		// Query the key at this index
		//
		Status = RegEnumKey(
				Key,
				Index,
				ChildStackEntry.KeyName,
				ChildStackEntry.KeyNameSize);

		//
		// Grow our key name buffer as required
		//
		if (Status == ERROR_MORE_DATA)
		{
			if (FAILED(ULongMult(
					ChildStackEntry.KeyNameSize,
					2,
					&ChildStackEntry.KeyNameSize)))
				break;
			else
				continue;
		}
		else if (Status == ERROR_SUCCESS)
		{
			LONG SubStatus;
			HKEY ChildKey;

			SubStatus = RegOpenKeyEx(
					Key,
					ChildStackEntry.KeyName,
					0,
					KEY_READ,
					&ChildKey);

			if (SubStatus == ERROR_SUCCESS)
			{
				DumpSpecificRegistryKeyAcls(
						LogFileHandle,
						ChildKey,
						&ChildStackEntry);
			}

			RegCloseKey(
					ChildKey);

			Index++;
		}
		else
			break;
	}

	if (ChildStackEntry.KeyName)
		free(ChildStackEntry.KeyName);
}

//
// Dump registry key ACL information
//
VOID 
DumpRegistryKeyAcls()
{
	KEY_NAME_STACK HklmStack;
	KEY_NAME_STACK HkuStack;
	HANDLE LogFileHandle;

	LogFileHandle = CreateFile(
				L"PersistedKey",
				GENERIC_WRITE,
				FILE_SHARE_READ,
				NULL,
				CREATE_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL);

	if (LogFileHandle == INVALID_HANDLE_VALUE)
	{
		wprintf(L"CreateFile('PersistedKey') failed, %lu\n", 
				GetLastError());
		return;
	}

	//
	// Dump HKEY_LOCAL_MACHINE
	//
	HklmStack.KeyName = L"HKEY_LOCAL_MACHINE";
	HklmStack.KeyNameSize = wcslen(L"HKEY_LOCAL_MACHINE");
	HklmStack.Parent = NULL;

	DumpSpecificRegistryKeyAcls(
			LogFileHandle,
			HKEY_LOCAL_MACHINE,
			&HklmStack);

	//
	// Dump HKEY_USERS
	//
	HkuStack.KeyName = L"HKEY_USERS";
	HkuStack.KeyNameSize = wcslen(L"HKEY_USERS");
	HkuStack.Parent = NULL;

	DumpSpecificRegistryKeyAcls(
			LogFileHandle,
			HKEY_USERS,
			&HkuStack);

	if (LogFileHandle)
		CloseHandle(LogFileHandle);
}

//
// Get security descriptor metadata for a service
//
BOOL 
GetServiceSecurityDescriptorMetadata(
	__in SC_HANDLE Service,
	__in ULONG ObjectKey,
	__out POBJECT_SECURITY_DESCRIPTOR_METADATA *OutSdMetadata)
{
	POBJECT_SECURITY_DESCRIPTOR_METADATA SdMetadata = NULL;
	ULONG SdMetadataLength = 0;
	BOOL Success = FALSE;

	do
	{
		//
		// Allocate our security descriptor metadata if we know what size we
		// should allocate it at
		//
		if (SdMetadataLength)
		{
			if (SdMetadata)
				free(SdMetadata);

			SdMetadataLength += sizeof(OBJECT_SECURITY_DESCRIPTOR_METADATA);

			SdMetadata = (POBJECT_SECURITY_DESCRIPTOR_METADATA)malloc(
					SdMetadataLength);

			if (!SdMetadata)
				break;

			//
			// Initialize the structure
			//
			SdMetadata->Header.Tag    = OBJECT_METADATA_TAG_SECURITY_DESC;
			SdMetadata->Header.Length = SdMetadataLength;
			SdMetadata->Header.Object = ObjectKey;
			SdMetadata->SdLength      = SdMetadataLength - sizeof(OBJECT_SECURITY_DESCRIPTOR_METADATA);
		}

		//
		// Query dacl, sacl, group, and owner information from the service.  This
		// call will directly copy the self-relative security descriptor into the
		// body of this metadata entry
		//
		Success = QueryServiceObjectSecurity(
				Service,
				DACL_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | OWNER_SECURITY_INFORMATION,
				SdMetadata
					? SdMetadata + 1
					: NULL,
				SdMetadataLength 
					? SdMetadataLength - sizeof(OBJECT_SECURITY_DESCRIPTOR_METADATA)
					: 0,
				&SdMetadataLength);

	} while ((!Success) && (GetLastError() == ERROR_INSUFFICIENT_BUFFER));

	if ((!Success) && (SdMetadata))
	{
		free(SdMetadata);
		SdMetadata = NULL;
	}	

	*OutSdMetadata = SdMetadata;

	return Success;
}

//
// Dumps service object metadata
//
VOID 
DumpServiceAcls()
{
	LPENUM_SERVICE_STATUS ServiceStatus = NULL;
	SC_HANDLE Scm = NULL;
	HANDLE LogFileHandle = NULL;
	ULONG ServiceStatusSize = 0;
	ULONG NumberOfServices;
	ULONG Index;
	BOOL Success;

	do
	{
		//
		// Create the log file that we will write our object metadata to
		//
		LogFileHandle = CreateFile(
				L"PersistedService",
				GENERIC_WRITE,
				FILE_SHARE_READ,
				NULL,
				CREATE_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL);

		if (LogFileHandle == INVALID_HANDLE_VALUE)
		{
			wprintf(L"CreateFile('PersistedService') failed, %lu\n", 
					GetLastError());
			break;
		}

		//
		// Open the SCM
		//
		Scm = OpenSCManager(
				NULL,
				NULL,
				SC_MANAGER_ENUMERATE_SERVICE);

		if (!Scm)
		{
			wprintf(L"OpenSCManager failed, %lu\n", GetLastError());
			break;
		}

		do
		{
			//
			// If we know the size of our allocation, allocate it now
			//
			if (ServiceStatusSize)
			{
				if (ServiceStatus)
					free(ServiceStatus);

				ServiceStatus = (LPENUM_SERVICE_STATUS)malloc(
						ServiceStatusSize);

				if (!ServiceStatus)
					break;
			}

			//
			// Enumerate services and pull out their information
			//
			Success = EnumServicesStatus(
					Scm,
					SERVICE_DRIVER | SERVICE_WIN32,
					SERVICE_STATE_ALL,
					ServiceStatus,
					ServiceStatusSize,
					&ServiceStatusSize,
					&NumberOfServices,
					NULL);

		} while ((!Success) && (GetLastError() == ERROR_MORE_DATA));

		if ((!Success) || (!ServiceStatus))
		{
			wprintf(L"EnumServiceStatus failed, %lu\n",
					GetLastError());
			break;
		}

		//
		// Now that we have service information, walk the array and open each
		// service so that we can pull out their respective security descriptors.
		//
		for (Index = 0; 
		     Index < NumberOfServices;
		     Index++)
		{
			POBJECT_SECURITY_DESCRIPTOR_METADATA SdMetadata;
			POBJECT_NAME_METADATA NameMetadata;
			SC_HANDLE Service;
			ULONG ObjectKey = ObjectKeyPool++;
			ULONG NameMetadataSize;
			ULONG BytesWritten;

			Service = OpenService(
					Scm,
					ServiceStatus[Index].lpServiceName,
					READ_CONTROL);

			if (!Service)
			{
				wprintf(L"OpenService('%s') failed, %lu\n",
						ServiceStatus[Index].lpServiceName,
						GetLastError());
				continue;
			}

			//
			// Query & log the service's security descriptor information
			//
			if (GetServiceSecurityDescriptorMetadata(
					Service,
					ObjectKey,
					&SdMetadata))
			{
				WriteFile(
						LogFileHandle,
						SdMetadata,
						SdMetadata->Header.Length,
						&BytesWritten,
						NULL);
			}
			else
			{
				wprintf(L"GetServiceSecurityDescriptorMetadata('%s') failed, %lu\n",
						ServiceStatus[Index].lpServiceName,
						GetLastError());
			}


			//
			// Save the service's name information
			//
			SaveObjectName(
					LogFileHandle,
					ObjectKey,
					ServiceStatus[Index].lpServiceName);

			CloseServiceHandle(
					Service);
		}

	} while (0);

	if (Scm)
		CloseServiceHandle(
				Scm);
	if (LogFileHandle)
		CloseHandle(
				LogFileHandle);
}

INT 
main(INT Argc, PCHAR *Argv)
{
	wprintf(L"Dumping file ACLs...\n");
	DumpFileAcls();

	wprintf(L"Dumping registry key ACLs...\n");
	DumpRegistryKeyAcls();

	wprintf(L"Dumping service ACLs...\n");
	DumpServiceAcls();

	return 0;
}
