#!/usr/bin/env ruby
#
# Trust boundary analysis interface
#
# mmiller@hick.org
# 4/2008
#

require 'optparse'
require 'tba/analyzer'

#
# Console class
#
class AnalyzerConsole

	FlowIndent   = "  - "
	DetailIndent = "    > "

	def initialize
		@sids       = []
		@trace_logs = []
		@details    = []
		@dump_ahorizon = false
		@dump_dhorizon = false
		
		# Create an instance of the analyzer and process the specified traces
		@tba = TBA::Analyzer.new
	
		# Parse command line options that were specified
		parse_options
	end

	def run
		# Process the trace logs that were specified
		@tba.process_traces(
			*@trace_logs)

		# Should we generate a trust boundary graph
		if @gen_tbg_path

			File.open(@gen_tbg_path, "w") { |fd|
				@tba.build_tbg.to_dot_graph('fd' => fd)
			}

		# Should we generate a trust boundary summary graph?
		elsif @gen_tbsg_path

			File.open(@gen_tbsg_path, "w") { |fd|
				@tba.build_tbg.summary_graph.to_dot_graph('fd' => fd)
			}

		# Are we checking for elevation paths between SIDs?
		elsif @dump_paths
			dump_elevation_paths(
				@fsid,
				@tsid)

		# Otherwise, we're just providing data
		else
			dump_attack_horizons  if @dump_ahorizon
			dump_defense_horizons if @dump_dhorizon
		end
	end

	# In reality, this should be implemented in terms of a graph traversal.
	# I'll do that in the future.
	def dump_elevation_paths(from_sid, to_sid)
		if from_sid.nil? or to_sid.nil?
			puts "You must specify a from SID (-F) and a to SID (-T)"
			return
		end

		puts 
		puts "Dumping elevation paths from #{from_sid} to #{to_sid}..."
		puts

		# Initialize our worklist starting from our from SID.  We'll look at
		# each SID's attack horizon trying to find paths to the to_sid. We could
		# go the other direction as well which might be more optimal.  It really
		# depends on which SID has a smaller attack/defense horizon.
		wl =  []
		wl << [ from_sid, [] ]

		# Set of SIDs that have been visited
		visited_sids = Set.new
		queued_sids  = Set.new

		until wl.empty?
			# Grab the current SID and path from the top of the stack
			current_sid, path_stack = wl.pop

			# We won't visit this SID again
			visited_sids << current_sid

			# Enumerate through the attack horizon for the current SID
			@tba.each_attack_horizon(current_sid) { |definition, use, threats|

				# Skip any attack paths that don't allow us to potentially elevate privileges
				#next if threats.in_priv.include?(TBA::Threats::ElevationOfPrivilege) == false

				# If this is the SID we are looking for then we'll display the
				# path.  Alternatively, if we have yet to visit the target SID,
				# then we will do so
				if visited_sids.include?(use.sid) == false

					dup_path_stack = path_stack.dup
					dup_path_stack << [definition, use, threats]

					# If we've reached the SID we were trying to reach, dump
					# information about the elevation path
					if use.sid == to_sid 

						puts
						puts "Elevation path from #{from_sid} to #{to_sid}:"

						dup_path_stack.reverse.each { |elevation_path_element|
							definition, use, threats = elevation_path_element

							dump_dataflow('attack', from_sid, definition, use, threats)
						}

					# Otherwise, we'll visited from this point forward
					else
						wl.push([ use.sid, dup_path_stack ])
					end

				end

			}
		end
	end

	# Dump the defense horizons for the specified SIDs
	def dump_defense_horizons
		@sids.each { |sid|

			puts
			puts "Defense horizon for #{sid}"
			puts

			@tba.each_defense_horizon(sid) { |definition, use, threats|
				dump_dataflow('defense', sid, definition, use, threats)
			}

		}
	end

	# Dump the attack horizons for the specified SIDs
	def dump_attack_horizons
		@sids.each { |sid|

			puts
			puts "Attack horizon for #{sid}"
			puts

			@tba.each_attack_horizon(sid) { |definition, use, threats|
				dump_dataflow('attack', sid, definition, use, threats)
			}
		}
	end

	# Dump a specific data flow
	def dump_dataflow(mode, sid, definition, use, threats)
		actual_threats = threats.in_priv_to_s

		puts "#{FlowIndent}#{definition} -> #{use} leading to #{actual_threats}"

		dump_dataflow_detail(definition)
		dump_dataflow_detail(use)
	end

	# Dump detail about a given flow
	def dump_dataflow_detail(flow)
		return if @details.length == 0

		# Get the metadata associated with this flow
		md = flow.metadata

		return if md.nil?

		puts "#{DetailIndent}Details for #{flow}"
		@details.each { |detail|

			case detail

				# Dump execution context information such as process name and
				# actualized tokens
				when "pinfo"
					if md.is_access?
						sc = md.security_context

						puts "#{DetailIndent}  Process: #{@tba.process_name(md.process_object_key)}"
						puts "#{DetailIndent}  Client Token: O:#{sc.client_token_owner_sid} G:#{sc.client_token_primary_group_sid}" if sc.client_token_owner_sid
						puts "#{DetailIndent}  Primary Token: O:#{sc.primary_token_owner_sid} G:#{sc.primary_token_primary_group_sid}" if sc.primary_token_owner_sid
					end

				# Dump the call stack that was used to 
				when "callstack"
					next if flow.can?
					
					puts "#{DetailIndent}  Call stack:"
					md.call_stack.frames.each { |frame|

						rip_str = @tba.find_image_rva(
							md.process_object_key,
							frame.rip)

						puts "#{DetailIndent}    #{rip_str}"
					}

			end

		}
	end

	# Parse command line options
	def parse_options
		opts = OptionParser.new { |opts|
			opts.banner = "Usage: analyzer [options]"

			opts.separator ""
			opts.separator "Specific options:"

			opts.on("-d", "-d", "Dump the defense horizon for one or more SIDs") { 
				@dump_dhorizon = true
			}

			opts.on("-a", "-a", "Dump the attack horizon for one or more SIDs") { 
				@dump_ahorizon = true
			}

			opts.on("-s", "-s <sid1,sid2,sid3>", "The SIDs to operate on (applies to -d, -a)") { |sids|
				sids.split(',').each { |sidstr|
					@sids << Sid.lookup_string(sidstr)
				}
			}

			opts.on("-i", "-i <log1,log2,log3>", "Path to one or more object trace logs to parse") { |logs|
				@trace_logs |= logs.split(',')
			}

			opts.on("-o", "-o <summary,callstack,pinfo>", "Specifies the output detail to that should be displayed") { |detail|
				@details = detail.split(",")
			}

			opts.on("-t", "-t <actualized|potential|all>", "Ignore results that are tied to actualized, potential, or both threats") { |threat_filter|
				@tba.ignore_actualized  = true if threat_filter == 'actualized'  or threat_filter == 'all'
				@tba.ignore_potential = true if threat_filter == 'potential' or threat_filter == 'all'
			}

			opts.on("-F", "-F sid", "Source SID for dumping elevation paths") { |fsidstr|
				@dump_paths = true
				@fsid = Sid.lookup_string(fsidstr)
			}

			opts.on("-T", "-T sid", "Target SID for dumping elevation paths") { |tsidstr|
				@dump_paths = true
				@tsid = Sid.lookup_string(tsidstr)
			}

			opts.on("-m", "-m <sid membership file>", "SID equivalence classes file") { |path|
				@tba.sid_membership_file = path
			}

			opts.on("-g", "-g <tbg>", "Generate a trust boundary graph") { |path|
				@gen_tbg_path = path
			}

			opts.on("-G", "-G <tbsg>", "Generate a trust boundary summary graph") { |path|
				@gen_tbsg_path = path
			}

		}

		opts.parse!(ARGV)
	end

end

AnalyzerConsole.new.run
