#!/usr/bin/env ruby
#
# Ruby implementation of classes that encapsulate the Windows SDDL.
#
# skape
# mmiller@hick.org
# 4/2008
#

# Simple cache table class
class CacheTable

	# Find an object find a raw description
	def self.lookup(table, klass, entry_raw)
		entry = table[entry_raw]

		if entry.nil?
			entry = klass.new(entry_raw)

			table[entry_raw] = entry
		end

		return entry
	end

end

#
# SID - Security Identifier
#
class Sid

	SidTable = {}

	# Well-known SID lookup table
	WellKnownSids =
		{
			"S-1-0"        => [ "Null Authority",                 nil  ],
			"S-1-0-0"      => [ "Nobody",                         nil  ],
			"S-1-1"        => [ "World Authority",                nil  ],
			"S-1-1-0"      => [ "Everyone",                       "WD" ],
			"S-1-2"        => [ "Local Authority",                nil  ],
			"S-1-3"        => [ "Creator Authority",              nil  ],
			"S-1-3-0"      => [ "Creator Owner",                  "CO" ],
			"S-1-3-1"      => [ "Creator Group",                  "CG" ],
			"S-1-3-2"      => [ "Creator Owner Server",           nil  ],
			"S-1-3-3"      => [ "Creator Group Server",           nil  ],
			"S-1-4"        => [ "Non-unique Authority",           nil  ],
			"S-1-5"        => [ "NT Authority",                   nil  ],
			"S-1-5-1"      => [ "Dialup",                         nil  ],
			"S-1-5-2"      => [ "Network",                        "NU" ],
			"S-1-5-3"      => [ "Batch",                          nil  ],
			"S-1-5-4"      => [ "Interactive",                    "IU" ],
			"S-1-5-6"      => [ "Service",                        "SU" ],
			"S-1-5-7"      => [ "Anonymous",                      "AN" ],
			"S-1-5-8"      => [ "Proxy",                          nil  ],
			"S-1-5-9"      => [ "Enterprise Domain Controllers",  "ED" ],
			"S-1-5-10"     => [ "Principal Self",                 nil  ],
			"S-1-5-11"     => [ "Authenticated Users",            "AU" ],
			"S-1-5-12"     => [ "Restricted Code",                "RC" ],
			"S-1-5-13"     => [ "Terminal Server Users",          "RD" ],
			"S-1-5-17"     => [ "IUSRS",                          nil  ],
			"S-1-5-18"     => [ "LocalSystem",                    "SY" ],
			"S-1-5-19"     => [ "Local Service",                  "LS" ],
			"S-1-5-20"     => [ "Network Service",                "NS" ],
			"S-1-5-32-544" => [ "Administrators",                 "BA" ],
			"S-1-5-32-545" => [ "Users",                          "BU" ],
			"S-1-5-32-546" => [ "Guests",                         "BG" ],
			"S-1-5-32-547" => [ "Power Users",                    "PU" ],
			"S-1-5-32-548" => [ "Account Operators",              "AO" ],
			"S-1-5-32-549" => [ "Server Operators",               "SO" ],
			"S-1-5-32-550" => [ "Print Operators",                "PO" ],
			"S-1-5-32-551" => [ "Backup Operators",               "BO" ],
			"S-1-5-32-552" => [ "Replicators",                    "RE" ],
			"S-1-16-4096"  => [ "Low Integrity",                  "LW" ],
			"S-1-16-8192"  => [ "Medium Integrity",               "ME" ],
			"S-1-16-12288" => [ "High Integrity",                 "HI" ],
			"S-1-16-16384" => [ "System Integrity",               "SI" ]
		}

	# Well-known variable SID lookup table
	WellKnownVariableSids =
		[
			[ /S-1-5-5-\d+-\d+/, [ "Logon Session",      "lS" ] ], # differentiate
			[ /S-1-5-\d+-500/,   [ "Administrator"       "LA" ] ],
			[ /S-1-5-\d+-501/,   [ "Guest",              "LG" ] ],
			[ /S-1-5-\d+-502/,   [ "KRBTGT"                   ] ],
			[ /S-1-5-\d+-503/,   [ "Standard User",      "LU" ] ],
			[ /S-1-5-\d+-512/,   [ "Domain Admins",      "DA" ] ],
			[ /S-1-5-\d+-513/,   [ "Domain Users",       "DU" ] ],
			[ /S-1-5-\d+-514/,   [ "Domain Guests",      "DG" ] ],
			[ /S-1-5-\d+-515/,   [ "Domain Computers",   "DC" ] ],
			[ /S-1-5-\d+-516/,   [ "Domain Controllers", "DD" ] ],
			[ /S-1-5-\d+-517/,   [ "Cert Publishers"          ] ],
		]

	# Constant SIDs
	def self.wd_sid
		Sid.lookup_string("S-1-1-0")
	end

	def self.sy_sid
		Sid.lookup_string("S-1-5-18")
	end

	def self.ba_sid
		Sid.lookup_string("S-1-5-32-544")
	end
	
	def self.ao_sid
		Sid.lookup_string("S-1-5-32-548")
	end
	
	def self.so_sid
		Sid.lookup_string("S-1-5-32-549")
	end
	
	def self.po_sid
		Sid.lookup_string("S-1-5-32-550")
	end
	
	def self.bo_sid
		Sid.lookup_string("S-1-5-32-551")
	end
	
	def self.ls_sid
		Sid.lookup_string("S-1-5-19")
	end

	def self.ns_sid
		Sid.lookup_string("S-1-5-20")
	end

	# Lookup a sid in the SID cache using the raw SID buffer
	def self.lookup(sid_raw)
		CacheTable.lookup(SidTable, Sid, sid_raw)
	end

	# Lookup a sid in the SID cache using a supplied SID string
	def self.lookup_string(sid_str)
		sid = SidTable[sid_str]

		if sid.nil?

			# If the SID string is not in standard form, then maybe the user
			# provided us will an alias for the SID.
			if sid_str !~ /^S-/
				WellKnownSids.each_pair { |rsid_str, cinfo|
					if cinfo[0] == sid_str or cinfo[1] == sid_str
						sid = from_sid_string(rsid_str)
						break
					end
				}
			end

			sid = from_sid_string(sid_str) if sid.nil?

			SidTable[sid_str] = sid
		end

		sid
	end

	# Creates a SID instance from a printable SID string
	def self.from_sid_string(sid_str)
		sid_raw   = ''
		sid_parts = sid_str.split(/-/)

		garbage   = sid_parts.shift
		revision  = sid_parts.shift.to_i
		id_auth   = sid_parts.shift.to_i
		sid_raw  << [ revision, sid_parts.length, 0, id_auth ].pack("CCvN")

		sid_parts.each { |part|
			sid_raw << [ part.to_i ].pack('V')
		}

		Sid.lookup(sid_raw)
	end

	# Initializes a SID from a raw SID buffer
	def initialize(sid_raw)
		@revision, sub_count = 
			sid_raw.unpack("CC")

		@identifier_authority = sid_raw[7]

		@sub_authorities = []

		sub_count.times { |cnt|
			@sub_authorities << sid_raw[8+(4*cnt),4].unpack("V")
		}
	end

	# Converts the SID to a string
	def to_s
		s = "S-#{@revision}-#{@identifier_authority}"

		@sub_authorities.each { |sub|
			s << "-#{sub}"
		}

		canonical_sid_string(s)
	end

	# Converts a sid string to a canonical (well-known) name
	def canonical_sid_string(sid_str)
		# Look up the name as if this were a static SID
		canonical_info = WellKnownSids[sid_str]

		# If we didn't find a well known name for this sid, try looking at
		# variable names
		if canonical_info.nil?

			WellKnownVariableSids.each { |vsid|

				if vsid[0].match(sid_str)
					canonical_info = vsid[1] 
					break
				end

			}

		end	

		# If we have canonical information for this SID, try to use the short
		# name first.  Otherwise, use the long name.  If we don't have canonical
		# information, we just use the SID as a string.
		if canonical_info
			canonical = canonical_info[1] || canonical_info[0]
		else
			canonical = sid_str
		end

		return canonical
	end

	def eql?(other)
		other.to_s == to_s
	end

	attr_reader :revision
	attr_reader :identifier_authority
	attr_reader :sub_authorities

end

# 
# ACE - Access Control Entry
#
class Ace

	# Global cache table for ACEs
	AceTable = {}

	# Maps ACE types to a string representation
	TypeString =
	{
		0  => "A",  # Access allowed
		1  => "D",  # Access denied
		5  => "OA", # Object access allowed
		6  => "OD", # Object access denied
		2  => "AU", # Audit
		3  => "AL", # Alarm
		7  => "OU", # Object audit
		8  => "OL", # Object alarm
		17 => "ML", # Integrity label
	}

	# Looks up an ACE in the global cache table using the raw ACE buffer
	def self.lookup(ace_raw)
		CacheTable.lookup(AceTable, Ace, ace_raw)
	end

	# Returns the length of the ace given a raw ACE buffer
	def self.ace_length(ace_raw)
		Ace.new(ace_raw).size
	end

	# Initializes an ACE from a raw ACE buffer
	def initialize(ace_raw)
		@type, @flags, @size, @rights = ace_raw.unpack("CCvV")

		@sid = Sid.lookup(ace_raw[8, @size - 8])
	end

	# Converts the ACE to its string representation
	def to_s
		s = ''
		s << "(#{TypeString[@type] || @type};"
		s << "#{flags_to_s};"
		s << "#{rights_to_s};"
		s << ";;"
		s << "#{sid})"
		s
	end

	# Is this an allow ace?
	def is_allow?
		@type == 0
	end

	# Is this a deny ace?
	def is_deny?
		@type == 1
	end

	# ACE type
	attr_reader :type
	# ACE flags
	attr_reader :flags
	# ACE size, in bytes
	attr_reader :size
	# ACE rights
	attr_reader :rights
	# ACE SID
	attr_reader :sid

private

	# Convert ace flags to readable string form
	def flags_to_s
		s = ''
		s << "OI" if (@flags & 0x01) != 0 # object inherit
		s << "CI" if (@flags & 0x02) != 0 # container inherit
		s << "NP" if (@flags & 0x04) != 0 # non-propagating
		s << "IO" if (@flags & 0x08) != 0 
		s << "ID" if (@flags & 0x10) != 0
		s << "SA" if (@flags & 0x40) != 0
		s << "FA" if (@flags & 0x80) != 0
		s
	end

	# Convert ace rights to readable string form
	def rights_to_s
		s = ''
		s << "SD" if (@rights & 0x00010000) != 0 # standard delete
		s << "RC" if (@rights & 0x00020000) != 0 # read control
		s << "WD" if (@rights & 0x00040000) != 0 # write dac
		s << "WO" if (@rights & 0x00080000) != 0 # write owner
		s << "SY" if (@rights & 0x00100000) != 0 # synchronize
		s << "GA" if (@rights & 0x10000000) != 0 # generic all
		s << "GX" if (@rights & 0x20000000) != 0 # generic execute
		s << "GW" if (@rights & 0x40000000) != 0 # generic write
		s << "GR" if (@rights & 0x80000000) != 0 # generic read
		s << "#{"SR(%.4x)" % (@rights & 0xffff)}" if (@rights & 0xffff) != 0
		s
	end

end

# 
# ACL - Access Control List
#
class Acl

	AclTable = {}

	# Lookup an ACL entry
	def self.lookup(acl_raw)
		CacheTable.lookup(AclTable, Acl, acl_raw)
	end

	# The length of the entire ACL
	def self.acl_length(acl_raw)
		Acl.new(acl_raw).size
	end

	# Initializes an ACL from a raw ACL buffer
	def initialize(acl_raw)
		@acl_revision, @sbz1, @size, @count, @sbz2 = 
			acl_raw.unpack("CCvvv")

		@aces  = []

		# Chomp off the head
		ace_raw = acl_raw[8, @size - 8]
		offset  = 0
		length  = ace_raw.length

		@count.times { |cnt|
			ace_len = Ace.ace_length(ace_raw[offset, length - offset])
			cur_ace = ace_raw[offset, ace_len]

			@aces << Ace.lookup(ace_raw[offset, ace_len])

			offset += ace_len
		}
	end

	# Converts the ACL to its string representation
	def to_s
		s = ''
		@aces.each { |ace|
			s << ace.to_s
		}
		s
	end

	attr_reader :acl_revision
	attr_reader :sbz1
	attr_reader :size
	attr_reader :count
	attr_reader :sbz2
	attr_reader :aces

end

# 
# A windows security descriptor (SD) composed of an owner SID, a group SID, a
# DACL, and a SACL.
#
class WindowsSecurityDescriptor

	SdTable = {}

	# Get the security descriptor encapsulating the supplied raw descriptor
	def self.lookup(sd_raw)
		CacheTable.lookup(SdTable, WindowsSecurityDescriptor, sd_raw)
	end
	
	# The length of a security descriptor -- assumed to be the full length of
	# the provided buffer
	def self.sd_length(sd_raw)
		sd_raw.length	
	end

	def initialize(sd_raw)
		@revision, @sbzl, @control, owner_off, group_off, sacl_off, dacl_off =
			sd_raw.unpack("CCvVVVV")

		@owner_sid = Sid.lookup(sd_raw[owner_off, sd_raw.length - owner_off]) if owner_off > 0
		@group_sid = Sid.lookup(sd_raw[group_off, sd_raw.length - group_off]) if group_off > 0
		@sacl      = Acl.lookup(sd_raw[sacl_off,  sd_raw.length - sacl_off]) if sacl_off > 0
		@dacl      = Acl.lookup(sd_raw[dacl_off,  sd_raw.length - dacl_off]) if dacl_off > 0
	end

	def to_s
# FIXME: why do these seem swapped in the metadata log?
#		s << "O:#{owner_sid}" if @owner_sid 
#		s << "G:#{group_sid}" if @group_sid

		s = ''
		s << "O:#{group_sid}" if @group_sid
		s << "G:#{owner_sid}" if @owner_sid 
		s << "D:#{dacl}" if @dacl
		s << "S:#{sacl}" if @sacl
		s
	end

	attr_reader :revision
	attr_reader :sbzl
	attr_reader :control
	attr_reader :owner_sid
	attr_reader :group_sid
	attr_reader :sacl
	attr_reader :dacl

end
