#!/usr/bin/env ruby

require 'common/sddl'

module ObjectTrace

class Metadata

	HeaderLength = 0x14

	def self.from_log(log)
		# Read the header of the next metadata entry
		data = log.read_raw(HeaderLength)

		raise EOFError if data.nil?
		raise EOFError if data.length < HeaderLength

		# Grab the tag and the length from the header
		tag, length = data.unpack("QV")

		# Read the body of the metadata entry
		data << log.read_raw(length - HeaderLength)

		# If we can't read the entire buffer, EOF
		raise EOFError if data.length < length

		# Create a metadata instance based on the tag of the entry
		metadata = nil

		case tag & 0xf
			when Access::Tag
				metadata = Access.new(data)
			when SecurityContext::Tag
				metadata = SecurityContext.new(data)
			when SecurityDescriptor::Tag
				metadata = SecurityDescriptor.new(data)
			when Name::Tag
				metadata = Name.new(data)
			when CallStack::Tag
				metadata = CallStack.new(data)
			when ImageInfo::Tag
				metadata = ImageInfo.new(data)
			else
				raise "Invalid tag encountered: #{tag}"
		end

		# Associate the metadata with the log it was read from
		metadata.log = log

		return metadata
	end

	# Initializes the base attributes associated with all metadata
	def initialize(data)
		tagseq, @length, @object_key = data.unpack("QVQ")

		@tag      = tagseq & 0xf
		@sequence = tagseq >> 4

		# Call deserialize to give derive classes a chance to process the body
		deserialize(data[HeaderLength, @length - HeaderLength])
	end

	# Deserializes the body of a metadata element
	def deserialize(body)
	end

	# Reads the next metadata entry from the log that this metadata was read
	# from
	def next
		log.read_metadata
	end

	# The object's type, as a string (e.g. Process)
	def object_type
		log.object_type
	end

	attr_reader :tag
	attr_reader :sequence
	attr_reader :length
	attr_reader :object_key
	attr_accessor :log

	def is_access?
		tag == Access::Tag
	end

	def is_name?
		tag == Name::Tag
	end

	class Access < Metadata
		Tag = 7

		def deserialize(body)
			@process_object_key, @access_flags, @access_mask, sc_len, sd_len, cs_len = body.unpack("QVVVVV")

			offset = 28

			# Deserialize the security context metadata
			if sc_len > 0
				@security_context = SecurityContext.new(body[offset,sc_len])
				offset += sc_len
			end

			# Deserialize the security descriptor metadata
			if sd_len > 0
				@security_descriptor = SecurityDescriptor.new(body[offset,sd_len])
				offset += sd_len
			# If there is no security descriptor and this is a definition, then
			# we create an empty security descriptor
			elsif is_def?
				@security_descriptor = SecurityDescriptor.new(nil)
			end

			# Deserialize the call stack metadata
			if cs_len > 0
				@call_stack = CallStack.new(body[offset,cs_len])
			end
		end

		# Is this access an object definition?
		def is_def?
			(@access_flags & 0x1) != 0
		end

		# Is this access an object use?
		def is_use?
			!is_def?
		end

		def to_s
			str = "Access"
			str << " DEF" if is_def?
			str << " USE" if is_use?
			str
		end

		# Attributes specific to the access metadata
		attr_reader :process_object_key
		attr_reader :access_flags
		attr_reader :access_mask

		# Sub-metadata elements
		attr_reader :security_context
		attr_reader :security_descriptor
		attr_reader :call_stack
	
		# Populated after-the-fact
		attr_accessor :process_object
	end

	class SecurityContext < Metadata
		Tag = 2

		def deserialize(body)
			@impersonation_level, cto_len, ctg_len, pto_len, ptg_len = body.unpack("VVVVV")

			offset = 20

			# FIXME: why do these seem swapped in the metadata log?
			
			if cto_len > 0
#				@client_token_owner_sid = Sid.lookup(body[offset,cto_len])
				@client_token_primary_group_sid = Sid.lookup(body[offset,cto_len])
				offset += cto_len
			end

			if ctg_len > 0
#				@client_token_primary_group_sid = Sid.lookup(body[offset,ctg_len])
				@client_token_user_sid = Sid.lookup(body[offset,ctg_len])
				offset += ctg_len
			end

			if pto_len > 0
#				@primary_token_owner_sid = Sid.lookup(body[offset,pto_len])
				@primary_token_primary_group_sid = Sid.lookup(body[offset,pto_len])
				offset += pto_len
			end

			if ptg_len > 0
#				@primary_token_primary_group_sid = Sid.lookup(body[offset,ptg_len])
				@primary_token_owner_sid = Sid.lookup(body[offset,ptg_len])
				offset += ptg_len
			end
		end

		attr_reader :impersonation_level
		attr_reader :client_token_owner_sid
		attr_reader :client_token_primary_group_sid
		attr_reader :primary_token_owner_sid
		attr_reader :primary_token_primary_group_sid
	end

	class SecurityDescriptor < Metadata
		Tag = 3

		def initialize(data)
			if data.nil?
				sd = nil
			else
				super
			end
		end

		def deserialize(body)
			sd_len = body.unpack("V")[0]

			if sd_len > 0
				@sd = WindowsSecurityDescriptor.lookup(body[4,sd_len])
			end
		end

		attr_reader :sd
	end

	class Name < Metadata
		Tag = 4

		def deserialize(body)
			name_len = body.unpack("v")[0]

			if name_len > 0
				@name = body[2,name_len].gsub("\x00", '')
			end
		end

		attr_reader :name
	end

	class CallStack < Metadata
		Tag = 5

		class StackFrame
			def initialize(rsp, rip, image_base)
				@rsp = rsp
				@rip = rip
				@image_base = image_base
			end

			def to_s
				"#{rip}"
			end

			attr_reader :rsp
			attr_reader :rip
			attr_reader :image_base
		end

		def deserialize(body)
			@frames = []

			return if body.nil?

			@call_stack_flags, @frame_count = body.unpack("VV")

			offset = 8

			while offset < body.length
				rsp, rip, image_base = body[offset, 24].unpack("QQQ")

				frames << StackFrame.new(rsp, rip, image_base)

				offset += 24
			end
		end

		def is_use?
			(call_stack_flags & 0x1) != 0
		end

		def dump(object)
			str = ''

			frames.each { |frame|
				
				# Find the image RVA associated with RIP at the time of this
				# invocation
				ripstr = find_image_rva(
					object,
					object.process_object,
					frame.rip)

				str << "\tRSP=%.16x RIP=#{ripstr}\n" % frame.rsp
			}

			return str
		end

		# Finds an image-relative RVA for an instruction pointer
		def find_image_rva(object, process_object, rip)
			# Try to search within the process object that was provided
			ripstr = find_image_rva_in_process(
				process_object,
				rip)

			# If we haven't found the RIP string and we have a valid system process,
			# try to search there
			if ripstr.nil? and object.object_space.system_process
				ripstr = find_image_rva_in_process(
					object.object_space.system_process,
					rip)
			end

			# If we didn't find an offset, the we just use the full address
			ripstr = "%.16x" % rip if ripstr.nil?

			return ripstr
		end

		# Searches the images within a process context for the image containing the
		# provided RIP
		def find_image_rva_in_process(process_object, rip)
			if process_object and process_object.images
				process_object.images.each { |image|
					if rip >= image.base_address and rip < image.base_address + image.size
						return "#{image.name}+#{"%x" % (rip - image.base_address)}"
					end
				}
			end

			nil
		end

		attr_reader :call_stack_flags
		attr_reader :frame_count
		attr_reader :frames
	
	end

	class ImageInfo < Metadata
		Tag = 6

		def deserialize(body)
			@base_address, @size, name_len = body.unpack("QQv")

			if name_len > 0
				@name = body[18,name_len].gsub("\x00", '')
			end
		end

		attr_reader :base_address
		attr_reader :size
		attr_reader :name
	end
	
end

end
