#!/usr/bin/env ruby

require 'rgl'

require 'common/sddl'
require 'objtrace/metadata'

module ObjectTrace

#
# Encapsulates an object trace log file and provides the routines needed to
# read metadata from a log
#
class ObjectLog

	def initialize(path)
		@object_type = File.basename(path).sub("Persisted", '')
		@fd = File.new(path, "rb")
		@read_buffer = ''
	end

	def read_metadata
		Metadata.from_log(self)
	end

	def read_raw(length)

		# Pull the data from the read buffer based on what is available
		if @read_buffer.length >= length
			raw_buffer = @read_buffer.slice!(0..(length-1))
		else
			raw_buffer = @read_buffer.slice!(0..(@read_buffer.length-1))
		end

		# Calculate the length left to read
		if length > raw_buffer.length
			remaining_length = length - raw_buffer.length
			pad_length       = (remaining_length + 0x1ffff) & ~0x1ffff
			raw_buffer      << (@fd.read(pad_length) || '')
			@read_buffer     = raw_buffer.slice!(length, pad_length-1)
		end

		raw_buffer
	end

	def close
		@fd.close if @fd
	end

	attr_reader :object_type
	attr_reader :read_buffer

end

#
# Encapsulates generic windows objects
#
class Object

	# Initializes an object of a given key
	def initialize
	end

	# Updates an object's internal state with new metadata
	def update(metadata)

		# Store the lowest sequence number for this object.  Since the object
		# simulator ensures in-sequence visitation, we don't need to do any
		# subsequent comparisons once it has been set.
		@sequence = metadata.sequence if @sequence.nil?

		case metadata.tag

			when Metadata::Access::Tag
				if metadata.is_def?
					on_definition(metadata)
				else
					on_use(metadata)
				end

				# Get the process object that was accessing this object and 
				# associate it with this metadata
				metadata.process_object = @object_space.find_object(
					metadata.process_object_key)

			when Metadata::SecurityDescriptor::Tag
				on_sd(metadata)

			when Metadata::Name::Tag
				on_name(metadata)

			when Metadata::ImageInfo::Tag
				on_imageinfo(metadata)

		end
	end

	# Called when the object is defined
	def on_definition(metadata)
		@definition = metadata
	end

	# Called when the object is used
	def on_use(metadata)
		@uses = [] if @uses.nil?
		@uses << metadata
	end

	# Called when the object is assigned a security descriptor
	def on_sd(metadata)
		@override_sd = metadata.sd
	end

	# Called when the object is assigned a name
	def on_name(metadata)
		@name = metadata.name
	end

	# Called when the object has image information tied to it
	def on_imageinfo(metadata)
		@images = [] if @images.nil?
		@images << metadata
	end

	# Returns the object's type.  Alias for object_log.object_type.
	def object_type
		object_log.object_type
	end

	# The process object that defined this object
	def process_object
		@definition ? @definition.process_object : nil
	end
	
	def to_s
		name || ("#{object_type}(%.16x)" % key)
	end

	# Returns true if the object has a valid 'definition' metadata
	def is_defined?
		@definition.nil? == false
	end

	# Returns the security descriptor associated with this object
	def security_descriptor
		sd = @override_sd

		if sd.nil? and @definition and @definition.security_descriptor
			sd = @definition.security_descriptor.sd
		end

		return sd
	end

	# Returns the security context of the object's definition
	def security_context
		@definition ? @definition.security_context : nil
	end

	# Returns the client token owner SID of the object's definition
	def client_owner_sid
		security_context ? security_context.client_token_owner_sid : nil
	end

	# Returns the client token primary group SID of the object's definition
	def client_group_sid
		security_context ? security_context.client_token_primary_group_sid : nil
	end

	# Returns the primary token owner SID of the object's definition
	def primary_owner_sid
		security_context ? security_context.primary_token_owner_sid : nil
	end

	# Returns the primary token primary group SID of the object's definition
	def primary_group_sid
		security_context ? security_context.primary_token_primary_group_sid : nil
	end

	# The simulator tied to this object
	def simulator
		@object_space ? @object_space.simulator : nil
	end

	# The context tied to this object's simulator
	def context
		simulator ? simulator.context : nil
	end

	# The object space this object is tied to
	attr_accessor :object_space

	# The object log this object is tied to
	attr_accessor :object_log

	# The object's unique key
	attr_accessor :key

	# Derived from object name metadata
	attr_reader :name

	# Instance of Access metadata describing a definition of an object
	attr_reader :definition

	# Instances of Access metadata describing a use of the object
	attr_reader :uses

	# If this is a process object, this contains a table of images that have
	# been loaded into the address space of the process
	attr_reader :images

	# The lowest sequence of the referenced object
	attr_reader :sequence

end
#
# This class encapsulates the known state of all Windows objects at any given
# point during an object trace simulation
#
class ObjectSpace

	def initialize(simulator = nil, object_klass = Object)
		@objects = {}
		@simulator = simulator
		@object_klass = object_klass
	end

	# Updates the object space with new metadata information
	def update(metadata)

		# Find the object using its object key
		object = find_object(metadata.object_key)

		# If this is a new object or this is a metadata create tag and the
		# current object already has a process defined (this indicates a new
		# object has been created)
		if (object.nil?) or
			(metadata.is_access? and metadata.is_def? and object.is_defined?)

			# If we're about to clobber an object, notify our callbacks so they
			# get a chance to look at it
			if object and @simulator
				@simulator.notify_final(object)
			end

			object = new_object(metadata)
		end

		# This should never happen, but if we ultimately failed to find the
		# object, something bad has happened
		if object.nil?
			puts "failed to find key #{"%.16x" % metadata.object_key} for #{metadata}"
			return nil
		end

		# Update the object with this new metadata
		object.update(metadata)

		# Return the object that we just updated
		return object
	end
	
	# Finds an object given an object key
	def find_object(key)
		@objects[key]
	end

	# Walk the active objects table
	def each(&b)
		@objects.values.each(&b)
	end

	# Sorts the object space by sequence number
	def sorted
		@objects.values.sort { |o1, o2|
			o1.sequence <=> o2.sequence	
		}
	end

	# The number of objects active in the object space
	def count
		@objects.length
	end

	# The system process object
	attr_reader :system_process

	# The simulator this object space is tied to
	attr_reader :simulator

private

	# Creates a new object and updates the object space hash
	def new_object(metadata)
		object = @object_klass.new

		object.key = metadata.object_key
		object.object_log = metadata.log
		object.object_space = self

		# The system process is assumed to be the first process defined
		if object.object_type == 'Process' and @system_process.nil?
			@system_process = object
		end

		@objects[object.key] = object
	end

end

#
# Simulates an object trace as contained within multiple object trace log
# files.  Each object log contains metadata that describes the definition and
# use of securable objects.
#
class ObjectSimulator

	def initialize(object_klass = Object)
		@object_klass = object_klass
	end

	# Runs the simulator against the supplied object log paths
	def run(*object_log_paths, &metadata_proc)

		# Empty out the object logs array
		@object_logs  = []

		begin

			# Open each object log
			object_log_paths.each { |path|
				@object_logs << ObjectLog.new(path)
			}
			
			on_simulation_begin()

			# Initialize an empty read queue.  The read queue is an ordered array
			# that contains the next-entry-to-be-read
			read_queue = initialize_read_queue

			# Until we've completely emptied the read queue, keep processing it
			until read_queue.empty?

				# Remove the next metadata entry from the queue
				metadata = read_queue.shift

				# Process the metadata, such as by calling the proc that was
				# supplied, if any
				on_metadata(metadata, &metadata_proc)
	
				# Read the next metadata entry and insert it into the read queue
				# at the appropriate position
				begin
					insert_read_queue(read_queue, metadata.next)
				rescue EOFError
				end

			end

			on_simulation_end()

		ensure

			# Close all of the object logs we've opened
			@object_logs.each { |log|
				log.close
			}

		end

	end

	# The object logs being read from
	attr_reader :object_logs

private

	def on_simulation_begin
	end

	def on_simulation_end
	end

	#
	# Initializes the read queue by reading the first entry from each object
	# log to figure out the starting sequence identifiers for each log.
	#
	def initialize_read_queue
		read_queue = []

		# Read the first metadata entry from each log file and insert the first
		# entry into the unsorted read queue
		@object_logs.each { |log|
			read_queue << log.read_metadata
		}

		# Sort the read queue so that the metadata entry with the lowest
		# sequence number is first
		read_queue.sort { |x, y|
			x.sequence <=> y.sequence
		}
	end

	# Inserts a metadata entry into the read queue at the appropriate index
	def insert_read_queue(read_queue, metadata)
		insert_at = nil

		# Walk the read queue
		read_queue.each_index { |idx|

			# If the metadata entry has a sequence that is below the sequence at
			# the current index, then we will insert before this index
			if metadata.sequence < read_queue[idx].sequence
				insert_at = idx
				break
			end

		}
	
		# If we found an index to insert at, then do so.  Otherwise, append the
		# metadata entry to the end of the queue
		if insert_at
			read_queue.insert(insert_at, metadata)
		else
			read_queue.push(metadata)
		end
	end

	# Called when a metadata entry is read from an object log
	def on_metadata(metadata, &metadata_proc)
		metadata_proc.call(metadata) if metadata_proc
	end

end

#
# Simulates an object trace by tracking object information over time
#
class ObjectSpaceSimulator < ObjectSimulator

	# Initializes the object space simulator
	def initialize(object_klass = Object)
		super

		@update_callbacks = []
		@final_callbacks = []
	end

	# Adds a callback procedure with a given set of criteria
	def add_update_callback(p, criteria)
		@update_callbacks << [p, criteria]
	end

	# Adds a callback procedure that will be notified when an object reaches
	# its final form
	def add_final_callback(p, criteria)
		@final_callbacks << [p, criteria]
	end

	# Runs the object trace
	def run(*object_log_paths, &metadata_proc)
		# Initialize the object space and object logs to a pristine state
		@object_space = ObjectSpace.new(self, @object_klass)

		# Pass the parameters on to the base class
		super

		# Walk the sequence-sorted active object space
		@object_space.sorted.each { |object|
			notify_final(object)
		}
	end

	# Notifies registered callbacks about an object being updated
	def notify_update(object, metadata)
		@update_callbacks.each { |cb|
			p, criteria = cb

			next if criteria and !meets_criteria?(object, criteria)

			p.call(object, metadata)
		}
	end

	# Notifies all final callbacks about the state of an object
	def notify_final(object)
		@final_callbacks.each { |cb|
			p, criteria = cb

			next if criteria and !meets_criteria?(object, criteria)

			p.call(object)
		}
	end

	attr_reader :object_space

private

	# Checks to see if an object meets criteria
	def meets_criteria?(object, criteria)

		# If the object is nil, we assume that it does not meet the criteria
		return false if object.nil?

		# Check to make sure that this object matches the criteria for the
		# object to be called
		if criteria

			# Is the name defined?
			return false if criteria['RequireName'] and object.name.nil?

			# Is the object properly defined?
			return false if criteria['RequireDefinition'] and object.is_defined? == false

			# Does the object have a security descriptor attached to it?
			return false if criteria['RequireSecurityDescriptor'] and object.security_descriptor.nil?
			
			# Does the object have a security context attached to it?
			return false if criteria['RequireSecurityContext'] and object.security_context.nil?

		end

		return true
	end

	# Called when a metadata entry is read from an object log
	def on_metadata(metadata, &metadata_proc)
		object = @object_space.update(metadata)

		# Notify callbacks of the object that was just changed
		notify_update(object, metadata)

		# After we've updated our state, call the callback via the base class
		# implementation
		super
	end


end

end
