#!/usr/bin/env ruby
#
# Trust boundary analysis library
#
# analyzer.rb: Front-end class that supports interpreting the trust boundary
# information that is collected
#
# mmiller@hick.org
# 4/2008
#

require 'common/sddl'
require 'common/graph'

require 'objtrace/objects'

require 'tba/flow'
require 'tba/simulator'
require 'tba/rights'
require 'tba/smg'
require 'tba/tbg'

module TBA

class Analyzer

	def initialize
		@simulator = Simulator.new
		@smg = SidMembershipGraph.from_file
	end

	# Processes the object traces in the supplied file paths
	def process_traces(*traces)
		# Add a relationship filter that will cause the simulator to ignore
		# certain relationships.  This procedure returns false if the use
		# SID is not more privileged than the definition SID.
		@simulator.relationship_filters << Proc.new { |definition, use|
			more_privileged?(use.sid, definition.sid)
		}

		# If a partial flow is tied to a valid object name then it's fine.
		# Otherwise, it is not.
		@simulator.vertex_filters << Proc.new { |pflow|
			not pflow.object_name.nil?
		}

		# Run the simulator against the supplied traces
		@simulator.run(*traces)
	end

	# Enumerates the set of all definitions from a given SID
	def each_attack_horizon(sid, &b)
		@simulator.each_sid_definition(sid) { |definition|

			definition.related_vertices.each { |use|

				b.call(
					definition, 
					use, 
					@simulator.flow_threats(
						definition.descriptor, 
						use.descriptor))

			}

		}
	end

	# Enumerates the set of all uses to a given SID
	def each_defense_horizon(sid, &b)
		@simulator.each_sid_use(sid) { |use|

			use.related_vertices.each { |definition|

				b.call(
					definition, 
					use, 
					@simulator.flow_threats(
						definition.descriptor, 
						use.descriptor))

			}

		}
	end

	# Builds a trust boundary graph from the current state of the simulator
	def build_tbg(sid = nil)
		@simulator.build_tbg(sid)
	end

	# Finds an image+rva for a given RIP in the context of a given process
	def find_image_rva(pobject_key, rip)
		@simulator.find_image_rva(pobject_key, rip)
	end

	def process_name(pobject_key)
		@simulator.process_name_cache[pobject_key] or "#{"%.16x" % pobject_key}"
	end

	# Is d1 less privileged than d2?
	def less_privileged?(d1, d2)
		@smg.less_privileged?(d1, d2)
	end

	# Is d1 more privileged than d2?
	def more_privileged?(d1, d2)
		@smg.more_privileged?(d1, d2)
	end

	def add_relationship_filter(filter)
		@simulator.relationship_filters << filter
	end

	def add_vertex_filter(filter)
		@simulator.vertex_filters << filter
	end

	def ignore_actualized=(b)
		@simulator.ignore_actualized = b
	end
	
	def ignore_potential=(b)
		@simulator.ignore_potential = b
	end

	def sid_membership_file=(path)
		@smg = SidMembershipGraph.from_file(path)
	end
	
	attr_reader :smg
	attr_reader :tbg

private

end

end
