#!/usr/bin/env ruby

module TBA

class Medium

	class Descriptor
		def initialize(name)
			@name = name
		end

		attr_reader :name
	end

end

#
# A data flow vertex describes a specific instance in which information flows from
# or to a domain of trust.  Partial flows are direction-sensitive and can
# either be an instance of an use (Use) or an definition (Definition).
#
class DataFlowVertex

	EmptyArray = []

	# 
	# A data flow vertex descriptor abstractly defines the vector through which
	# information can flow through a given medium
	#
	class Descriptor

		# Initializes the data flow vertex descriptor by assigning it a name, the
		# medium descriptor being operated upon, and a hash of any additional
		# parameters needed to describe this descriptor.  
		#
		# params can contain
		#
		# MatchDefinition => t/f
		# RightsRequired => string
		# Implicit => t/f
		# ImplicitSid => sid/proc
		# AssociatedProc => proc
		def initialize(name, medium_descriptor, params)
			@name = name
			@medium_descriptor = medium_descriptor
			@params = params
			
			@object_type = @medium_descriptor.name
		end

		# The rights required, if any, to match this descriptor
		def rights_required
			@params['RightsRequired']
		end

		# True if this descriptor should match definitions
		def match_definition?
			@params['MatchDefinition'] || false
		end

		# True if this is a descriptor for an implicit flow.  This means that
		# the flow is not derived from any specific operation but naturally
		# occurs as a result of another flow
		def implicit?
			@params['Implicit'] || false
		end

		# The implicit SID that should be used for flows of this type
		def implicit_sid(from_vertex)
			isid = @params['ImplicitSid']

			if isid and isid.kind_of?(Proc)
				isid.call(from_vertex)
			else
				isid
			end
		end

		# The name of the medium through which data would flow (e.g. Process)
		def medium
			@medium_descriptor.name
		end

		# Creates an implicit data flow vertex
		def new_implicit(from_vertex)
			flow = flow_class.new(
				implicit_sid(from_vertex), 
				from_vertex.object_key, 
				self)
			flow.object_name = from_vertex.object_name
			flow
		end

		# Creates a new does define data flow vertex
		def new_does_def(sid, does_def_md)
			flow = flow_class.new(sid, does_def_md.object_key, self)
			flow.does_def_md = does_def_md
			flow
		end

		# Creates a new does use data flow vertex
		def new_does_use(sid, does_use_md)
			flow = flow_class.new(sid, does_use_md.object_key, self)
			flow.does_use_md = does_use_md
			flow
		end

		# Creates a new can use data flow vertex
		def new_can_use(sid, can_use_md)
			flow = flow_class.new(sid, can_use_md.object_key, self)
			flow.can_use_md = can_use_md
			flow.can = true
			flow
		end

		# Does this data flow vertex descriptor match the defintion?
		def match_does_def?(definition)
			match_definition?
		end

		# Does this data flow vertex descriptor match a does use?
		def match_does_use?(use)

			# If we require that certain rights be granted in order to match, then
			# check to see if those rights were granted
			if rights_required and has_required_rights(use.access_mask)
				return true
			end

			return false

		end

		# Does this data flow vertex descriptor match a can use?
		def match_can_use?(sd, ace)

			# If we require that certain rights be granted in order to match, then
			# check to see if those rights were granted
			if rights_required and has_required_rights(ace.rights)
				return true
			end

			return false

		end

		# Returns in a string in MediumName(name) form
		def to_s
			"#{@name}"
		end

		# True if this flow descriptor represents an use
		def is_use?
			flow_class == Use
		end

		# True if this flow descriptor represents an definition
		def is_def?
			not is_use?
		end

		# The class that should be instantiated when creating an instance of a
		# data flow vertex from new_xxx
		def flow_class
			raise NotImplementedError
		end

		# The object type involved in the flow
		attr_reader :object_type

		# The friendly name of this flow descriptor
		attr_reader :name

	private

		# Implicit rights.  If rights grants write_dac, all rights can be obtained.
		def implicit_rights(rights)
			if rights and (rights & 0x40000) == 0x40000
				rights = 0xffffffff
			end

			rights
		end

		# Checks to see if the rights supplied meet those that are required
		def has_required_rights(rights)
			required = Medium::Rights[medium][rights_required]
			granted  = implicit_rights(rights)

			return (required & granted) == required
		end

	end

	# Initializes an instance of a data flow vertex as relating to a SID, the key
	# of the object being operated upon, and the data flow vertex descriptor
	# describing the manner of the operation
	def initialize(sid, object_key, descriptor)
		@sid = sid
		@object_key = object_key
		@descriptor = descriptor
	end

	# True if this is an use
	def is_use?
		kind_of?(Use)
	end

	# True if this is an definition
	def is_def?
		kind_of?(Definition)
	end

	# True if this is a can-vertex (meaning it is possible based on ACLs, but
	# does not explicitly occur).  A can-vertex also happens if this is the
	# result of a definition.
	def can?
		@can
	end

	# True if this is a does-vertex (meaning the flow does occur as a result of
	# behavior or is implicit)
	def does?
		not can?
	end
	
	# True if this is an implicit data flow vertex
	def implicit?
		descriptor.implicit?
	end

	# True if this is an explicit data flow vertex
	def explicit?
		not implicit?
	end

	# Checks to see if this data flow vertex instance is associated with another
	# data flow vertex instance.  For instance, if the two flows are not operating
	# on the same object key, they are not associated.
	def associated_with?(vertex)
		self.object_key == vertex.object_key
	end

	# The object type of the medium that is involved in this flow
	def object_type
		descriptor.object_type
	end

	# The object's name as a string
	def object_name_to_s
		"#{object_type}(#{@object_name || ("%.16x" % object_key)})"
	end

	# The raw, unformatted object name
	def object_name
		@object_name
	end

	# Assigns the object an explicit name
	def object_name=(name)
		@object_name = name
	end

	# Returns a string in the format domain,vector,medium
	def to_s
		"#{sid},#{object_name_to_s},#{can? ? "can" : "does"}#{implicit? ? " implicitly" : ""} #{descriptor}"
	end

	# Adds a flow as being related to this data flow vertex
	def add_related(flow)
		(@related_vertices ||= []) << flow
	end

	# The collection of flows related to this data flow vertex
	def related_vertices
		@related_vertices || EmptyArray
	end

	def metadata
		return does_def_md if does_def_md
		return does_use_md if does_use_md
		return can_use_md if can_use_md
	end

	# The SID involved in this flow
	attr_reader :sid

	# The descriptor that describes the vector of the flow
	attr_reader :descriptor

	# The object being operated upon
	attr_reader :object_key

	# True or false depending on whether this is a can flow versus a does flow
	attr_accessor :can

	# If this data flow vertex is a does use
	attr_accessor :does_def_md

	# If this data flow vertex is a does use
	attr_accessor :does_use_md

	# If this data flow vertex is a can use
	attr_accessor :can_use_md

end

#
# A partial use
#
class Use < DataFlowVertex

	#
	# A partial use descriptor
	#
	class Descriptor < DataFlowVertex::Descriptor
		
		# See base class
		def initialize(name, medium_descriptor, params)
			super
		end

		# See base class
		def flow_class
			Use
		end

	end

	# See base class
	def initialize(sid, object_key, descriptor)
		super
	end
end

#
# A partial definition
#
class Definition < DataFlowVertex

	#
	# A partial definition descriptor
	#
	class Descriptor < DataFlowVertex::Descriptor

		# See base class
		def initialize(name, medium_descriptor, params)
			super
		end

		# See base class
		def flow_class
			Definition
		end

	end

	# See base class
	def initialize(sid, object_key, descriptor)
		super
	end

end

class DataFlow

	class Descriptor

		def initialize(definition_desc, use_desc, threats, params)
			@definition_desc = definition_desc
			@use_desc  = use_desc
			@threats = threats
			@params       = params || {}
		end

		# The associate procedure, if any
		def associate_proc
			@params["AssociateProc"]
		end

		# The definition descriptor
		attr_reader :definition_desc

		# The use descriptor
		attr_reader :use_desc

		# The threats of the dataflow
		attr_reader :threats

	end

end

#
# Implications of a given flow
#
class Threats
	DenialOfService = "DoS"
	ElevationOfPrivilege = "EoP"
	InformationDisclosure = "Info Disc"

	def initialize(params)
		@params = params
	end

	# A collection of threats if the use side is privileged
	def in_priv
		@params['InPriv']
	end

	def in_priv_to_s
		(in_priv || []).join(",")
	end

	# A collection of threats if the definition side is privileged
	def out_priv
		@params['OutPriv']
	end

	def out_priv_to_s
		(out_priv || []).join(",")
	end
end

#
# A directed graph 
#
class DataFlowDescriptorGraph < DirectedGraph

	def initialize
		super

		# Initialize the has of data flow vertex descriptors to an empty hash 
		@vertex_descriptors = {}

		#
		# ALPC port object
		#
		alpc_md = Medium::Descriptor.new("ALPC Port")

		alpc_fd_writereq = Definition::Descriptor.new(
			"Write request",
			alpc_md,
			"RightsRequired" => "Connect")
		alpc_fd_readreq = Use::Descriptor.new(
			"Read request",
			alpc_md,
			"MatchDefinition" => true)

		alpc_fd_writerep = Definition::Descriptor.new(
			"Write reply",
			alpc_md,
			"MatchDefinition" => true)
		alpc_fd_readrep = Use::Descriptor.new(
			"Read reply",
			alpc_md,
			"RightsRequired" => "Connect")

		# Write a request message which is then read
		add_dataflow_desc(
			alpc_fd_writereq,
			alpc_fd_readreq,
			Threats.new(
				"InPriv"  => [ Threats::ElevationOfPrivilege  ],
				"OutPriv" => [ Threats::InformationDisclosure ]))

		# Write a reply to a request which is then read
		add_dataflow_desc(
			alpc_fd_writerep,
			alpc_fd_readrep,
			Threats.new(
				"InPriv"  => [ Threats::ElevationOfPrivilege  ],
				"OutPriv" => [ Threats::InformationDisclosure ]))

		#
		# File object
		#
		file_md = Medium::Descriptor.new("File")

		file_fd_readdata = Use::Descriptor.new(
			"Read file data", 
			file_md,
			"RightsRequired" => "ReadData")
		file_fd_writedata = Definition::Descriptor.new(
			"Write file data", 
			file_md,
			"RightsRequired" => "WriteData")

		# Too expensive at the moment
		# Write data to a file which is then read
		add_dataflow_desc(
			file_fd_writedata,
			file_fd_readdata,
			Threats.new(
				"InPriv"  => [ Threats::ElevationOfPrivilege  ],
				"OutPriv" => [ Threats::InformationDisclosure ]))

		#
		# Process object
		#
		process_md = Medium::Descriptor.new("Process")

		process_fd_readvm  = Use::Descriptor.new(
			"Read process VM", 
			process_md,
			"RightsRequired" => "VmRead")
		process_fd_writevm = Definition::Descriptor.new(
			"Write process VM", 
			process_md,
			"RightsRequired" => "VmWrite")
		process_fd_execute = Use::Descriptor.new(
			"Execute process",
			process_md,
			"MatchDefinition" => true)

		# Write data to process memory which is then executed
		add_dataflow_desc(
			process_fd_writevm,
			process_fd_execute,
			Threats.new(
				"InPriv" => [ Threats::ElevationOfPrivilege ]))

		# Write data to a file which is then executed
		add_dataflow_desc(
			file_fd_writedata, 
			process_fd_execute,
			Threats.new(
				"InPriv" => [ Threats::ElevationOfPrivilege ]),

			# We must manually associate data flow vertexs in this case.  We
			# associate where a user CAN write to a file that DOES get executed.
			"AssociateProc" => Proc.new { |definition, use|

				# Process file name will appear as an NT file name
				proc_fname = use.object_name

				# Chop off the drive letter from the file name since it will
				# appear as a DOS file name rather than an NT file name
				file_fname = (definition.object_name || "")[2..-1]

				# If the process and file name are valid and the process file name
				# ends with the name of the file that was opened for write, then
				# we will create an association
				file_fname and file_fname.length > 0 and
					proc_fname and 
					not proc_fname.match(/#{Regexp.escape(file_fname)}$/i).nil?

			})

		#
		# Section object
		#
		section_md = Medium::Descriptor.new("Section")

		section_fd_write = Definition::Descriptor.new(
			"Write",
			section_md,
			"RightsRequired" => "MapWrite")
		section_fd_read = Use::Descriptor.new(
			"Read",
			section_md,
			"RightsRequired" => "MapRead")
		section_fd_execute = Use::Descriptor.new(
			"Execute",
			section_md,
			"RightsRequired" => "MapExecute")

		# Write to a section which is then read from
		add_dataflow_desc(
			section_fd_write,
			section_fd_read,
			Threats.new(
				"InPriv"  => [ Threats::ElevationOfPrivilege  ],
				"OutPriv" => [ Threats::InformationDisclosure ]))

		# Write to a section that is then executed from
		add_dataflow_desc(
			section_fd_write,
			section_fd_execute,
			Threats.new(
				"InPriv" => [ Threats::ElevationOfPrivilege ]))

		#
		# Service object
		#
		service_md = Medium::Descriptor.new("Service")

		service_fd_changeconfig = Definition::Descriptor.new(
			"Change configuration",
			service_md,
			"RightsRequired" => "ChangeConfig")
		service_fd_start = Use::Descriptor.new(
			"Start service",
			service_md,
			"Implicit"    => true,
			"ImplicitSid" => Sid.sy_sid)

		# Change service configuration and then start the service as SY
		add_dataflow_desc(
			service_fd_changeconfig,
			service_fd_start,
			Threats.new(
				"InPriv" => [ Threats::ElevationOfPrivilege ]))

		# Thread object
		thread_md = Medium::Descriptor.new("Thread")

		thread_fd_setctx = Definition::Descriptor.new(
			"Set thread context",
			thread_md,
			"RightsRequired" => "SetContext")
		thread_fd_execute = Use::Descriptor.new(
			"Execute thread",
			thread_md,
			"MatchDefinition" => true)

		# Set the context of a thread that is executed
		add_dataflow_desc(
			thread_fd_setctx,
			thread_fd_execute,
			Threats.new(
				"InPriv" => [ Threats::ElevationOfPrivilege ]))

	end

	# Enumerates the data flow vertex descriptors tied to a given object type
	def each_vertex_desc(object_type, &b)
		(@vertex_descriptors[object_type] || []).each(&b)
	end

	# Returns the threats of a given flow
	def threats(definition_desc, use_desc)
		successors_wa(definition_desc).each { |curr_use_desc, dataflow_desc|
			return dataflow_desc.threats if use_desc == curr_use_desc
		}
		nil
	end

	# Determines if an definition is associated with an use according to
	# descriptor relationships
	def associated_with?(definition, use)
		successors_wa(definition.descriptor).each { |curr_use_desc, dataflow_desc|
			if use.descriptor == curr_use_desc

				associate_proc = dataflow_desc.associate_proc

				# If no associate procedure is defined or the defined one returns
				# true, then the definition and use are said to be associated
				if associate_proc.nil? or
					associate_proc.call(definition, use)

					return true 

				end
			end
		}

		false
	end

private

	# Adds a dataflow descriptor linking an definition with an use
	def add_dataflow_desc(definition_desc, use_desc, threats, params = {})
		dataflow_desc = DataFlow::Descriptor.new(
			definition_desc,
			use_desc,
			threats,
			params)

		add_vd(dataflow_desc.definition_desc)
		add_vd(dataflow_desc.use_desc)
		add_edge(
			dataflow_desc.definition_desc,
			dataflow_desc.use_desc,
			dataflow_desc)
	end

	# Adds a data flow vertex descriptor
	def add_vd(fd)
		(@vertex_descriptors[fd.object_type] ||= []) << fd
	end

end

end
