#!/usr/bin/env ruby
#
# Trust boundary analysis library
#
# simulator.rb: Simulates a collection of object trace logs to extract data
# flow information needed to interpret trust boundaries
#
# mmiller@hick.org
# 4/2008
#

require 'common/graph'

require 'objtrace/objects'

require 'tba/flow'
require 'tba/rights'
require 'tba/tbg'

module TBA

#
# Custom simulator that collects data flow information from object trace logs
#
class Simulator < ObjectTrace::ObjectSimulator

	EmptyArray = []

	# Constructor
	def initialize
		super

		@flow_desc_graph = DataFlowDescriptorGraph.new

		# Populated during simulation
		@vertices_by_object = {}

		# Populated post-simulation
		@vertices_by_desc = {}
		@vertices_by_sid = {}

		# Empty array of relationship filters
		@relationship_filters = []
		# Empty array of data flow vertex filters
		@vertex_filters = []

		# We don't ignore anything by default
		@ignore_actualized  = false
		@ignore_potential = false

		@process_cache = {}
		@process_name_cache = {}
		@imageinfo_cache = {}
	end

	# Returns the threats of a given flow as composed of an out-flow and
	# an in-flow
	def flow_threats(definition_desc, use_desc)
		@flow_desc_graph.threats(definition_desc, use_desc)
	end

	# Enumerates each in-flow to a given SID
	def each_sid_use(sid, &b)
		(@vertices_by_sid[sid] || EmptyArray).each { |vertex|
			next if vertex.is_def?
			b.call(vertex)
		}
	end

	# Enumerates each out-flow from a given SID
	def each_sid_definition(sid, &b)
		(@vertices_by_sid[sid] || EmptyArray).each { |vertex|
			next if vertex.is_use?
			b.call(vertex)
		}
	end

	# Build a trust boundary graph from the current state
	def build_tbg(sid = nil)
		TrustBoundaryGraph.new(self, sid)	
	end

	# Finds the image+rva for a given RIP in the context of a given process
	def find_image_rva(pobject_key, rip)
		ripstr = find_image_rva_in_process(
			pobject_key, 
			rip)

		if ripstr.nil? and @system_process
			ripstr = find_image_rva_in_process(
				@system_process,
				rip)
		end

		ripstr = "%.16x" % rip if ripstr.nil?

		ripstr
	end

	# Ignore active data flows such as where a data flow's in-flow is DOES.
	# Default is false.
	attr_accessor :ignore_actualized

	# Ignore passive data flows such as where a data flow's in-flow is CAN.
	# Default is false.
	attr_accessor :ignore_potential

	# The system process
	attr_reader :system_process

	# Graph of flow descriptors and their relationships to one another
	attr_reader :flow_desc_graph

	# A hash of data flow vertexs indexed by a flow descriptor
	attr_reader :vertices_by_desc

	# A hash of data flow vertexs indexed by SID
	attr_reader :vertices_by_sid

	# Cache associating a process object key with its name
	attr_reader :process_name_cache

	# Cache associating a process object key with its definition metadata
	attr_reader :process_cache

	# Cache associating a process object key with an array of imageinfo
	# metadata instances
	attr_reader :imageinfo_cache

	# An array of procedures that can be used to filter out valid/invalid
	# data flow vertex relationships.  Each proc is passed definition,use.
	attr_reader :relationship_filters

	# An array of procedures that can be used to filter invalid data flow vertexs,
	# such as those that do not have a valid name.  Each proc is passed a
	# vertex.
	attr_reader :vertex_filters

private

	# When the simulation completes, we go through each data flow vertex and add
	# any implicit data flow vertexs that are needed
	def on_simulation_end

		# Walk each object key that has been associated with data flow vertexs
		@vertices_by_object.delete_if { |object_key, flows|

			# Walk the array of data flow vertexs associated with this object
			0.upto(flows.length-1) { |oidx|
			
				# Alias the data flow vertex we'll be operating upon
				vertex1 = flows[oidx]

				# If the data flow vertex is not acceptable, skip it
				next if not acceptable_vertex?(vertex1)

				# If this is an out flow, attempt to determine if this partial
				# flow has any implicit relationships
				definition = nil
				use  = nil

				if vertex1.is_def?
					add_implicit_relationships(vertex1)

					definition = vertex1 
				end
				
				# Regardless of what happens, we always track the data flow vertexs
				# into the next stage as we may need to associate them later.
				add_vertex_stg2(vertex1)

				# Now, attempt to figure out what flows this flow might be
				# associated with
				(oidx+1).upto(flows.length-1) { |iidx|
				
					# Alias the second data flow vertex
					vertex2 = flows[iidx]

					# Determine which data flow vertex represents in/out
					if definition and vertex2.is_use?
						use = vertex2
					elsif vertex2.is_def?
						use  = vertex1
						definition = vertex2
					end

					# If both data flow vertexs have the same direction, skip them
					next if use.nil? or definition.nil?

					# If this relationship is not acceptable based on a dynamically
					# specified relationship filter, then ignore it
					next if not acceptable_relationship?(
						definition,
						use)

					# Are the data flow vertexs associated with one another according
					# to the information stored in the flow descriptor graph?  If
					# not, we skip.
					next if not @flow_desc_graph.associated_with?(
						definition,
						use)

					# If the descriptors are associated, then the data flow vertex
					# instances are associated
					definition.add_related(use)
					use.add_related(definition)
				}

			}

			# Always remove entries as we go
			true

		}

		# Now that we've built relationships between data flow vertexs that are
		# explicitly associated (based on the fact that they access the same
		# object), we must now go back through and add relationships between
		# data flow vertexs that may not be explicitly associated.  This operation
		# can be very slow.
		@flow_desc_graph.each_edge_annotations { |definition_desc, use_desc, dataflow_desc|
		
			# If this complete flow doesn't require dynamic association, then
			# skip it
			associate_proc = dataflow_desc.associate_proc
			next if associate_proc.nil?

			@vertices_by_desc[definition_desc].each { |definition|

				@vertices_by_desc[use_desc].each { |use|

					if associate_proc.call(definition, use)

						# If this relationship is not acceptable based on a
						# dynamically specified relationship filter, then ignore it
						next if not acceptable_relationship?(
							definition,
							use)

						definition.add_related(use)
						use.add_related(definition)

					end

				} if @vertices_by_desc[use_desc]

			} if @vertices_by_desc[definition_desc]

		}
	end

	# Is the provided data flow vertex acceptable?
	def acceptable_vertex?(vertex)

		# If this form of data flow is explicitly ignored, then it is not
		# acceptable
		return false if ignore_actualized and vertex.is_use? and vertex.does?
		return false if ignore_potential and vertex.is_use? and vertex.can?

		@vertex_filters.each { |filter_proc|
			return false if not filter_proc.call(vertex)
		}
		return true
	end

	# Is the provided definition/use pair acceptable in combination?
	def acceptable_relationship?(definition, use)
		@relationship_filters.each { |filter_proc|
			return false if not filter_proc.call(definition, use)
		}
		return true
	end

	# Add any implicit relationships for the provided definition
	def add_implicit_relationships(definition)
		@flow_desc_graph.successors_wa(definition.descriptor).each { |use_desc, threats|

			next if not use_desc.implicit?

			use = use_desc.new_implicit(
				definition)

			# Is this in-flow acceptable?
			next if not acceptable_vertex?(
				use)

			# If this relationship is not acceptable based on a dynamically
			# specified relationship filter, then ignore it
			next if not acceptable_relationship?(
				definition,
				use)

			definition.add_related(use)
			use.add_related(definition)

			add_vertex_stg2(use)

		}
	end

	# Called by the base class when a metadata element is being simulated
	def on_metadata(metadata, &metadata_proc)
		case metadata.tag
			when ObjectTrace::Metadata::Access::Tag
				if metadata.is_use?
					process_use(metadata) 
				else
					process_definition(metadata)
				end

			when ObjectTrace::Metadata::SecurityDescriptor::Tag
				process_sd(metadata, metadata.sd)

			when ObjectTrace::Metadata::Name::Tag
				process_name(metadata)

			when ObjectTrace::Metadata::ImageInfo::Tag
				process_imageinfo(metadata)

		end

		super
	end

	# Processes an access definition metadata instance
	def process_definition(definition)
		object_type = definition.object_type

		# If this is defining a process, track that fact
		if object_type == 'Process'
			@system_process = definition.object_key if @system_process.nil?

			@process_cache[definition.object_key] = definition
		end
	
		# Walk each data flow vertex descriptor for this object type.  If a
		# descriptor matches this definition, add a data flow vertex to track it.
		each_vertex_desc(definition.object_type) { |desc|
	
			#
			#next if not desc.match_does_def?(definition)
			#
			# Implicitly, we allow definitions to match all in-flow descriptors
			# as the definer could be subject to any of them.
			next if desc.is_def?

			sid = nil

			# TODO: In some cases, a security descriptor cannot be obtained
			# during definition.  Right now we assume that when this occurs that
			# the sid of the object is equal to that of the client token (or
			# primary token).  This isn't a safe assumption.
			if definition.security_descriptor and 
				definition.security_descriptor.sd
				sid = definition.security_descriptor.sd.owner_sid
			else
				if definition.security_context.client_token_owner_sid
					sid = definition.security_context.client_token_owner_sid
				else
					sid = definition.security_context.primary_token_owner_sid
				end
			end

			add_vertex_stg1(
				desc.new_does_def(
					sid,
					definition))

		}

		# If the object was assigned a security descriptor, we should evaluate
		# it
		if definition.security_descriptor and
			definition.security_descriptor.sd

			process_sd(
				definition,
				definition.security_descriptor.sd)

		end
	end

	# When an object is used, determine the non-conditional data flow vertexs that
	# acctually occur
	def process_use(use)

		#
		# At this point we have a does-use.  We need to translate this into an
		# instance of a data flow vertex whether it be input or output.
		#
		each_vertex_desc(use.object_type) { |desc|

			# If this descriptor does not accept this use, then we skip it
			next if not desc.match_does_use?(use)
	
			# Create a new does flows from this use
			sc = use.security_context

			add_vertex_stg1(
				desc.new_does_use(
					sc.primary_token_owner_sid,
					use)) if sc.primary_token_owner_sid
			add_vertex_stg1(
				desc.new_does_use(
					sc.primary_token_primary_group_sid,
					use)) if sc.primary_token_primary_group_sid
			add_vertex_stg1(
				desc.new_does_use(
					sc.client_token_owner_sid,
					use)) if sc.client_token_owner_sid
			add_vertex_stg1(
				desc.new_does_use(
					sc.client_token_primary_group_sid,
					use)) if sc.client_token_primary_group_sid
		}
		
	end
	
	# When a security descriptor assigned, we determine the data flow vertexs
	# possible
	def process_sd(metadata, sd) 

		# If the object has a valid DACL...
		if sd.dacl
			sd.dacl.aces.each { |ace|

				# Skip deny aces
				next if ace.is_deny?

				# At this point we have a can-use.  We need to translate this into an
				# instance of a data flow vertex whether it be input or output.
				each_vertex_desc(metadata.object_type) { |desc|

					# If this descriptor does not accept this use, then we skip it
					next if not desc.match_can_use?(sd, ace)
			
					# Create a new can flow from this use
					add_vertex_stg1(
						desc.new_can_use(
							ace.sid,
							metadata))
				}
				
			}

		# If the object has a null DACL then everyone can implicitly match any
		# data flow vertex
		else
				
			each_vertex_desc(metadata.object_type) { |desc|

				add_vertex_stg1(
					desc.new_can_use(
						Sid.wd_sid,
						nil))
			
			}

		end

	end

	# Assign name information to all data flow vertexs currently related to this
	# object key
	def process_name(name_md)
		# Cache the process name globally
		if @process_cache[name_md.object_key]
			@process_name_cache[name_md.object_key] = name_md.name
		end

		@vertices_by_object[name_md.object_key].each { |vertex|
			vertex.object_name = name_md.name
		} if @vertices_by_object[name_md.object_key]
	end
	def find_image_rva_in_process(pobject_key, rip)
		if pobject_key and @imageinfo_cache[pobject_key]
			@imageinfo_cache[pobject_key].each { |image|
				if rip >= image.base_address and rip < image.base_address + image.size
					return "#{image.name}+#{"%x" % (rip - image.base_address)}"
				end
			}
		end

		nil
	end

	# Caches image information as being associated with a given process object
	def process_imageinfo(info)
		(@imageinfo_cache[info.object_key] ||= []) << info
	end

	# Enumerates each data flow vertex descriptor
	def each_vertex_desc(object_type, &b)
		@flow_desc_graph.each_vertex_desc(object_type, &b)	
	end

	# Populates the data flow vertex stage 1 hashes (during simulation)
	def add_vertex_stg1(flow)
		(@vertices_by_object[flow.object_key] ||= []) << flow

		# Snag the name of this object if one has been assigned.  If one has
		# not, this flow will be updated once name information is assigned to
		# the object.
		first = @vertices_by_object[flow.object_key].first
		flow.object_name = first.object_name if first != flow
	end

	# Populates the data flow vertex stage 2 hashes (after simulation)
	def add_vertex_stg2(flow)
		(@vertices_by_desc[flow.descriptor] ||= []) << flow
		(@vertices_by_sid[flow.sid] ||= []) << flow
	end

end

end
