#!/usr/bin/env ruby
#
# Trust boundary analysis library
#
# smg.rb: SID membership graph -- describes domains of trust by group
# association
#
# mmiller@hick.org
# 4/2008
#

require 'common/graph'
require 'common/sddl'

module TBA

class SidMembershipGraph < DirectedGraph

	#
	# Populates the graph from a file that should be in the format:
	#
	# s1 -> s2 
	#
	# where s1 is a member of s2
	#
	def self.from_file(path = nil)
		graph = SidMembershipGraph.new
		
		File.open(path, "r") { |fd|
			until fd.eof?
				s1_str, op, s2_str = fd.readline.split(/\s+/)

				next if s1_str.nil? or s2_str.nil?

				s1_sid = Sid.lookup_string(s1_str)
				s2_sid = Sid.lookup_string(s2_str)

				if s1_sid.nil?
					puts "couldn't build sid for '#{s1_str}'"
				elsif s2_sid.nil?
					puts "couldn't build sid for '#{s2_str}'"
				end

				graph.add_edge(
					s1_sid,
					s2_sid)
			end
		} if path

		graph.add_implicit_edges
		graph.calculate_memberships
		graph
	end

	def add_implicit_edges
		# Grab a snapshot of the vertices we have so far
		vertices = []
		each_vertex { |v|
			vertices << v
		}
		
		# All SIDs are implicitly members of Everyone
		everyone = Sid.wd_sid

		vertices.each { |v|
			add_edge(
				v,
				everyone)
		}

		# TODO: Users that can logon are members of Interactive and Logon Session
	end

	def calculate_memberships
		@tcb      = Set.new
		@std_user = Set.new

		# Calculate all of the SIDs that belong to the TCB
		wl = [ 
			Sid.ba_sid, 
			Sid.sy_sid,
			Sid.ao_sid,
			Sid.so_sid,
			Sid.po_sid,
			Sid.bo_sid,

			# For now, we also consider LocalService and NetworkService to be
			# part of the TCB
			Sid.ls_sid,
			Sid.ns_sid
		]

		until wl.empty?
			tcb_sid = wl.shift
			@tcb << tcb_sid
			predecesors(tcb_sid).each { |psid|
				wl << psid
			}
		end
	end

	# TODO: both of these need work 
	
	def less_privileged?(sid1, sid2)
		return false if sid1 == sid2

		if @tcb.member?(sid1) and @tcb.member?(sid2)
			return false
		end

		return true
	end

	def more_privileged?(sid1, sid2)
		return false if sid1 == sid2

		@tcb.member?(sid1) and (@tcb.member?(sid2) == false)
	end

	attr_reader :tcb
	attr_reader :std_user

end

end
