#!/usr/bin/env ruby
#
# Trust boundary analysis library
#
# tbg.rb: Trust boundary graph composed of nodes being SIDs, out-vertices, and
# in-vertices.  
#
# mmiller@hick.org
# 4/2008
#

require 'tba/flow'
require 'tba/simulator'

module TBA

#
# A trust boundary graph illustrates connectedness between multiple domains of
# trust as facilitated in and out vertices
#
# D1 -> definition --- [implications] ---> use -> D2
#
class TrustBoundaryGraph < DirectedGraph

	def initialize(simulator, focus_sid)
		super(Set)

		@simulator = simulator

		# Enumerate data flow vertices based on SID
		@simulator.vertices_by_sid.each_pair { |sid, vertices|

			# If we were supplied a focus SID and the current SID doesn't match,
			# skip it
			next if focus_sid and sid != focus_sid

			# Enumerate each of the data flow vertices associated with this SID
			vertices.each { |vertex|

				if vertex.is_use?
					related = vertex.related_vertices
					next if related.length == 0

					related.each { |rvertex|
						add_edge(rvertex.sid, sid, [rvertex, vertex])
					}
				else
					related = vertex.related_vertices
					next if related.length == 0

					related.each { |rvertex|
						add_edge(sid, rvertex.sid, [vertex, rvertex])
					}
				end

			}
		}
	end

	def summary_graph
		TrustBoundarySummaryGraph.new(self)
	end

	attr_reader :simulator

private

end

class TrustBoundarySummaryGraph < DirectedGraph
	def initialize(tbg)
		super(Set)

		dataflow_counts = {}

		tbg.each_edge_annotations { |a0, a1, annotation|
			dataflow_counts[a0]      = {} if dataflow_counts[a0].nil?
			dataflow_counts[a0][a1]  = 0  if dataflow_counts[a0][a1].nil?
			dataflow_counts[a0][a1] += 1 
		}
		
		dataflow_counts.each_pair { |a0, hash|
			hash.each_pair { |a1, count|
				add_edge(a0, a1, count)
			}
		}
	end

	def to_dot_elabel(u, v, a)
		a.to_s
	end

end

end
