﻿//
// A simple data flow graph implementation that relates data
// definition contexts with data use contexts
//
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using TBA.Database;

namespace TBA.Graphing
{
    /// <summary>
    /// A simple data flow graph that relates data definitions with data uses
    /// </summary>
    public class DataFlowGraph
        : UR.Graphing.DirectedGraph
    {
        /// <summary>
        /// Initializes the data flow graph using records from the supplied data context
        /// </summary>
        /// <param name="db">The data context</param>
        public DataFlowGraph(
            TBADataContext db,
            FilterDataFlowDelegate filter)
        {
            this.Db = db;

            // Extract the TBDFG from the database
            var q =
                from df in Db.DataFlows
                where 
                    df.DataFlowUse.ActorId != df.DataFlowDefinition.ActorId &&
                    df.DataFlowUse.SnapshotId == db.SnapshotId
                select df;

            // Create edges between the definition and use for each data flow
            foreach (DataFlow df in q)
            {
                if (filter != null &&
                    filter(df))
                    continue;

                AddEdge(
                    df.DataFlowDefinition,
                    df.DataFlowUse);
            }
        }

        public UR.Graphing.Graph Summarize()
        {
            Dictionary<string, int> relationCounts =
                new Dictionary<string, int>();
            UR.Graphing.DirectedGraph graph =
                new UR.Graphing.DirectedGraph();

            foreach (UR.Graphing.Edge e in Edges)
            {
                IDataFlowHalf d = e.Source as IDataFlowHalf;
                IDataFlowHalf u = e.Target as IDataFlowHalf;

                string relation =
                    String.Format(
                        "{0}.{1}",
                        d.Actor,
                        u.Actor);

                if (relationCounts.ContainsKey(relation))
                    relationCounts[relation]++;
                else
                    relationCounts[relation] = 1;
            }

            foreach (KeyValuePair<string, int> pair in relationCounts)
            {
                string[] actors = pair.Key.Split(new char [] { '.' });

                graph.AddEdge(
                    graph.CreateEdge(
                        actors[0],
                        actors[1],
                        pair.Value));
            }

            return graph;
        }

        /// <summary>
        /// Returns the defense horizon for the supplied actor
        /// </summary>
        /// <param name="actor">The name of the actor to get the defense horizon for</param>
        /// <returns>An enumerable of edges</returns>
        public IEnumerable<UR.Graphing.Edge> GetDefenseHorizon(string actor)
        {
            foreach (UR.Graphing.Edge e in Edges)
            {
                IDataFlowHalf u = e.Target as IDataFlowHalf;

                // Does the actor match?
                if (u.Actor.Name != actor)
                    continue;

                yield return e;
            }
        }

        /// <summary>
        /// Returns the attack horizon for the supplied actor
        /// </summary>
        /// <param name="actor">The name of the actor to get the attack horizon for</param>
        /// <returns>An enumerable of edges</returns>
        public IEnumerable<UR.Graphing.Edge> GetAttackHorizon(string actor)
        {
            foreach (UR.Graphing.Edge e in Edges)
            {
                IDataFlowHalf d = e.Source as IDataFlowHalf;

                // Does the actor match?
                if (d.Actor.Name != actor)
                    continue;

                yield return e;
            }
        }

        /// <summary>
        /// Database context
        /// </summary>
        private TBADataContext Db { get; set; }
    }
}
