﻿//
// Object trace log parsing classes
//
using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime;
using System.Runtime.InteropServices;
using System.Security.Principal;
using System.Security.AccessControl;

using TBA.Database;

namespace TBA.Analysis
{
    /// <summary>
    /// Encapsulates a securable object as related to an object key
    /// </summary>
    public class SecurableObject
    {
        /// <summary>
        /// Initializes the securable object instance
        /// </summary>
        /// <param name="key">The unique key associated with this object</param>
        public SecurableObject(UInt64 key)
        {
            this.Key = key;
        }

        /// <summary>
        /// The object's unique key
        /// </summary>
        public UInt64 Key { get; private set; }
    }

    /// <summary>
    /// A process securable object
    /// </summary>
    public class ProcessObject
        : SecurableObject
    {
        /// <summary>
        /// Initializes the process object
        /// </summary>
        /// <param name="key">The unique key associated with this object</param>
        public ProcessObject(UInt64 key)
            : base(key)
        {
            this.ImageRecords = new List<ImageInfoRecord>();
        }

        /// <summary>
        /// A list of mapped image records associated with this process 
        /// </summary>
        public List<ImageInfoRecord> ImageRecords { get; private set; }

        /// <summary>
        /// The name of the process object
        /// </summary>
        public string Name { get; set; }
    }

    #region Object log records

    /// <summary>
    /// A record read in from an object trace log
    /// </summary>
    public class ObjectLogRecord
    {
        /// <summary>
        /// Initializes the object trace log record
        /// </summary>
        /// <param name="log">The log the record was read in from</param>
        /// <param name="header">The record's header</param>
        /// <param name="body">The body of the record</param>
        public ObjectLogRecord(
            ObjectTraceLog log,
            RecordHeader header,
            object body)
        {
            this.TraceLog = log;
            this.Header = header;
            this.RecordBody = body;
        }

        /// <summary>
        /// Deserializes a record from an object trace log
        /// </summary>
        /// <param name="log">The log to deserialize from</param>
        /// <returns>An object log record instance</returns>
        public static ObjectLogRecord FromObjectLog(ObjectTraceLog log)
        {
            // Read the record header from the stream
            RecordHeader header = MarshalHeader(log);

            // Marshal the body depending on what type of record it is
            switch (header.Type)
            {
                case RecordType.Access:
                    return MarshalAccessRecord(log, header);

                case RecordType.Name:
                    return MarshalNameRecord(log, header);

                case RecordType.ImageInfo:
                    return MarshalImageInfoRecord(log, header);

                case RecordType.CallStack:
                    return MarshalCallStackRecord(log, header);

                case RecordType.SecurityDescriptor:
                    return MarshalSecurityDescriptorRecord(log, header);

                default:
                    return null;
            }
        }

        /// <summary>
        /// Marshals the header of a log record from the provided trace log
        /// </summary>
        /// <param name="log">The log to deserialize from</param>
        /// <returns>The header associated with the record</returns>
        private static RecordHeader MarshalHeader(
            ObjectTraceLog log)
        {
            log.Stream.Read(
                log.HeaderBuffer,
                0,
                log.HeaderBuffer.Length);

            return (RecordHeader)Marshal.PtrToStructure(
                    log.HeaderPtr,
                    typeof(RecordHeader));
        }

        private static AccessRecord MarshalAccessRecord(
            ObjectTraceLog log,
            RecordHeader header)
        {
            SecurityContextRecord securityContext = null;
            SecurityDescriptorRecord securityDescriptor = null;
            CallStackRecord callStack = null;
            AccessRecord.Body body =
                MarshalStructureFromObjectLog<AccessRecord.Body>(
                    Marshal.SizeOf(typeof(AccessRecord.Body)),
                    log);

            if (body.SecurityContextLength != 0)
            {
                securityContext =
                    MarshalSecurityContextRecord(
                        log);
            }

            if (body.SecurityDescriptorLength != 0)
            {
                securityDescriptor =
                    MarshalSecurityDescriptorRecord(
                        log);
            }

            if (body.CallStackLength != 0)
            {
                callStack =
                    MarshalCallStackRecord(
                        log);
            }

            return new AccessRecord(
                log,
                header,
                body,
                securityContext,
                securityDescriptor,
                callStack);
        }

        private static NameRecord MarshalNameRecord(
            ObjectTraceLog log,
            RecordHeader header)
        {
            NameRecord.Body body = 
                MarshalStructureFromObjectLog<NameRecord.Body>(
                    Marshal.SizeOf(typeof(NameRecord.Body)),
                    log);

            return new NameRecord(
                log,
                header,
                body,
                MarshalObjectFromObjectLog<string>(
                    body.NameLength,
                    log));
        }

        private static ImageInfoRecord MarshalImageInfoRecord(
            ObjectTraceLog log,
            RecordHeader header)
        {
            ImageInfoRecord.Body body =
                MarshalStructureFromObjectLog<ImageInfoRecord.Body>(
                    Marshal.SizeOf(typeof(ImageInfoRecord.Body)),
                    log);

            return new ImageInfoRecord(
                log,
                header,
                body,
                MarshalObjectFromObjectLog<string>(
                    body.ImageNameLength,
                    log));
        }

        private static CallStackRecord MarshalCallStackRecord(
            ObjectTraceLog log)
        {
            return MarshalCallStackRecord(
                log,
                MarshalHeader(
                    log));
        }

        private static CallStackRecord MarshalCallStackRecord(
            ObjectTraceLog log,
            RecordHeader header)
        {
            CallStackRecord.Body body =
                MarshalStructureFromObjectLog<CallStackRecord.Body>(
                    Marshal.SizeOf(typeof(CallStackRecord.Body)),
                    log);
            CallStackFrameRecord[] frames = new CallStackFrameRecord[body.NumberOfFrames];

            for (int index = 0; index < body.NumberOfFrames; index++)
                frames[index] = MarshalStructureFromObjectLog<CallStackFrameRecord>(
                    Marshal.SizeOf(typeof(CallStackFrameRecord)),
                    log);

            return new CallStackRecord(
                log,
                header,
                body,
                frames);
        }

        private static SecurityContextRecord MarshalSecurityContextRecord(
            ObjectTraceLog log)
        {
            return MarshalSecurityContextRecord(
                log,
                MarshalHeader(
                    log));
        }

        private static SecurityContextRecord MarshalSecurityContextRecord(
            ObjectTraceLog log,
            RecordHeader header)
        {
            SecurityIdentifier clientTokenOwnerSid = null;
            SecurityIdentifier clientTokenPrimaryGroupSid = null;
            SecurityIdentifier primaryTokenOwnerSid = null;
            SecurityIdentifier primaryTokenPrimaryGroupSid = null;
            SecurityContextRecord.Body body =
                MarshalStructureFromObjectLog<SecurityContextRecord.Body>(
                    Marshal.SizeOf(typeof(SecurityContextRecord.Body)),
                    log);

            if (body.ClientTokenOwnerSidLength != 0)
            {
                clientTokenOwnerSid =
                    MarshalObjectFromObjectLog<SecurityIdentifier>(
                        (int)body.ClientTokenOwnerSidLength,
                        log);
            }

            if (body.ClientTokenPrimaryGroupSidLength != 0)
            {
                clientTokenPrimaryGroupSid =
                    MarshalObjectFromObjectLog<SecurityIdentifier>(
                        (int)body.ClientTokenPrimaryGroupSidLength,
                        log);
            }

            if (body.PrimaryTokenOwnerSidLength != 0)
            {
                primaryTokenOwnerSid =
                    MarshalObjectFromObjectLog<SecurityIdentifier>(
                        (int)body.PrimaryTokenOwnerSidLength,
                        log);
            }

            if (body.PrimaryTokenPrimaryGroupSidLength != 0)
            {
                primaryTokenPrimaryGroupSid =
                    MarshalObjectFromObjectLog<SecurityIdentifier>(
                        (int)body.PrimaryTokenPrimaryGroupSidLength,
                        log);
            }

            return new SecurityContextRecord(
                log,
                header,
                body,
                clientTokenOwnerSid,
                clientTokenPrimaryGroupSid,
                primaryTokenOwnerSid,
                primaryTokenPrimaryGroupSid);
        }

        private static SecurityDescriptorRecord MarshalSecurityDescriptorRecord(
            ObjectTraceLog log)
        {
            return MarshalSecurityDescriptorRecord(
                log,
                MarshalHeader(
                    log));
        }

        private static SecurityDescriptorRecord MarshalSecurityDescriptorRecord(
            ObjectTraceLog log,
            RecordHeader header)
        {
            SecurityDescriptorRecord.Body body =
                MarshalStructureFromObjectLog<SecurityDescriptorRecord.Body>(
                    Marshal.SizeOf(typeof(SecurityDescriptorRecord.Body)),
                    log);

            if (body.SdLength == 0)
                return null;

            return new SecurityDescriptorRecord(
                log,
                header,
                body,
                MarshalObjectFromObjectLog<CommonSecurityDescriptor>(
                    (int)body.SdLength,
                    log));
        }

        private static T MarshalStructureFromObjectLog<T>(
            int length,
            ObjectTraceLog log)
        {
            return MarshalFromObjectLog<T>(
                length, 
                log, 
                MarshalStructure);
        }

        private static T MarshalObjectFromObjectLog<T>(
            int length,
            ObjectTraceLog log)
        {
            return MarshalFromObjectLog<T>(
                length,
                log,
                MarshalObject);
        }

        private delegate object MarshalDelegate(
            System.Type type,
            ObjectTraceLog log,
            int length,
            byte[] buffer,
            IntPtr ptr);

        private static object MarshalStructure(
            System.Type structType,
            ObjectTraceLog log,
            int length,
            byte[] buffer,
            IntPtr ptr)
        {
            return Marshal.PtrToStructure(
                ptr,
                structType);
        }

        private static object MarshalObject(
            System.Type objectType,
            ObjectTraceLog log,
            int length,
            byte[] buffer,
            IntPtr ptr)
        {
            if (objectType == typeof(String))
                return Marshal.PtrToStringUni(
                    ptr,
                    length / 2);
            else if (objectType == typeof(SecurityIdentifier))
                return new SecurityIdentifier(
                    ptr);
            else if (objectType == typeof(CommonSecurityDescriptor))
                return new CommonSecurityDescriptor(
                    false,
                    log.ObjectType == "Directory" ? true : false,
                    buffer,
                    0);
            else
                return null;
        }

        private static T MarshalFromObjectLog<T>(
            int length, 
            ObjectTraceLog log,
            MarshalDelegate marshaller)
        {
            GCHandle temporaryHandle = default(GCHandle);
            byte[] buffer;
            IntPtr ptr;

            if (length > log.BodySize)
            {
                buffer = new byte[length];
                temporaryHandle = GCHandle.Alloc(buffer, GCHandleType.Pinned);
                ptr = temporaryHandle.AddrOfPinnedObject();
            }
            else
            {
                buffer = log.BodyBuffer;
                ptr = log.BodyPtr;
            }

            // Populate the buffer from the stream
            if ((length > 0) && 
                (log.Stream.Read(buffer, 0, length) != length))
                throw new EndOfStreamException("Partial or incomplete read");

            try
            {
                return (T)marshaller.Invoke(
                    typeof(T),
                    log,
                    length,
                    buffer,
                    ptr);
            }
            finally
            {
                if (temporaryHandle.IsAllocated)
                    temporaryHandle.Free();
            }
        }

        public RecordHeader Header { get; private set; }

        public RecordType Type
        {
            get { return Header.Type; }
        }

        public UInt64 Sequence
        {
            get { return Header.Sequence; }
        }

        public UInt64 ObjectKey
        {
            get { return Header.ObjectKey; }
        }

        /// <summary>
        /// The record's opaque body
        /// </summary>
        public object RecordBody { get; private set; }

        public ObjectTraceLog TraceLog { get; private set; }
    }

    public enum RecordType
    {
        Create = 1, // Unused
        SecurityContext = 2,
        SecurityDescriptor = 3,
        Name = 4,
        CallStack = 5,
        ImageInfo = 6,
        Access = 7
    }

    [StructLayout(LayoutKind.Sequential, Pack=1)]
    public struct RecordHeader
    {
        public UInt64 Tag;
        public UInt32 Length;
        public UInt64 ObjectKey;

        public UInt64 Sequence
        {
            get
            {
                return Tag >> 4;
            }
        }

        public RecordType Type
        {
            get
            {
                return (RecordType)(Tag & 0xf);
            }
        }
    }

    public class SecurityContextRecord
        : ObjectLogRecord
    {
        public SecurityContextRecord(
            ObjectTraceLog log,
            RecordHeader header,
            SecurityContextRecord.Body body,
            SecurityIdentifier clientTokenOwnerSid,
            SecurityIdentifier clientTokenPrimaryGroupSid,
            SecurityIdentifier primaryTokenOwnerSid,
            SecurityIdentifier primaryTokenPrimaryGroupSid)
            : base(log, header, body)
        {
            this.ClientTokenOwnerSid = clientTokenOwnerSid;
            this.ClientTokenPrimaryGroupSid = clientTokenPrimaryGroupSid;
            this.PrimaryTokenOwnerSid = primaryTokenOwnerSid;
            this.PrimaryTokenPrimaryGroupSid = primaryTokenPrimaryGroupSid;
        }

        public new Body RecordBody { get { return (Body)base.RecordBody; } }
        public SecurityIdentifier ClientTokenOwnerSid { get; private set; }
        public SecurityIdentifier ClientTokenPrimaryGroupSid { get; private set; }
        public SecurityIdentifier PrimaryTokenOwnerSid { get; private set; }
        public SecurityIdentifier PrimaryTokenPrimaryGroupSid { get; private set; }

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        public struct Body
        {
            public UInt32 ImpersonationLevel;
            public UInt32 ClientTokenOwnerSidLength;
            public UInt32 ClientTokenPrimaryGroupSidLength;
            public UInt32 PrimaryTokenOwnerSidLength;
            public UInt32 PrimaryTokenPrimaryGroupSidLength;

            //
            // client token & primary token in order of lengths
            //
        }
    }

    public class SecurityDescriptorRecord
        : ObjectLogRecord
    {
        public SecurityDescriptorRecord(
            ObjectTraceLog log,
            RecordHeader header,
            SecurityDescriptorRecord.Body body,
            CommonSecurityDescriptor securityDescriptor)
            : base(log, header, body)
        {
            this.SecurityDescriptor = securityDescriptor;
        }

        public new Body RecordBody { get { return (Body)base.RecordBody; } }
        public CommonSecurityDescriptor SecurityDescriptor { get; private set; }

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        public struct Body
        {
            public UInt32 SdLength;

            //
            // security descriptor
            //
        }
    }

    public class NameRecord
        : ObjectLogRecord
    {
        public NameRecord(
            ObjectTraceLog log,
            RecordHeader header,
            NameRecord.Body body,
            string name)
            : base(log, header, body)
        {
            this.Name = name;
        }

        public new Body RecordBody { get { return (Body)base.RecordBody; } }
        public string Name { get; private set; }

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        public struct Body
        {
            public UInt16 NameLength;

            //
            // name
            //
        }
    }

    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public struct CallStackFrameRecord
    {
        public UInt64 Rsp;
        public UInt64 Rip;
        public UInt64 ImageBase;
    }


    public class CallStackRecord
        : ObjectLogRecord
    {
        public CallStackRecord(
            ObjectTraceLog log,
            RecordHeader header,
            CallStackRecord.Body body,
            CallStackFrameRecord[] frames)
            : base(log, header, body)
        {
            this.Frames = frames;
        }

        public new Body RecordBody { get { return (Body)base.RecordBody; } }
        public CallStackFrameRecord[] Frames { get; private set; }

        public Int64 Hash
        {
            get
            {
                StringBuilder builder = new StringBuilder();

                foreach (CallStackFrameRecord frameRecord in Frames)
                    builder.AppendFormat("{0}-{1}-{2}",
                        frameRecord.Rip,
                        frameRecord.Rsp,
                        frameRecord.ImageBase);

                return builder.ToString().GetHashCode();
            }
        }

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        public struct Body
        {
            public UInt32 Flags;
            public UInt32 NumberOfFrames;

            //
            // array of CallStackFrameRecords
            //
        }
    }

    public class ImageInfoRecord
        : ObjectLogRecord
    {
        public ImageInfoRecord(
            ObjectTraceLog log,
            RecordHeader header,
            ImageInfoRecord.Body body,
            string imageName)
            : base(log, header, body)
        {
            this.ImageName = imageName;
        }

        public new Body RecordBody { get { return (Body)base.RecordBody; } }
        public string ImageName { get; private set; }

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        public struct Body
        {
            public UInt64 BaseAddress;
            public UInt64 ImageSize;
            public UInt16 ImageNameLength;

            public UInt64 EndAddress
            {
                get
                {
                    return BaseAddress + ImageSize;
                }
            }


            //
            // image name
            //
        }
    }

    
    public class AccessRecord
        : ObjectLogRecord
    {
        public AccessRecord(
            ObjectTraceLog log,
            RecordHeader header,
            AccessRecord.Body body,
            SecurityContextRecord securityContext,
            SecurityDescriptorRecord securityDescriptor,
            CallStackRecord callStack)
            : base(log, header, body)
        {
            this.SecurityContext = securityContext;
            this.SecurityDescriptor = securityDescriptor;
            this.CallStack = callStack;
        }

        public new Body RecordBody { get { return (Body)base.RecordBody; } }
        public SecurityContextRecord SecurityContext { get; private set; }
        public SecurityDescriptorRecord SecurityDescriptor { get; private set; }
        public CallStackRecord CallStack { get; private set; }

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        public struct Body
        {
            public UInt64 ProcessObjectKey;
            public UInt32 Flags;
            public UInt32 AccessMask;
            public UInt32 SecurityContextLength;
            public UInt32 SecurityDescriptorLength;
            public UInt32 CallStackLength;

            public bool IsDefinition
            {
                get { return (Flags & 0x1) != 0; }
            }

            public bool IsUse
            {
                get { return !IsDefinition; }
            }

            //
            // security context, security descriptor, and call stack
            //
        }
    }

    #endregion
    
    public class ObjectTraceLog : 
        IEnumerable<ObjectLogRecord>,
        IEnumerator<ObjectLogRecord>,
        IDisposable
    {
        /// <summary>
        /// Creates an object log that is tied to a file
        /// </summary>
        /// <param name="path">The path that contains object log records</param>
        /// <returns></returns>
        public static ObjectTraceLog Open(string path)
        {
            return new ObjectTraceLog(
                Path.GetFileNameWithoutExtension(path).Replace("Persisted", ""), 
                new BufferedStream(File.OpenRead(path)));
        }

        /// <summary>
        /// Initializes the object log to use a specific stream
        /// </summary>
        /// <param name="objectType">The type of object the stream records are associated with</param>
        /// <param name="stream">The stream to interact with</param>
        public ObjectTraceLog(string objectType, Stream stream)
        {
            this.ObjectType = objectType;
            this.Stream = stream;

            // Per-object log state used when reading to improve performance
            HeaderBuffer = new byte[Marshal.SizeOf(typeof(RecordHeader))];
            HeaderGCHandle = GCHandle.Alloc(HeaderBuffer, GCHandleType.Pinned);
            HeaderPtr = HeaderGCHandle.AddrOfPinnedObject();

            BodySize = 8192;
            BodyBuffer = new byte[BodySize];
            BodyGCHandle = GCHandle.Alloc(BodyBuffer, GCHandleType.Pinned);
            BodyPtr = BodyGCHandle.AddrOfPinnedObject();
        }

        /// <summary>
        /// Disposes of the object and its associated stream
        /// </summary>
        public void Dispose()
        {
            if (Stream != null)
            {
                Stream.Dispose();
                Stream = null;
            }

            if (HeaderGCHandle.IsAllocated)
                HeaderGCHandle.Free();

            if (BodyGCHandle.IsAllocated)
                BodyGCHandle.Free();
        }

        /// <summary>
        /// Fetches the next record.  If one is not found, null is returned.
        /// </summary>
        /// <returns></returns>
        public ObjectLogRecord FetchNext()
        {
            MoveNext();

            return Current;
        }

        /// <summary>
        /// Gets the next object record
        /// </summary>
        /// <returns>True if an object record was obtained</returns>
        public bool MoveNext()
        {
            try
            {
                Current = ObjectLogRecord.FromObjectLog(this);
            }
            catch (EndOfStreamException)
            {
                Current = null;
            }

            return Current != null;
        }

        /// <summary>
        /// Reset the object log back to the beginning
        /// </summary>
        public void Reset()
        {
            Stream.Seek(0, SeekOrigin.Begin);

            Current = null;
        }

        public IEnumerator<ObjectLogRecord> GetEnumerator()
        {
            throw new NotImplementedException();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }

        public ObjectLogRecord Current { get; private set; }
        object IEnumerator.Current
        {
            get { return Current;  }
        }

        public string ObjectType { get; private set; }
        public Stream Stream { get; private set; }

        public byte[] HeaderBuffer { get; private set; }
        public IntPtr HeaderPtr { get; private set; }
        public GCHandle HeaderGCHandle { get; private set; }

        public int BodySize { get; private set; }
        public byte[] BodyBuffer { get; private set; }
        public IntPtr BodyPtr { get; private set; }
        public GCHandle BodyGCHandle { get; private set; }
    }
}
