﻿//
// This program extracts data flow information from object trace logs
// and persists it to a normalized database.
//
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Security.Principal;
using System.Security.AccessControl;

using TBA.Database;
using TBA.Graphing;
using TBA.Analysis;

using UR.Common;
using UR.Graphing;
using UR.Ui;

namespace TBA
{
    /// <summary>
    /// Compares two sids to see if sid1 is less privileged than sid2
    /// </summary>
    /// <param name="sid1">The first SID to compare</param>
    /// <param name="sid2">The second SID to compare</param>
    /// <returns>True if sid1 is less privileged than sid2</returns>
    public delegate bool LessPrivilegedThanDelegate(
        SecurityIdentifier sid1,
        SecurityIdentifier sid2);

    /// <summary>
    /// Entry point class
    /// </summary>
    class Program
    {
        /// <summary>
        /// Entry point.  Kicks off the import simulator on the provided object
        /// trace logs.
        /// </summary>
        /// <param name="args">One or more object trace log paths</param>
        static void Main(
            string[] args)
        {
            Program program = new Program(
                new CommandLineArguments(args),
                new TBADataContext());

            program.Run();
        }

        /// <summary>
        /// Initializes an instance of the program
        /// </summary>
        /// <param name="args">Command line argument vector</param>
        /// <param name="db">Database connection context</param>
        public Program(
            CommandLineArguments args,
            TBADataContext db)
        {
            this.Args = args;
            this.Db = db;
        }

        /// <summary>
        /// Runs the program
        /// </summary>
        public void Run()
        {
            DateTime startTime = DateTime.Now;

            // Display the help banner if no arguments have been provided.
            if (Args.NoArguments)
            {
                Args.DisplayHelp();
                return;
            }

            if (Args.DisplayTime)
                Console.WriteLine("Start time: {0}", 
                    startTime);

            // Initialize the simulator with filters to limit the amount of data
            // that we must consider
            Simulator = new ImportSimulator(Db) {
                FilterDataFlow = this.FilterDataFlow,
                FilterDataFlowDefinition = this.FilterDataFlowDefinition,
                FilterDataFlowUse = this.FilterDataFlowUse
            };

            // If there are traces to be imported, import them now.
            if (Args.Traces.Length > 0)
            {
                Simulator.Simulate(Args.Traces);

                SnapshotId = Simulator.Snapshot.Id;
            }

            // Generate a trust boundary data flow graph if requested
            if (Args.TbdfgFilePath != null ||
                Args.TbsdfgFilePath != null)
            {
                if (Args.TbdfgFilePath != null)
                    DataFlowGraph.ToGraphML(Args.TbdfgFilePath);
                if (Args.TbsdfgFilePath != null)
                    DataFlowGraph.Summarize().ToGraphML(Args.TbsdfgFilePath);
            }

            // Dump elevation paths, attack horizons, and defense horizons as requested
            if (Args.DumpElevationPaths)
                DumpElevationPaths();

            if (Args.DumpAttackHorizon)
                DumpAttackHorizons();

            if (Args.DumpDefenseHorizon)
                DumpDefenseHorizons();

            if (Args.DisplayTime)
                Console.WriteLine("Stop time: {0} [{1} elapsed", 
                    DateTime.Now,
                    DateTime.Now - startTime);
        }

        /// <summary>
        /// Dumps the attack horizon for a given SID
        /// </summary>
        public void DumpAttackHorizons()
        {
            DumpHorizons(false);
        }

        /// <summary>
        /// Dumps the defense horizon for a given SID
        /// </summary>
        public void DumpDefenseHorizons()
        {
            DumpHorizons(true);
        }

        /// <summary>
        /// Dumps elevation paths from one SID to another SID
        /// </summary>
        public void DumpElevationPaths()
        {
            DirectedGraph elevationGraph = new DirectedGraph();
            DefaultSet visisted = new DefaultSet();
            WorkList wl = new WorkList();
            string fromSid = Args.FromSid;
            string toSid = Args.ToSid;

            if (fromSid == null || toSid == null)
            {
                Console.WriteLine(
                    "You must specify a from and to sid (/F, /T)");
                return;
            }

            wl.Add(fromSid);

            while (!wl.IsEmpty)
            {
                string currentSid = wl.Dequeue() as string;

                if (visisted.Contains(currentSid))
                    continue;

                visisted.Add(currentSid);

                // Select data flows originating from the current SID that threaten
                // the SID that uses the data
                var q =
                    from df in Db.DataFlows
                    where
                        df.DataFlowDefinition.Actor.Name == currentSid &&
                        ((ThreatBits)df.DataFlowUse.Verb.ThreatBitMask & ThreatBits.EoP) != 0 &&
                        df.DataFlowDefinition.SnapshotId == SnapshotId
                    select df;

                foreach (DataFlow df in q)
                {
                    elevationGraph.AddEdge(
                        df.DataFlowDefinition.Actor,
                        df.DataFlowUse.Actor);

                    // If the use actor is not our target SID, add it to the work list
                    if (df.DataFlowUse.Actor.Name != toSid)
                        wl.Add(df.DataFlowUse.Actor.Name);
                }                        
            }

            elevationGraph.ToGraphML("elevationGraph.graphml");
        }

        /// <summary>
        /// Dumps the attack or defense horizon for a given SID
        /// </summary>
        private void DumpHorizons(bool defenseHorizon)
        {            
            // Get horizon information for each SID
            foreach (string sid in Args.Sids)
            {
                // Use the appropriate query depending on whether or not we're
                // dumping attack or defense horizons
                var q =
                    (defenseHorizon) ?
                        (from df in Db.DataFlows
                        where
                            df.DataFlowUse.Actor.Name == sid &&
                            df.DataFlowUse.ActorId != df.DataFlowDefinition.ActorId &&
                            df.DataFlowUse.SnapshotId == SnapshotId                            
                        select df) :
                        (from df in Db.DataFlows
                        where
                            df.DataFlowDefinition.Actor.Name == sid &&
                            df.DataFlowUse.ActorId != df.DataFlowDefinition.ActorId &&
                            df.DataFlowUse.SnapshotId == SnapshotId                    
                        select df);

                Console.WriteLine(
                    "\n" +
                    "Dumping {0} horizon for SID: {1}\n" +
                    "\n",
                    defenseHorizon ? "defense" : "attack",
                    sid);

                // Display information about each data flow representing the 
                // horizon for the current SID
                int index = 1;

                foreach (DataFlow df in q)
                {
                    // If the definition SID is not less privileged than the use SID,
                    // then we don't care about this data flow
                    if (!IsLessPrivilegedThan(
                        df.DataFlowDefinition.Actor.Sid,
                        df.DataFlowUse.Actor.Sid))
                        continue;

                    DumpDataFlow(index, df);

                    index++;
                }
            }
        }

        /// <summary>
        /// Dumps information about the provided data flow to the console
        /// </summary>
        /// <param name="index">The index of the data flow</param>
        /// <param name="df">The data flow to display information about</param>
        private void DumpDataFlow(
            int index,
            DataFlow df)
        {
            Console.WriteLine(
                "{1}.\t{0}",
                df,
                index);

            if (!Args.DumpCallStack && !Args.DumpProcessContext)
                return;

            DumpDataFlowDetail(df.DataFlowDefinition);
            DumpDataFlowDetail(df.DataFlowUse);
        }

        private void DumpDataFlowDetail(IDataFlowHalf half)
        {
            Console.WriteLine("\t   Details for {0}", half);

            if (Args.DumpCallStack &&
                half.ExecutionContext != null &&
                half.CallStack != null)
            {
                Console.WriteLine("\t   Call stack:\n");

                foreach (CallStackFrame frame in half.CallStack.CallStackFrames)
                {
                    Console.WriteLine(
                        "\t\t{0}", 
                        FrameToString(
                            half.ExecutionContext.ProcessObjectKey,
                            frame));
                }
            }

            // Dump process context information
            if (Args.DumpProcessContext &&
                half.ExecutionContext != null)
            {
                Console.WriteLine(
                    "\t   Process: {0}\n" +
                    "\t   Primary Token: O:{1} G:{2}",
                    Simulator.GetProcessName(
                        (ulong)half.ExecutionContext.ProcessObjectKey),
                    half.ExecutionContext.PrimaryTokenOwner.Sid.GetNtAccountName(),
                    half.ExecutionContext.PrimaryTokenPrimaryGroup.Sid.GetNtAccountName());

                if (half.ExecutionContext.ClientTokenOwner != null)
                {
                    Console.WriteLine(
                        "\t   Client Token: O:{0} G:{1}",
                        half.ExecutionContext.ClientTokenOwner.Sid.GetNtAccountName(),
                        half.ExecutionContext.ClientTokenPrimaryGroup.Sid.GetNtAccountName());
                }
            }
        }

        /// <summary>
        /// Converts a call stack frame to a string
        /// </summary>
        /// <param name="processObjectKey">The process object to lookup the frame in</param>
        /// <param name="frame">The frame to lookup</param>
        /// <returns>A readable string describing the frame</returns>
        private string FrameToString(
            long processObjectKey,
            CallStackFrame frame)
        {
            string imageName;
            string str;
            int rva;

            if (Simulator.FindImageRva(
                (ulong)processObjectKey,
                (ulong)frame.Rip,
                out imageName,
                out rva))
            {
                str = String.Format(
                    "{0}+{1:X}",
                    imageName,
                    rva);
            }
            else
                str = String.Format(
                    "{0:X16}", 
                    frame.Rip);

            return str;
        }

        /// <summary>
        /// Checks to see if sid1 is less privileged than sid2
        /// </summary>
        /// <remarks>
        /// If no LessPrivilegedThanDelegate has been defined, this routine will return
        /// false if sid1 is in the TCB and true otherwise.
        /// </remarks>
        /// <param name="sid1">The first SID to compare</param>
        /// <param name="sid2">The second SID to compare</param>
        /// <returns>True if sid1 is less privileged than sid2</returns>
        public bool IsLessPrivilegedThan(
            SecurityIdentifier sid1,
            SecurityIdentifier sid2)
        {
            if (LessPrivilegedThanDelegate != null)
                return LessPrivilegedThanDelegate(sid1, sid2);
            else
            {
                if (IsTcbSid(sid1))
                    return false;

                return true;
            }
        }

        /// <summary>
        /// Filters a data flow
        /// </summary>
        /// <remarks>
        /// If the definition SID is not less privileged than the use SID, return false
        /// </remarks>
        /// <param name="flow">The data flow to filter</param>
        /// <returns>True if the data flow should be filtered out</returns>
        private bool FilterDataFlow(
            DataFlow flow)
        {
            return !IsLessPrivilegedThan(
                flow.DataFlowDefinition.Actor.Sid,
                flow.DataFlowUse.Actor.Sid);
        }

        /// <summary>
        /// Filter data flow definition vertices
        /// </summary>
        /// <param name="def">Data flow definition</param>
        /// <returns>True if the data flow definition should be filtered out</returns>
        private bool FilterDataFlowDefinition(
            DataFlowDefinition def)
        {
            // If the definition actor is in the TCB then we don't care since we're
            // only interested in elevation paths
            return IsTcbSid(def.Actor.Sid);
        }

        /// <summary>
        /// Filter data flow use vertices
        /// </summary>
        /// <param name="use">Data flow use</param>
        /// <returns>True if the data flow use should be filtered out</returns>
        private bool FilterDataFlowUse(
            DataFlowUse use)
        {
            // If the use actor is not being threatened with EoP, skip this
            if (((ThreatBits)use.Verb.ThreatBitMask & ThreatBits.EoP) == 0)
                return true;

            // If the use actor is not part of the TCB, then we filter it
            return !IsTcbSid(use.Actor.Sid);
        }

        /// <summary>
        /// Checks to see if the provided SID is part of the TCB on Windows
        /// </summary>
        /// <param name="sid">The SID to check</param>
        /// <returns>True if the SID is part of the TCB</returns>
        private bool IsTcbSid(SecurityIdentifier sid)
        {
            string value = sid.Value;

            return
                value == "S-1-5-18" || // LocalSystem
                value == "S-1-5-19" || // LocalService
                value == "S-1-5-20" || // NetworkService
                value == "S-1-5-32-544" || // Administrators
                value == "S-1-5-32-548" || // Account Operators
                value == "S-1-5-32-549" || // Server Operators
                value == "S-1-5-32-550" || // Print Operators
                value == "S-5-1-32-551" || // Backup Operators
                value.EndsWith("-500") || // Administrator
                value.EndsWith("-512"); // DomainAdmins
        }

        /// <summary>
        /// Arguments to the program
        /// </summary>
        public CommandLineArguments Args { get; private set; }
        /// <summary>
        /// The program's database context to use for LINQ queries
        /// </summary>
        public TBADataContext Db { get; private set; }
        /// <summary>
        /// The identifier of the snapshot that is being analyzed
        /// </summary>
        public int SnapshotId
        {
            get
            {
                return Db.SnapshotId;
            }
            private set
            {
                dataFlowGraph = null;

                Db.SnapshotId = value;
            }
        }
        /// <summary>
        /// Delegates that checks to see if a SID is less privileged than another SID
        /// </summary>
        public LessPrivilegedThanDelegate LessPrivilegedThanDelegate { get; set; }

        /// <summary>
        /// Simulator instance
        /// </summary>
        private ImportSimulator Simulator { get; set; }

        /// <summary>
        /// A data flow graph for the current snapshot
        /// </summary>
        public DataFlowGraph DataFlowGraph
        {
            get
            {
                if (dataFlowGraph == null)
                    dataFlowGraph =  new DataFlowGraph(
                        Db,
                        FilterDataFlow);

                return dataFlowGraph;
            }
        }
        private DataFlowGraph dataFlowGraph;

        /// <summary>
        /// Command line argument parser for this application
        /// </summary>
        public class CommandLineArguments
        {
            /// <summary>
            /// Parses the provided command line arguments
            /// </summary>
            /// <param name="args"></param>
            public CommandLineArguments(string[] args)
            {
                CommandLineParser parser = new CommandLineParser(
                    new CommandLineSwitch[] {
                        new CommandLineSwitch("/d", SwDumpDefenseHorizon),
                        new CommandLineSwitch("/a", SwDumpAttackHorizon),
                        new CommandLineSwitch("/s", true, SwSids),
                        new CommandLineSwitch("/i", true, SwImportTraces),
                        new CommandLineSwitch("/o", true, SwOutputDetail),
                        new CommandLineSwitch("/t", true, SwThreatType),
                        new CommandLineSwitch("/F", true, SwFromSid),
                        new CommandLineSwitch("/T", true, SwToSid),
                        new CommandLineSwitch("/m", true, SwSidMembershipFile),
                        new CommandLineSwitch("/g", true, SwGenerateTbdfg),
                        new CommandLineSwitch("/G", true, SwGenerateTbsdfg),
                        new CommandLineSwitch("/time", false, SwTime),
                        new CommandLineSwitch("/help", SwHelp)
                    });

                if (args.Length == 0)
                    NoArguments = true;

                // Initialize to sane defaults
                DumpDefenseHorizon = false;
                DumpAttackHorizon = false;
                DisplayTime = false;
                TbdfgFilePath = null;
                TbsdfgFilePath = null;
                Traces = new string[0];

                // Now parse the command line arguments
                parser.Parse(args);
            }

            /// <summary>
            /// Displays the help banner
            /// </summary>
            public void DisplayHelp()
            {
                Console.WriteLine(
                    "TODO");
            }

            /// <summary>
            /// Processes the /help command line switch
            /// </summary>
            /// <param name="context">Unused</param>
            /// <param name="value">The value provided to this switch, if any</param>
            /// <returns>True on success</returns>
            public bool SwHelp(object context, string value)
            {
                DisplayHelp();
                return true;
            }

            public bool SwDumpDefenseHorizon(object context, string value)
            {
                DumpDefenseHorizon = true;
                return true;
            }
            public bool SwDumpAttackHorizon(object context, string value)
            {
                DumpAttackHorizon = true;
                return true;
            }
            public bool SwSids(object context, string value)
            {
                Sids = value.Split(new char[] { ',' });
                return true;
            }
            public bool SwOutputDetail(object context, string value)
            {
                foreach (string v in value.Split(new char[] { ',' }))
                {
                    if (v == "callstack")
                        DumpCallStack = true;
                    if (v == "pinfo")
                        DumpProcessContext = true;
                }

                return true;
            }
            public bool SwThreatType(object context, string value)
            {
                return true;
            }
            public bool SwFromSid(object context, string value)
            {
                DumpElevationPaths = true;
                FromSid = value;
                return true;
            }
            public bool SwToSid(object context, string value)
            {
                DumpElevationPaths = true;
                ToSid = value;
                return true;
            }
            public bool SwSidMembershipFile(object context, string value)
            {
                return true;
            }
            public bool SwGenerateTbdfg(object context, string value)
            {
                TbdfgFilePath = value;
                return true;
            }
            public bool SwGenerateTbsdfg(object context, string value)
            {
                TbsdfgFilePath = value;
                return true;
            }
            public bool SwTime(object context, string value)
            {
                DisplayTime = true;
                return true;
            }


            /// <summary>
            /// Processes the /import command line switch
            /// </summary>
            /// <param name="context">Unused</param>
            /// <param name="value">The value provided to this switch, if any</param>
            /// <returns>True on success</returns>
            public bool SwImportTraces(object context, string value)
            {
                Traces = value.Split(new char[] { ',' });
                return true;
            }

            /// <summary>
            /// The list of trace files to be imported
            /// </summary>
            public string[] Traces { get; private set; }
            /// <summary>
            /// True if no arguments were provided
            /// </summary>
            public bool NoArguments { get; private set; }
            public bool DumpDefenseHorizon { get; private set; }
            public bool DumpAttackHorizon { get; private set; }
            public string TbdfgFilePath { get; private set; }
            public string TbsdfgFilePath { get; private set; }
            public bool DumpElevationPaths { get; private set; }
            public string FromSid { get; private set; }
            public string ToSid { get; private set; }
            public string[] Sids { get; private set; }
            public bool DisplayTime { get; private set; }
            public bool DumpCallStack { get; private set; }
            public bool DumpProcessContext { get; private set; }
        }
    }

    /// <summary>
    /// Filters a data flow
    /// </summary>
    /// <param name="flow">The data flow to filter</param>
    /// <returns>True if the flow should be filtered out</returns>
    public delegate bool FilterDataFlowDelegate(
        DataFlow flow);
    /// <summary>
    /// Filters a data flow definition
    /// </summary>
    /// <param name="def">The data flow definition to filter</param>
    /// <returns>True if the definition should be filtered out</returns>
    public delegate bool FilterDataFlowDefinitionDelegate(
        DataFlowDefinition def);
    /// <summary>
    /// Filters a data flow use
    /// </summary>
    /// <param name="use">The data flow to filter</param>
    /// <returns>True if the use should be filtered out</returns>
    public delegate bool FilterDataFlowUseDelegate(
        DataFlowUse use);

    /// <summary>
    /// Custom object simulator that imports information to a normalized database
    /// </summary>
    public class ImportSimulator
        : ObjectSimulator
    {
        /// <summary>
        /// Initializes the simulator
        /// </summary>
        public ImportSimulator()
            : this(new TBADataContext())
        {
        }

        /// <summary>
        /// Initializes the simulator
        /// </summary>
        /// <param name="db">The database context to use</param>
        public ImportSimulator(TBADataContext db)
        {
            this.Db = db;
        }

        /// <summary>
        /// Called when simulation is starting.  Initializes flags that cause
        /// the underlying simulator to track process objects and the images
        /// contained within process objects.  This is needed for determining
        /// which module a virtual address is contained within.
        /// 
        /// Additionally, a snapshot is created that will uniquely identify the
        /// data imported during this run.
        /// 
        /// A verb graph is also created which describes the methods through
        /// which data can be transmitted for each of the various object types.
        /// </summary>
        public override void OnBegin()
        {
            // Set flags to control how the base class collects data
            TrackProcesses = true;
            TrackProcessImages = true;

            // Acquire our snapshot
            this.Snapshot = new Snapshot();
            Snapshot.Name = "Default snapshot";
            Db.Snapshots.InsertOnSubmit(Snapshot);
            Db.SubmitChanges();

            // Initialize the verb graph
            this.VerbGraph = new VerbGraph(Db);

            // Initialize our dictionaries
            this.CallStackDictionary = new Dictionary<long, CallStack>();
            this.ModuleDictionary = new Dictionary<string, Module>();
            this.ActorDictionary = new Dictionary<string, Actor>();
            this.MediumDictionary = new Dictionary<ulong, Medium>();
            this.ExecutionContextDictionary = new Dictionary<int, ExecutionContext>();
        }

        /// <summary>
        /// When simulation is finished, the last step is to determine
        /// relationships between definition and use data flow parts.  These
        /// relationships are found whenever a data flow definition and a 
        /// data flow use are operating on the same medium with complementary
        /// verbs.  Verbs are complementary when they are dependent on one
        /// another in order to produce a complete data flow.  For example
        /// the writing to a file and reading from a file are complementary.
        /// </summary>
        public override void OnEnd()
        {
            // Submit whatever pending changes we have at this point
            Db.SubmitChanges();

            var q =
                from def in Db.DataFlowDefinitions
                from use in Db.DataFlowUses
                where use.MediumId == def.MediumId
                select new { Use = use, Definition = def };

            foreach (var row in q)
            {
                if (!VerbGraph.IsRelated(
                    row.Use.Verb,
                    row.Definition.Verb))
                {
                    continue;
                }

                DataFlow flow = new DataFlow();

                flow.DataFlowDefinition = row.Definition;
                flow.DataFlowUse = row.Use;

                // Should this data flow be filtered and not inserted?
                if (FilterDataFlow(flow))
                {
                    continue;
                }

                Db.DataFlows.InsertOnSubmit(flow);
            }

            Db.SubmitChanges();
        }

        /// <summary>
        /// Called whenever a record is encountered during simulation.  A record
        /// can take a number of different forms.
        /// </summary>
        /// <param name="record">The record being processed</param>
        public override void OnRecord(
            ObjectLogRecord record)
        {
            // Call the base class first to give it an opportunity to track this
            base.OnRecord(record);

            // Now extract the information we need to import
            switch (record.Type)
            {
                // When an object is defined or used
                case RecordType.Access:
                    ProcessAccessRecord(
                        record as AccessRecord);
                    break;

                // When an object's name is determined
                case RecordType.Name:
                    ProcessNameRecord(
                        record as NameRecord);
                    break;

                // When an object's security descriptor is assigned or modified
                case RecordType.SecurityDescriptor:
                    ProcessSecurityDescriptorRecord(
                        record as SecurityDescriptorRecord);
                    break;

                default:
                    break;
            }
        }

        /// <summary>
        /// Called when an access record is encountered during simulation.  An
        /// access required describes the context when an object is defined or 
        /// used.
        /// 
        /// To process an access record, each verb in the verb graph is enumerated.
        /// A verb matches the access record in terms of object type, whether or not
        /// it is a definition or a use, and the access rights granted.
        /// </summary>
        /// <param name="record">The access record to process</param>
        private void ProcessAccessRecord(
            AccessRecord record)
        {
            // Enumerate each verb in the verb graph to determine if it is
            // compatible with this access record
            foreach (Verb verb in VerbGraph.Vertices)
            {
                // Figure out our access mask
                int accessMask =
                    (record.RecordBody.IsDefinition) 
                        ? unchecked((int)0xffffffff) 
                        : unchecked((int)record.RecordBody.AccessMask);

                // Does this verb match what we're looking for?
                if (!verb.IsMatch(
                    record.TraceLog.ObjectType,
                    record.RecordBody.IsDefinition,
                    accessMask))
                {
                    continue;
                }

                // Get the medium associated with the object that is being
                // defined or used.  Currently the device driver doesn't correctly
                // differentiate ALPC server/client object definitions, so we always
                // re-use previous meduims.  This can lead to some inaccuracy/missed
                // data until the driver is fixed.
                Medium medium =
                    GetMediumFromObjectKey(
                        record.TraceLog.ObjectType == "ALPC Port" ? false : record.RecordBody.IsDefinition,
                        record);

                // Get the call stack that is associated with this definition
                // or use
                CallStack callStack =
                    GetCallStackFromRecord(
                        record.RecordBody.ProcessObjectKey,
                        record.CallStack);

                // Get the active execution context
                ExecutionContext executionContext =
                    GetExecutionContext(
                        record.RecordBody.ProcessObjectKey,
                        record.SecurityContext);

                // If this is a definition, create a data flow vertex
                // where the actor is either the security descriptor owner sid
                // or the active security context client or primary owner sid.
                if (record.RecordBody.IsDefinition)
                {
                    SecurityIdentifier sid = null;

                    // If the definition has a security descriptor, we assume
                    // the owner SID of the security descriptor
                    if (record.SecurityDescriptor != null &&
                        record.SecurityDescriptor.SecurityDescriptor.Owner != null)
                    {
                        sid = record.SecurityDescriptor.SecurityDescriptor.Owner;

                        ProcessSecurityDescriptorRecord(
                            record.SecurityDescriptor);
                    }
                    // Otherwise, we look at the security context using the
                    // client token owner SID in preference to the primary token
                    // owner SID
                    else
                    {
                        if (record.SecurityContext.ClientTokenOwnerSid != null)
                            sid = record.SecurityContext.ClientTokenOwnerSid;
                        else
                            sid = record.SecurityContext.PrimaryTokenOwnerSid;
                    }

                    if (sid == null)
                        continue;

                    // Create a vertex for this data flow component
                    CreateDataFlowVertex(
                        true,
                        executionContext,
                        sid,
                        medium,
                        verb, 
                        callStack);
                }
                // If this is a use, create vertices for each of the SIDs
                // in the active security context
                else
                {
                    // Currently we ignore uses for ALPC port objects as they
                    // do not accurately capture the usage behavior
                    if (record.TraceLog.ObjectType == "ALPC Port")
                        continue;

                    if (record.SecurityContext.ClientTokenOwnerSid != null)
                        CreateDataFlowVertex(
                            true,
                            executionContext,
                            record.SecurityContext.ClientTokenOwnerSid,
                            medium,
                            verb,
                            callStack);

                    if (record.SecurityContext.ClientTokenPrimaryGroupSid != null)
                        CreateDataFlowVertex(
                            true,
                            executionContext,
                            record.SecurityContext.ClientTokenPrimaryGroupSid,
                            medium,
                            verb,
                            callStack);

                    if (record.SecurityContext.PrimaryTokenOwnerSid != null)
                        CreateDataFlowVertex(
                            true,
                            executionContext,
                            record.SecurityContext.PrimaryTokenOwnerSid,
                            medium,
                            verb,
                            callStack);

                    if (record.SecurityContext.PrimaryTokenPrimaryGroupSid != null)
                        CreateDataFlowVertex(
                            true,
                            executionContext,
                            record.SecurityContext.PrimaryTokenPrimaryGroupSid,
                            medium,
                            verb,
                            callStack);
                }
            }
        }

        /// <summary>
        /// Called when a name record is encountered during simulation.
        /// </summary>
        /// <remarks>
        /// A name record provides the name given to an object.  This method
        /// associates the name with the medium that encapsulates the object.
        /// </remarks>
        /// <param name="record">The name record to process</param>
        private void ProcessNameRecord(
            NameRecord record)
        {
            Medium medium = 
                GetMediumFromObjectKey(
                    false,
                    record);

            // Update the object medium's name
            medium.Name = record.Name;
        }

        /// <summary>
        /// Called when an record describing an object's security descriptor is encountered during simulation.
        /// </summary>
        /// <remarks>
        /// This occurs when an object's security descriptor is assigned or modified.
        /// 
        /// Each ACE in the security descriptor's DACL describe the extent to which
        /// individual SIDs can interact with the object.  These access rights define the
        /// extent to which each SID can potentially influence one another through the object.
        /// As such, data flow vertices are created for each ace where the actor is the ACE's SID.
        /// </remarks>
        /// <param name="record"></param>
        private void ProcessSecurityDescriptorRecord(
            SecurityDescriptorRecord record)
        {
            // Walk each ACE in the security descriptor's DACL so that we can
            // determine compatible verbs
            foreach (CommonAce ace in record.SecurityDescriptor.DiscretionaryAcl)
            {
                // Skip access denied ACEs.  We only care about what /can/ happen.
                if (ace.AceType == AceType.AccessDenied)
                    continue;

                // Enumerate each verb in the verb graph to find cases where
                // the ACE's access rights and object type match the criteria.
                foreach (Verb verb in VerbGraph.Vertices)
                {
                    // Does the current ACE match this verb?
                    if (!verb.IsMatch(
                        record.TraceLog.ObjectType,
                        false,
                        ace.AccessMask))
                    {
                        continue;
                    }

                    // Get the medium and actor being affected
                    Medium medium = 
                        GetMediumFromObjectKey(
                            false,
                            record);

                    CreateDataFlowVertex(
                        false,
                        null,
                        ace.SecurityIdentifier,
                        medium,
                        verb,
                        null);
                }
            }
        }

        /// <summary>
        /// Creates a data flow vertex which associates an actor, medium,
        /// and verb as either defining or using data.
        /// </summary>
        /// <param name="actualized">True if the data flow was derived from dynamic access rights</param>
        /// <param name="executionContext">The execution context that the action occurs in</param>
        /// <param name="sid">The SID associated with the data flow</param>
        /// <param name="medium">The medium through which data is flowing</param>
        /// <param name="verb">The manner through which data is flowing</param>
        /// <param name="callStack">The call stack that enabled this data flow</param>
        private void CreateDataFlowVertex(
            bool actualized,
            ExecutionContext executionContext,
            SecurityIdentifier sid,
            Medium medium,
            Verb verb,
            CallStack callStack)
        {
            Actor actor = GetActorFromSid(
                sid);

            if (verb.IsDefinition)
            {
                DataFlowDefinition definition = new DataFlowDefinition();

                definition.Snapshot = Snapshot;
                definition.Verb = verb;
                definition.Medium = medium;
                definition.Actor = actor;
                definition.CallStack = callStack;
                definition.Actualized = actualized;
                definition.ExecutionContext = executionContext;

                if (FilterDataFlowDefinition != null &&
                    FilterDataFlowDefinition(definition))
                    return;

                Db.DataFlowDefinitions.InsertOnSubmit(definition);

                // Check if this definition has any associated implicit uses
                foreach (UR.Graphing.Edge edge in VerbGraph.Successors(verb))
                {
                    Verb useVerb = edge.Target as Verb;

                    // FIXME: Right now we assume implicit use leads to
                    // LocalSystem -- this is not accurate.  It depends
                    // on the context in which the use occurs.
                    if (useVerb.MatchImplicitUse)
                    {
                        CreateDataFlowVertex(
                            true,
                            null,
                            new SecurityIdentifier("S-1-5-18"),
                            medium,
                            useVerb,
                            null);
                    }
                }
            }
            else
            {
                DataFlowUse use = new DataFlowUse();

                use.Snapshot = Snapshot;
                use.Verb = verb;
                use.Medium = medium;
                use.Actor = actor;
                use.CallStack = callStack;
                use.Actualized = actualized;
                use.ExecutionContext = executionContext;

                if (FilterDataFlowUse != null &&
                    FilterDataFlowUse(use))
                    return;

                Db.DataFlowUses.InsertOnSubmit(use);
            }
        }

        private ExecutionContext GetExecutionContext(
            UInt64 processObjectKey,
            SecurityContextRecord sc)
        {
            Actor pto = GetActorFromSid(sc.PrimaryTokenOwnerSid);
            Actor ptpg = GetActorFromSid(sc.PrimaryTokenPrimaryGroupSid);
            Actor cto = null;
            Actor ctpg = null;

            if (sc.ClientTokenOwnerSid != null)
                cto = GetActorFromSid(sc.ClientTokenOwnerSid);
            if (sc.ClientTokenPrimaryGroupSid != null)
                ctpg = GetActorFromSid(sc.ClientTokenPrimaryGroupSid);

            // Compute the hash key to lookup
            int hashKey = String.Format(
                "{0}-{1}-{2}-{3}-{4}",
                sc.PrimaryTokenOwnerSid.Value,
                sc.PrimaryTokenPrimaryGroupSid.Value,
                (cto != null) ? sc.ClientTokenOwnerSid.Value : "0",
                (ctpg != null) ? sc.ClientTokenPrimaryGroupSid.Value : "0",
                processObjectKey).GetHashCode();

            if (ExecutionContextDictionary.ContainsKey(hashKey))
                return ExecutionContextDictionary[hashKey];

            // Otherwise, we insert a new context into the database
            ExecutionContext context = new ExecutionContext()
            {
                ProcessObjectKey = (long)processObjectKey,
                ClientTokenOwner = cto,
                ClientTokenPrimaryGroup = ctpg,
                PrimaryTokenOwner = pto,
                PrimaryTokenPrimaryGroup = ptpg,
                ProcessName = GetProcessName(processObjectKey)
            };

            Db.ExecutionContexts.InsertOnSubmit(context);

            ExecutionContextDictionary[hashKey] = context;

            return context;
        }

        /// <summary>
        /// Gets the call stack associated with a call stack record
        /// </summary>
        /// <param name="processObjectKey">The process in which the call stack occurred</param>
        /// <param name="record">The call stack record</param>
        /// <returns>The call stack</returns>
        private CallStack GetCallStackFromRecord(
            UInt64 processObjectKey,
            CallStackRecord record)
        {
            Int64 callStackHash = record.Hash;

            // Check our local cache first before hitting the database
            if (CallStackDictionary.ContainsKey(callStackHash))
                return CallStackDictionary[callStackHash];

            // Is there a call stack in the database that matches this hash?
            CallStack callStack = new CallStack();

            callStack.Hash = callStackHash;
            callStack.Snapshot = Snapshot;

            // Encapsulate each call stack frame record
            foreach (CallStackFrameRecord frameRecord in record.Frames)
            {
                CallStackFrame frame = new CallStackFrame();
                Module module;
                int rva;

                GetModuleFromRip(
                    processObjectKey,
                    frameRecord.Rip,
                    out module,
                    out rva);

                frame.Rip = (long)frameRecord.Rip;
                frame.Rsp = (long)frameRecord.Rsp;
                frame.Module = module;
                frame.RipRva = rva;

                callStack.CallStackFrames.Add(frame);
            }

            Db.CallStacks.InsertOnSubmit(callStack);

            // Cache this for future reference
            CallStackDictionary[callStackHash] = callStack;

            return callStack;
        }

        /// <summary>
        /// Gets the module associated with an instruction pointer
        /// </summary>
        /// <remarks>
        /// Uses FindImageRva to get the image name associated with the module.
        /// From there, a module instance is obtained using the image name.  If
        /// either of these steps fail, a null module is passed as an output
        /// parameter and false is returned.
        /// </remarks>
        /// <param name="processObjectKey">The process object context</param>
        /// <param name="rip">The instruction pointer's virtual address</param>
        /// <param name="module">Output parameter that will hold the module</param>
        /// <param name="rva">Output parameter that will hold the RVA</param>
        /// <returns></returns>
        private bool GetModuleFromRip(
            UInt64 processObjectKey,
            UInt64 rip,
            out Module module,
            out int rva)
        {
            string imageName;

            if (FindImageRva(
                processObjectKey,
                rip,
                out imageName,
                out rva))
            {
                module = GetModuleFromImageName(
                    imageName);

                return true;
            }
            else
            {
                module = null;
                rva = 0;

                return false;
            }
        }

        /// <summary>
        /// Gets a module from an image name
        /// </summary>
        /// <param name="imageName">The module's image name</param>
        /// <returns>The module associated with the image name</returns>
        private Module GetModuleFromImageName(
            string imageName)
        {
            // Check our local cache first before hitting the database
            if (ModuleDictionary.ContainsKey(imageName))
                return ModuleDictionary[imageName];

            Module module = new Module();

            module.Name = imageName;
            Db.Modules.InsertOnSubmit(module);

            // Cache this for future reference
            ModuleDictionary[imageName] = module;

            return module;
        }

        /// <summary>
        /// Gets the actor that encapsulates the provided SID
        /// </summary>
        /// <param name="sid">The SID to encapsulate</param>
        /// <returns>The actor associated with the SID</returns>
        private Actor GetActorFromSid(
            SecurityIdentifier sid)
        {
            string sidName = sid.ToString();

            // Check the local cache first
            if (ActorDictionary.ContainsKey(sidName))
                return ActorDictionary[sidName];

            Actor actor = new Actor();

            actor.Name = sid.ToString();
            actor.ActorTypeId = 1;
            actor.Snapshot = Snapshot;

            Db.Actors.InsertOnSubmit(actor);

            // Update the local cache
            ActorDictionary[sidName] = actor;

            return actor;
        }

        /// <summary>
        /// Gets the medium that encapsulates the supplied object key
        /// </summary>
        /// <param name="createNew">If true, a new medium will be created</param>
        /// <param name="record"></param>
        /// <returns></returns>
        private Medium GetMediumFromObjectKey(
            bool createNew,
            ObjectLogRecord record)
        {
            Medium medium = null;
            bool insert = false;

            // If we shouldn't create a new instance then check the database
            // for the most recent one
            if (!createNew &&
                MediumDictionary.ContainsKey(record.ObjectKey))
            {
                return MediumDictionary[record.ObjectKey];
            }

            // If we still don't have a medium at this point, create a new one
            if (medium == null)
            {
                medium = new Medium();
                insert = true;
            }

            medium.Name = String.Format("{0}({1})", 
                record.TraceLog.ObjectType,
                record.ObjectKey);

            medium.MediumTypeId = 1;
            medium.UniqueKey = (long)record.ObjectKey;
            medium.CreationSequence = (long)record.Sequence;
            medium.Snapshot = Snapshot;

            if (insert)
                Db.Mediums.InsertOnSubmit(medium);

            // Cache this medium
            MediumDictionary[record.ObjectKey] = medium;

            return medium;
        }

        /// <summary>
        /// The snapshot the simulator is associated with
        /// </summary>
        public Snapshot Snapshot { get; private set; }
        /// <summary>
        /// LINQ front object
        /// </summary>
        private TBADataContext Db { get; set; }
        /// <summary>
        /// The verb graph describing the relationship between verbs
        /// </summary>
        private VerbGraph VerbGraph { get; set; }

        /// <summary>
        /// Delegate that allows data flows to be filtered
        /// </summary>
        public FilterDataFlowDelegate FilterDataFlow { get; set; }
        /// <summary>
        /// Delegate that allows data flow definitions to be filtered
        /// </summary>
        public FilterDataFlowDefinitionDelegate FilterDataFlowDefinition { get; set; }
        /// <summary>
        /// Delegate that allows data flow uses to be filtered
        /// </summary>
        public FilterDataFlowUseDelegate FilterDataFlowUse { get; set; }

        /// <summary>
        /// In-memory cache of actor name to actor association
        /// </summary>
        private Dictionary<string, Actor> ActorDictionary { get; set; }
        /// <summary>
        /// In-memory cache of module image name to module association
        /// </summary>
        private Dictionary<string, Module> ModuleDictionary { get; set; }
        /// <summary>
        /// In-memory cache of call stack hash to call stack association
        /// </summary>
        private Dictionary<Int64, CallStack> CallStackDictionary { get; set; }
        /// <summary>
        /// In-memory cache of object key to medium association
        /// </summary>
        private Dictionary<UInt64, Medium> MediumDictionary { get; set; }
        /// <summary>
        /// In-memory cache of execution contexts
        /// </summary>
        private Dictionary<int, ExecutionContext> ExecutionContextDictionary { get; set; }
    }

    /// <summary>
    /// Extends the verb class to with a method that checks to see if
    /// some given data matches the verb's criteria.
    /// </summary>
    public static class ExtensionMethods
    {
        /// <summary>
        /// Check if the provided data matches the verb instance's criteria
        /// </summary>
        /// <param name="verb">The verb this pointer</param>
        /// <param name="objectType">The object's type</param>
        /// <param name="isObjectDefinition">True if the object is being defined</param>
        /// <param name="accessMask">The access rights being granted</param>
        /// <returns>True if the verb's criteria matches the provided data</returns>
        public static bool IsMatch(
            this Verb verb,
            string objectType,
            bool isObjectDefinition,
            int accessMask)
        {
            // If the object types to do not match then the verb
            // cannot match
            if (objectType != verb.ObjectType)
                return false;

            // If the context is defining the object and we should match implicitly
            // when an object is defined, then we have found a match.
            if (isObjectDefinition && verb.MatchImplicitDefinition)
                return true;

            // If we should compare access rights and the rights required by the
            // verb are comparable to the access rights being granted, then we
            // have a match
            if (verb.MatchRights && ((verb.RightsRequired & accessMask) != 0))
                return true;

            // Otherwise, there is no match
            return false;
        }
    }
}
