﻿//
// Simulates the definitions and uses of objects by processing the records
// in object trace logs in execution order.
//
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace TBA.Analysis
{
    /// <summary>
    /// Simulates an object trace log
    /// </summary>
    public class ObjectSimulator
    {
        /// <summary>
        /// Initializes the default simulator flags
        /// </summary>
        public ObjectSimulator()
        {
            this.TrackProcesses = false;
            this.TrackProcessImages = false;
        }

        /// <summary>
        /// Processes the records of the supplied object trace logs in sequence order
        /// </summary>
        /// <param name="traceLogPaths">An array of file paths that contain object trace log data</param>
        public void Simulate(string[] traceLogPaths)
        {
            Reset();

            OnBegin();

            try
            {
                // Open each object trace log and add the first log record to the read queue.
                // We then sort the read queue so that we have them in order.
                foreach (string path in traceLogPaths)
                {
                    ObjectTraceLog log = ObjectTraceLog.Open(path);

                    if (!log.MoveNext())
                        continue;

                    TraceLogs.Add(log);
                    ReadQueue.Add(log.Current);
                }

                // Until the read queue is empty, we keep pulling records from the
                // queue, processing them, and then enqueuing the next record from
                // an object trace log
                while (ReadQueue.Count > 0)
                {
                    ObjectLogRecord record = DequeueFromReadQueue();

                    OnRecord(record);

                    ObjectLogRecord next = record.TraceLog.FetchNext();

                    if (next != null)
                        EnqueueToReadQueue(next);
                }

            }
            finally
            {
                OnEnd();

                foreach (ObjectTraceLog log in TraceLogs)
                    log.Dispose();
            }
        }

        /// <summary>
        /// Called when simulation is starting before trace logs have been opened
        /// </summary>
        public virtual void OnBegin()
        {
        }

        /// <summary>
        /// Called when simulation has completed but before trace logs have been closed
        /// </summary>
        public virtual void OnEnd()
        {
        }

        /// <summary>
        /// Called when a record is being processed from the queue
        /// </summary>
        /// <param name="record">The trace log record to process</param>
        public virtual void OnRecord(ObjectLogRecord record)
        {
            // If we are tracking processes and this record is an access to
            // a process, we process it
            if (TrackProcesses &&
                record.TraceLog.ObjectType == "Process")
            {
                // Track the creation of new processes
                if (record.Type == RecordType.Access)
                    TrackProcess(record as AccessRecord);
                // Update the process object's name when we encounter a name record
                else if (record.Type == RecordType.Name)
                {
                    ProcessObject process =
                        GetProcessObject(
                            record.ObjectKey);

                    process.Name = (record as NameRecord).Name;
                }
                // If we are tracking process images and this is an image info record,
                // we process it further
                else if (
                    TrackProcessImages &&
                    record.Type == RecordType.ImageInfo)
                {
                    TrackProcessImage(record as ImageInfoRecord);
                }
            }
        }

        /// <summary>
        /// Gets the name associated with a process
        /// </summary>
        /// <param name="processObjectKey">The process' object key</param>
        /// <returns>The name of the process executable</returns>
        public string GetProcessName(
            UInt64 processObjectKey)
        {
            ProcessObject process = null;

            if (ProcessObjectDictionary.ContainsKey(processObjectKey))
                process = ProcessObjectDictionary[processObjectKey];

            if (process == null ||
                process.Name == null)
                return String.Format("{0:X16}", processObjectKey);

            return process.Name;
        }

        /// <summary>
        /// Finds the RVA and image name associated with a given RIP in a given process context
        /// </summary>
        /// <param name="processObjectKey">The key of the process object context</param>
        /// <param name="rip">The RIP value</param>
        /// <param name="imageName">Stores the name of the image RIP is contained within, or null</param>
        /// <param name="rva">Stores the RVA of the RIP value within the image, or 0</param>
        /// <returns>True if the RVA was found to be within an image</returns>
        /// <remarks>
        /// TrackProcesses and TrackProcessImages must be enabled for this function to
        /// work properly.
        /// </remarks>
        public bool FindImageRva(
            UInt64 processObjectKey,
            UInt64 rip,
            out string imageName,
            out int rva)
        {
            // If we can't find the process associated with this key, we bail
            if (!ProcessObjectDictionary.ContainsKey(processObjectKey))
            {
                imageName = null;
                rva = 0;
                return false;
            }

            ProcessObject process = ProcessObjectDictionary[processObjectKey];

            // Enumerate each image record in the process' address space
            foreach (ImageInfoRecord record in process.ImageRecords)
            {
                // If the RIP value isn't within this image, we go on to the next one
                if (rip < record.RecordBody.BaseAddress)
                    continue;

                if (rip >= record.RecordBody.EndAddress)
                    continue;

                // If it's within the image region, we've found what we're looking for
                imageName = record.ImageName;
                rva = (int)(rip - record.RecordBody.BaseAddress);

                return true;
            }

            // Try resolving within the system process if we failed
            if (SystemProcess != null &&
                SystemProcess.Key != processObjectKey)
            {
                return FindImageRva(
                    SystemProcess.Key,
                    rip,
                    out imageName,
                    out rva);
            }
            else
            {
                // If we exit the loop, then no match was found
                imageName = null;
                rva = 0;

                return false;
            }
        }

        /// <summary>
        /// Creates a process object to correspond to the record's object key
        /// </summary>
        /// <param name="record">The record to process</param>
        private void TrackProcess(AccessRecord record)
        {
            // We only care about object definitions
            if (!record.RecordBody.IsDefinition)
                return;

            NewProcessObject(
                record.ObjectKey);
        }

        /// <summary>
        /// Adds an image record to the process object associated with the record's object key
        /// </summary>
        /// <param name="record">The image information record to process</param>
        private void TrackProcessImage(ImageInfoRecord record)
        {
            ProcessObject process =
                GetProcessObject(
                    record.ObjectKey);

            if (process != null)
                process.ImageRecords.Insert(0, record);
        }

        /// <summary>
        /// Gets the process object associated with the supplied object key
        /// </summary>
        /// <param name="processObjectKey">The process object key</param>
        /// <returns>The process object</returns>
        private ProcessObject GetProcessObject(ulong processObjectKey)
        {
            if (!ProcessObjectDictionary.ContainsKey(processObjectKey))
            {
                NewProcessObject(
                    processObjectKey);
            }

            return ProcessObjectDictionary[processObjectKey];
        }

        /// <summary>
        /// Creates a new process object and inserts it into the cache
        /// </summary>
        /// <param name="processObjectKey">The process object key</param>
        /// <returns>The corresponding process object</returns>
        private ProcessObject NewProcessObject(ulong processObjectKey)
        {
            ProcessObject process =
                new ProcessObject(processObjectKey);

            ProcessObjectDictionary[processObjectKey] =
                process;

            if (SystemProcess == null)
            {
                SystemProcess = process;
                SystemProcess.Name = "System";
            }

            return process;
        }

        /// <summary>
        /// Inserts the supplied record in execution order to the read queue
        /// </summary>
        /// <param name="record">The record to insert</param>
        private void EnqueueToReadQueue(ObjectLogRecord record)
        {
            int index;

            // Start searching from the front of the read queue under the assumption
            // that this record will likely execute next
            for (index = 0;
                 index < ReadQueue.Count && record.Sequence > ReadQueue[index].Sequence;
                 index++) ;

            ReadQueue.Insert(index, record);
        }

        /// <summary>
        /// Dequeue a record from the read queue
        /// </summary>
        /// <returns></returns>
        private ObjectLogRecord DequeueFromReadQueue()
        {
            ObjectLogRecord record = ReadQueue[0];
            ReadQueue.RemoveAt(0);

            return record;
        }

        /// <summary>
        /// Reset the state associated with this simulator
        /// </summary>
        public void Reset()
        {
            SystemProcess = null;

            TraceLogs = new List<ObjectTraceLog>();
            ReadQueue = new List<ObjectLogRecord>();
            ProcessObjectDictionary = new Dictionary<ulong, ProcessObject>();
        }

        /// <summary>
        /// True if process objects should be tracked by the simulator
        /// </summary>
        /// <remarks>
        /// Required for FindImageRva to work
        /// </remarks>
        public bool TrackProcesses { get; set; }
        /// <summary>
        /// True if mapped image information should be tracked by the simulator
        /// </summary>
        /// <remarks>
        /// Required for FindImageRva to work
        /// </remarks>
        public bool TrackProcessImages { get; set; }

        /// <summary>
        /// The System process object
        /// </summary>
        private ProcessObject SystemProcess { get; set; }
        /// <summary>
        /// A dictionary relating object key with process object instance
        /// </summary>
        private Dictionary<UInt64, ProcessObject> ProcessObjectDictionary { get; set; }
        /// <summary>
        /// A list of trace logs
        /// </summary>
        private List<ObjectTraceLog> TraceLogs { get; set; }
        /// <summary>
        /// A queue of log records to be processed
        /// </summary>
        private List<ObjectLogRecord> ReadQueue { get; set; }
    }
}
