﻿//
// A graph relating definition verbs with use verbs.  A verb represents an 
// action that is supported by an object type that either defines or uses
// data.
//
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using TBA.Database;

namespace TBA.Graphing
{
    /// <summary>
    /// Bits associated with each threat category from STRIDE
    /// </summary>
    [Flags]
    public enum ThreatBits
    {
        /// <summary>
        /// No threats
        /// </summary>
        Nil = 0,
        /// <summary>
        /// Elevation of Privilege
        /// </summary>
        EoP = 0x01,
        /// <summary>
        /// Denial of Service
        /// </summary>
        DoS = 0x02,
        /// <summary>
        /// Spoofing
        /// </summary>
        Spf = 0x04,
        /// <summary>
        /// Information Disclosure
        /// </summary>
        Inf = 0x08,
        /// <summary>
        /// Repudiation
        /// </summary>
        Rep = 0x10,
        /// <summary>
        /// Tampering
        /// </summary>
        Tam = 0x20,

        /// <summary>
        /// All threats
        /// </summary>
        Stride = (EoP | DoS | Spf | Inf | Rep | Tam)
    }

    /// <summary>
    /// A graph relating definition verbs with use verbs
    /// </summary>
    public class VerbGraph
        : UR.Graphing.DirectedGraph
    {
        /// <summary>
        /// Initializes the verb graph
        /// </summary>
        /// <param name="db">The data context to operate on</param>
        public VerbGraph(TBADataContext db)
        {
            this.Db = db;

            PopulateGraph();
        }

        /// <summary>
        /// Checks to see if two verbs are related
        /// </summary>
        /// <param name="use">A use verb</param>
        /// <param name="definition">A definition verb</param>
        /// <returns></returns>
        public bool IsRelated(
            Verb use,
            Verb definition)
        {
            return HasEdge(
                definition,
                use);
        }

        /// <summary>
        /// Populates the graph by adding edges between each related verb pair
        /// </summary>
        private void PopulateGraph()
        {
            PopulateAlpcPortVerbs();
            PopulateFileVerbs();
            PopulateProcessVerbs();
            PopulateServiceVerbs();
        }

        /// <summary>
        /// Populates the graph's ALPC port verbs
        /// </summary>
        private void PopulateAlpcPortVerbs()
        {
            string objectType = "ALPC Port";
            
            // a_d writes a request that is read by a_u
            AddRelation(
                GetVerb(
                    "Write request",
                    objectType,
                    true,
                    0x1, // CONNECT
                    false,
                    false,
                    ThreatBits.Inf
                    ),
                GetVerb(
                    "Read request",
                    objectType,
                    false,
                    0,
                    false,
                    true,
                    ThreatBits.Stride
                    ));

            // a_d writes a request that is read by a_u
            AddRelation(
                GetVerb(
                    "Write reply",
                    objectType,
                    true,
                    0,
                    false,
                    true,
                    ThreatBits.Inf
                    ),
                GetVerb(
                    "Read reply",
                    objectType,
                    false,
                    0x1, // CONNECT
                    false,
                    false,
                    ThreatBits.Stride
                    ));
        }

        /// <summary>
        /// Populates the graph's File verbs
        /// </summary>
        private void PopulateFileVerbs()
        {
            string objectType = "File";

            // a_d may write data to a file that may read by a_u
            AddRelation(
                GetVerb(
                    "Write data",
                    objectType,
                    true,
                    0x2,
                    false,
                    false,
                    ThreatBits.Inf),
                GetVerb(
                    "Read data",
                    objectType,
                    false,
                    0x1,
                    false,
                    false,
                    ThreatBits.Stride));
        }

        /// <summary>
        /// Populates the graph's Process verbs
        /// </summary>
        private void PopulateProcessVerbs()
        {
            string objectType = "Process";

            // a_d may write memory to a process that is then executed as a_u
            AddRelation(
                GetVerb(
                    "Write process memory",
                    objectType,
                    true,
                    0x20, // VM_WRITE
                    false,
                    false,
                    ThreatBits.Inf),
                GetVerb(
                    "Execute process memory",
                    objectType,
                    false,
                    0,
                    true,
                    false,
                    ThreatBits.Stride));

        }

        /// <summary>
        /// Populates the graph's Service verbs
        /// </summary>
        private void PopulateServiceVerbs()
        {
            string objectType = "Service";

            AddRelation(
                GetVerb(
                    "Change configuration",
                    objectType,
                    true,
                    0x2,
                    false,
                    false,
                    ThreatBits.Nil),
                GetVerb(
                    "Start service",
                    objectType,
                    false,
                    0,
                    true,
                    false,
                    ThreatBits.Stride));
        }

        /// <summary>
        /// Adds an edge relating the definition verb and the use verb
        /// </summary>
        /// <param name="definition">The definition verb</param>
        /// <param name="use">The use verb</param>
        private void AddRelation(
            Verb definition,
            Verb use)
        {
            VerbRelation relation;

            // If the relationship has not been defined in the database, define it now
            if ((from rel in Db.VerbRelations
                 where rel.DefinitionVerbId == definition.Id && rel.UseVerbId == use.Id
                 select rel).Count() == 0)
            {
                relation = new VerbRelation();
                relation.DefinitionVerbId = definition.Id;
                relation.UseVerbId = use.Id;

                Db.VerbRelations.InsertOnSubmit(relation);
            }

            // Persist these changes and add an edge to the graph for each verb
            Db.SubmitChanges();

            AddEdge(
                definition,
                use);
        }

        /// <summary>
        /// Gets a verb instance from the verb's information
        /// </summary>
        /// <param name="name">The verb's name</param>
        /// <param name="objectType">The object type the verb operates on</param>
        /// <param name="isDefinition">True if this is a definition verb</param>
        /// <param name="rightsRequired">Bit vector of rights required to access this verb</param>
        /// <param name="implicitUse">True if this represents an implicit use verb</param>
        /// <param name="implicitDef">True if this represents an implicit definition verb</param>
        /// <param name="threatBits">The threats to the actor using this verb</param>
        /// <returns></returns>
        private Verb GetVerb(
            string name,
            string objectType,
            bool isDefinition,
            int rightsRequired,
            bool implicitUse,
            bool implicitDef,
            ThreatBits threatBits)
        {
            Verb verb;
            var q =
                from v in Db.Verbs
                where
                     v.Name == name &&
                     v.ObjectType == objectType
                select v;

            if (q == null ||
                q.Count() == 0)
            {
                verb = new Verb();

                verb.Name = name;
                verb.ObjectType = objectType;
                verb.IsDefinition = isDefinition;
                verb.MatchImplicitDefinition = implicitDef;
                verb.MatchImplicitUse = implicitUse;
                verb.ThreatBitMask = (int)threatBits;
                verb.RightsRequired = rightsRequired;
                verb.MatchRights = rightsRequired != 0;

                Db.Verbs.InsertOnSubmit(verb);
                Db.SubmitChanges();
            }
            else
                verb = q.First() as Verb;

            return verb;
        }

        /// <summary>
        /// THe data context
        /// </summary>
        private TBADataContext Db { get; set; }
    }
}
