//
// Dynamic trust boundary metadata collection driver
//
// This driver is designed to collect metadata about securable objects such as
// Process, File, Event, and Section objects.  The metadata is used to construct
// a chronological object log which conveys information about object definitions
// and uses.  When an object is defined or used, metadata is logged about the
// environment in which the object is created including the security context,
// call stack, object name information, and the security descriptor assigned to
// the object.
//
// This driver is only designed to run on Windows Vista SP1 x64 due to the fact
// that it requires the new object manager callback APIs and also due to the
// fact that it requires x64-specific facilities to obtain reliable call stack
// information.  It also requires a debugger to be attached at boot time due to
// the fact that it instruments ObCreateObject in order to prevent PatchGuard
// from striking.  Welp, at least this code is only meant to be used for
// research purposes, right?
//
// skape
// mmiller@hick.org
// 3/2008
//
#include "precomp.h"

NTSTATUS 
ObCreateObjectHook( 
	__in KPROCESSOR_MODE ProbeMode, 
	__in POBJECT_TYPE ObjectType, 
	__in POBJECT_ATTRIBUTES ObjectAttributes OPTIONAL, 
	__in KPROCESSOR_MODE OwnershipMode, 
	__inout PVOID ParseContext OPTIONAL, 
	__in ULONG ObjectBodySize, 
	__in ULONG PagedPoolCharge, 
	__in ULONG NonPagedPoolCharge, 
	__out PVOID *Object);

extern NTSTATUS CallOrigObCreateObject( 
	__in KPROCESSOR_MODE ProbeMode, 
	__in POBJECT_TYPE ObjectType, 
	__in POBJECT_ATTRIBUTES ObjectAttributes OPTIONAL, 
	__in KPROCESSOR_MODE OwnershipMode, 
	__inout PVOID ParseContext OPTIONAL, 
	__in ULONG ObjectBodySize, 
	__in ULONG PagedPoolCharge, 
	__in ULONG NonPagedPoolCharge, 
	__out PVOID *Object);

//
// An instrumented function description & state
//
typedef struct _INSTRUMENTATION_INFORMATION
{
	PWCHAR SymbolName;
	PVOID HookRoutine;
	PVOID OrigRoutine;
	PUCHAR Trampoline;
	ULONG TrampolineSize;
	ULONG PreambleSize;
	ULONG_PTR RestorePointer;
} INSTRUMENTATION_INFORMATION, *PINSTRUMENTATION_INFORMATION;

//
// Cache the system process object for use in our load image notify routine
//
PVOID SystemProcessObject = NULL;

//
// Trampoline stub that is used to overwrite the ObCreateObject preamble
//
// mov rax, address
// jmp rax
//
UCHAR 
ObCreateObjectTrampoline[] =
	"\x48\xb8\xaa\xaa\xaa\xaa\xbb\xbb\xbb\xbb\xff\xe0";

INSTRUMENTATION_INFORMATION ObCreateObjectInfo =
{
	L"ObCreateObject",
	ObCreateObjectHook,
	CallOrigObCreateObject,
	ObCreateObjectTrampoline,
	0xc,
	0xd,
	0
};

//
// Hook routine for ObCreateObject that logs metadata about the object being
// defined.  The metadata includes call stack, security context, and security
// descriptor information.  However, we cannot query name information at this
// point as some objects will require the object to have been initialized to a
// certain point before it is safe to query the name.  This is a bummer as it
// means we must wait until a later point to query this information.
//
NTSTATUS 
ObCreateObjectHook( 
	__in KPROCESSOR_MODE ProbeMode, 
	__in POBJECT_TYPE ObjectType, 
	__in POBJECT_ATTRIBUTES ObjectAttributes OPTIONAL, 
	__in KPROCESSOR_MODE OwnershipMode, 
	__inout PVOID ParseContext OPTIONAL, 
	__in ULONG ObjectBodySize, 
	__in ULONG PagedPoolCharge, 
	__in ULONG NonPagedPoolCharge, 
	__out PVOID *OutObject)
{
	NTSTATUS Status;

	//
	// Call the original ObCreateObject routine which will either succeed or
	// fail.  We only care about successful calls.
	//
	Status = CallOrigObCreateObject(
			ProbeMode, 
			ObjectType, 
			ObjectAttributes, 
			OwnershipMode, 
			ParseContext,
			ObjectBodySize,
			PagedPoolCharge, 
			NonPagedPoolCharge, 
			OutObject);

	//
	// If the call succeeds we'll log the definition in terms of the type of
	// object that we defined, the security descriptor that was applied to it,
	// and the code that was responsible for creating it.  We don't pass along
	// the error to the caller if our write fails.
	//
	// We also check to make sure that we are not running in the context of the
	// object log worker thread.  If we are, we don't attempt to log anything.
	// This is done to prevent recursion where objects are created on behalf of
	// the worker thread.
	//
	if ((NT_SUCCESS(Status)) &&
	    (PsGetCurrentThread() != ObjectLogWorkerThread))
	{
		OBJECT_ACCESS_METADATA_FLAGS AccessFlags;
		POBJECT_ACCESS_METADATA AccessMetadata;
		POBJECT_LOG_WORKITEM WorkItem;
		POBJECT_LOG ObjectLog;
		NTSTATUS SubStatus;
		PVOID Object = *OutObject;

		//
		// Lookup the object log at this point so that we can make sure that
		// anything that needs to be hooked for the newly created object's object
		// type will be hooked.
		//
		LookupObjectLog(
				Object,
				&ObjectLog);

		//
		// Allocate and initialize the work item that we will posted our logging
		// effort on to
		//
		WorkItem = AllocateObjectLogWorkItem(
				ObjectLogTaskWrite);

		if (WorkItem)
		{
			//
			// Get the access metadata describing this object's definition.
			//
			AccessFlags.AsUlong = 0;
			AccessFlags.Definition = 1;

			SubStatus = GetObjectAccessMetadata(
					Object,
					AccessFlags,
					0,
					&WorkItem->Write.AccessMetadata);

			if (NT_SUCCESS(SubStatus))
			{
				//
				// Snag a reference to this object so that we can keep it alive in
				// our work item context
				//
				ObReferenceObject(
						Object);

				WorkItem->Write.Object = Object;

				QueueObjectLogWorkItem(
						WorkItem);
			}	
			else
			{
				DebugPrint(("ObCreateObjectHook: GetObjectAccessMetadata failed, %.8x", 
						SubStatus));

				FreeObjectLogWorkItem(
						WorkItem);
			}
		}
	}

	return Status;
}

//
// No-operation
//
OB_PREOP_CALLBACK_STATUS
ObPreCallback(
	__in POBJECT_LOG ObjectLog,
	__in POB_PRE_OPERATION_INFORMATION Information)
{
	return OB_PREOP_SUCCESS;
}

//
// Called after each operation on an object which allows us to log additional
// metadata about a use of an object.  This metadata includes the name of the
// object (since it should be safe to obtain it now) as well as the security
// context and call stack of the caller.
//
VOID
ObPostCallback(
	__in POBJECT_LOG ObjectLog,
	__in POB_POST_OPERATION_INFORMATION Information)
{
	POBJECT_SECURITY_DESCRIPTOR_METADATA SecurityDescriptorMetadata = NULL;
	OBJECT_ACCESS_METADATA_FLAGS AccessFlags;
	POBJECT_ACCESS_METADATA AccessMetadata = NULL;
	POBJECT_NAME_METADATA NameMetadata = NULL;
	POBJECT_LOG_WORKITEM WorkItem;
	POBJECT_HEADER ObjectHeader;
	ACCESS_MASK AccessMask;
	PETHREAD Thread = PsGetCurrentThread();
	NTSTATUS Status = STATUS_SUCCESS;
	BOOLEAN Recurse = FALSE;
	PVOID Object;

	do
	{
		Object = Information->Object;

		//
		// If the operation did not complete successfully, we don't care.
		//
		if (Information->ReturnStatus != STATUS_SUCCESS)
			break;

		//
		// Don't proceed in the slightest if we are running in the object log
		// worker thread
		//
		if (Thread == ObjectLogWorkerThread)
			break;

		//
		// If we are being called recursively, bail.  Otherwise, toggle the
		// recurse flag
		//
		if (CHECK_RECURSE_FLAG(Thread))
			break;

		SET_RECURSE_FLAG(Thread);
		Recurse = TRUE;

		//
		// Get the object's name information
		//
		Status = GetObjectNameMetadata(
				Object,
				&NameMetadata);

		if (!NT_SUCCESS(Status))
		{
			DebugPrint(("GetObjectNameMetadata(%p) failed, %.8x",
					Object, Status));
		}

		//
		// If the handle and pointer count are less than or equal to one, if we
		// assume that this is a definition, let's get the security descriptor
		// that was ultimately assigned to the object as we may not have been able
		// to do so at the time of ObCreateObject
		//
		ObjectHeader = GET_OBJECT_HEADER(Object);

		if ((ObjectHeader->HandleCount <= 1) &&
		    (ObjectHeader->PointerCount <= 1))
		{
			GetObjectSecurityDescriptorMetadata(
					Object,
					TRUE,
					&SecurityDescriptorMetadata);
		}

		//
		// Log the use of this object and the access mask that was granted at the
		// time of the use
		//
		AccessFlags.AsUlong = 0;

		if (Information->Operation == OB_OPERATION_HANDLE_CREATE)
			AccessMask = Information->Parameters->CreateHandleInformation.GrantedAccess;
		else if (Information->Operation == OB_OPERATION_HANDLE_DUPLICATE)
			AccessMask = Information->Parameters->DuplicateHandleInformation.GrantedAccess;

		GetObjectAccessMetadata(
				Object,
				AccessFlags,
				AccessMask,
				&AccessMetadata);
	
		//
		// If we failed to acquire access, name, and sd information, then bail
		//
		if (!AccessMetadata && !NameMetadata && !SecurityDescriptorMetadata)
			break;

		//
		// Allocate & queue our work item to log what we've found
		//
		WorkItem = AllocateObjectLogWorkItem(
				ObjectLogTaskWrite);

		if (WorkItem)
		{
			ObReferenceObject(
					Object);

			WorkItem->Write.Object = Object;
			WorkItem->Write.NameMetadata = NameMetadata;
			WorkItem->Write.AccessMetadata = AccessMetadata;
			WorkItem->Write.SecurityDescriptorMetadata = SecurityDescriptorMetadata;

			QueueObjectLogWorkItem(
					WorkItem);
		}
		else
		{
			if (SecurityDescriptorMetadata)
				ExFreePool(
						SecurityDescriptorMetadata);
			if (AccessMetadata)
				ExFreePool(
						AccessMetadata);
			if (NameMetadata)
				ExFreePool(
						NameMetadata);
		}

	} while (0);

	//
	// Clear the recurse flag if it was set
	//
	if (Recurse)
		CLEAR_RECURSE_FLAG(Thread);
}

//
// Procedure called whenever an object type's SecurityProcedure is invoked.  The
// purpose of this routine is to log an object's security descriptor
// information.
//
NTSTATUS 
ObjectTypeSecurityProcedure(
	__in PVOID Object,
	__in SECURITY_OPERATION Operation,
	__in PSECURITY_INFORMATION Information,
	__in PSECURITY_DESCRIPTOR SecurityDescriptor,
	__inout PULONG ReturnLength,
	__inout PSECURITY_DESCRIPTOR *ObjectSecurityDescriptor,
	__in POOL_TYPE PoolType,
	__in PGENERIC_MAPPING GenericMapping)
{
	POBJECT_LOG ObjectLog;
	NTSTATUS Status;

	//
	// Look up the object log associated with this object.  The object log MUST
	// be valid -- we can't proceed if it is not
	//
	LookupObjectLog(
			Object,
			&ObjectLog);

	ASSERT(ObjectLog != NULL);

	//
	// Invoke the original object type security procedure.  If it succeeds, we'll
	// investigate further.
	//
	if (ObjectLog->SecurityProcedure)
		Status = ObjectLog->SecurityProcedure(
				Object,
				Operation,
				Information,
				SecurityDescriptor,
				ReturnLength,
				ObjectSecurityDescriptor,
				PoolType,
				GenericMapping);
	else
		Status = STATUS_SUCCESS;

	//
	// If the security operation was set or assign & the operation succeeded, we
	// log information about the security descriptor
	//
	if (((Operation == OpSetSecurityDescriptor) ||
	     (Operation == OpAssignSecurityDescriptor)) &&
	    (NT_SUCCESS(Status)))
	{
		POBJECT_SECURITY_DESCRIPTOR_METADATA SecurityDescriptorMetadata = NULL;
		POBJECT_LOG_WORKITEM WorkItem = NULL;
		NTSTATUS SubStatus;

		SubStatus = GetObjectSecurityDescriptorMetadata(
				Object,
				FALSE,
				&SecurityDescriptorMetadata);

		//
		// Allocate and queue the work item that will write the information to
		// the object log for this object
		//
		if (SecurityDescriptorMetadata)
		{
			WorkItem = AllocateObjectLogWorkItem(
					ObjectLogTaskWrite);

			if (WorkItem)
			{
				WorkItem->Write.ObjectLog = ObjectLog;
				WorkItem->Write.SecurityDescriptorMetadata = SecurityDescriptorMetadata;

				QueueObjectLogWorkItem(
						WorkItem);
			}
			else
			{
				ExFreePool(
						SecurityDescriptorMetadata);
			}
		}

	}

	return Status;
}


//
// Queues access and name metadata for an object
//
VOID
QueueObjectAccessAndNameMetadata(
	__in PVOID Object,
	__in BOOLEAN QueueNameMetadata,
	__in BOOLEAN QueueAccessMetadata,
	__in BOOLEAN IsDefinition)
{
	POBJECT_ACCESS_METADATA AccessMetadata = NULL;
	POBJECT_NAME_METADATA NameMetadata = NULL;
	POBJECT_LOG_WORKITEM WorkItem;

	//
	// If requested, get the object's create metadata
	//
	if (QueueAccessMetadata)
	{
		OBJECT_ACCESS_METADATA_FLAGS AccessFlags;

		AccessFlags.AsUlong = 0;
		AccessFlags.Definition = IsDefinition;

		GetObjectAccessMetadata(
				Object,
				AccessFlags,
				0,
				&AccessMetadata);
	}

	//
	// Query the process' image file path name which we will log as metadata
	//
	if (QueueNameMetadata)
		GetObjectNameMetadata(
				Object,
				&NameMetadata);

	if (AccessMetadata || NameMetadata)
	{
		//
		// Allocate a write work item and post it off to a worker thread so that
		// we do not block the calling context
		//
		WorkItem = AllocateObjectLogWorkItem(
				ObjectLogTaskWrite);

		if (WorkItem)
		{
			ObReferenceObject(
					Object);

			WorkItem->Write.Object         = Object;
			WorkItem->Write.NameMetadata   = NameMetadata;
			WorkItem->Write.AccessMetadata = AccessMetadata;

			QueueObjectLogWorkItem(
					WorkItem);
		}
		else
		{
			if (AccessMetadata)
				ExFreePool(
						AccessMetadata);
			if (NameMetadata)
				ExFreePool(
						NameMetadata);
		}
	}
}

//
// Queues image information metadata
//
NTSTATUS
QueueImageInfoMetadata(
	__in PEPROCESS ProcessObject,
	__in PUNICODE_STRING ImageName,
	__in ULONG64 ImageBase,
	__in ULONG ImageSize)
{
	POBJECT_IMAGEINFO_METADATA ImageInfoMetadata = NULL;
	POBJECT_LOG_WORKITEM WorkItem;
	POBJECT_LOG ObjectLog = NULL;
	NTSTATUS Status;
	BOOLEAN Queued = FALSE;
	ULONG ImageInfoMetadataSize;

	//
	// Initialize our image info metadata
	//
	ImageInfoMetadataSize = 
		sizeof(OBJECT_IMAGEINFO_METADATA) + ImageName->Length;

	ImageInfoMetadata = (POBJECT_IMAGEINFO_METADATA)ExAllocatePoolWithTag(
			PagedPool,
			ImageInfoMetadataSize,
			ALLOC_TAG);

	if (!ImageInfoMetadata)
		return STATUS_INSUFFICIENT_RESOURCES;

	InitializeObjectMetadata(
			ImageInfoMetadata,
			OBJECT_METADATA_TAG_IMAGE_INFO,
			ImageInfoMetadataSize,
			ProcessObject);

	ImageInfoMetadata->BaseAddress     = ImageBase;
	ImageInfoMetadata->ImageSize       = ImageSize;
	ImageInfoMetadata->ImageNameLength = ImageName->Length;

	RtlCopyMemory(
			(PUCHAR)ImageInfoMetadata + sizeof(OBJECT_IMAGEINFO_METADATA),
			ImageName->Buffer,
			ImageName->Length);

	//
	// Allocate a work item to write our metadata to the appropriate log
	//
	WorkItem = AllocateObjectLogWorkItem(
			ObjectLogTaskWrite);

	if (WorkItem)
	{
		//
		// Take a reference to the process object
		//
		ObReferenceObject(
				ProcessObject);

		WorkItem->Write.Object            = ProcessObject;
		WorkItem->Write.ImageInfoMetadata = ImageInfoMetadata;

		QueueObjectLogWorkItem(
				WorkItem);

		Status = STATUS_SUCCESS;
	}
	else
	{
		ExFreePool(
				ImageInfoMetadata);

		Status = STATUS_INSUFFICIENT_RESOURCES;
	}

	return Status;
}

//
// Queues information about ntoskrnl's image mapping
//
VOID
QueueNtosImageInfo(
	__in PEPROCESS SystemProcess)
{
	UNICODE_STRING NtosImageName;
	UNICODE_STRING SymbolName;
	PVOID SymbolAddress;
	PVOID ImageBaseAddress = NULL;

	RtlInitUnicodeString(
			&NtosImageName,
			L"ntoskrnl.exe");
	RtlInitUnicodeString(
			&SymbolName,
			L"ObCreateObject");

	//
	// Resolve the location of a public symbol within ntoskrnl so that we can
	// locate its image base address
	//
	SymbolAddress = MmGetSystemRoutineAddress(
			&SymbolName);

	if (!SymbolAddress)
	{
		ASSERT(SymbolAddress != NULL);
		return;
	}

	//
	// Get the base address of the symbol
	//
	if (RtlPcToFileHeader(
			SymbolAddress,
			&ImageBaseAddress) != NULL)
	{
		PIMAGE_NT_HEADERS Nt;

		Nt = RtlImageNtHeader(
				ImageBaseAddress);

		if (Nt)
			QueueImageInfoMetadata(
					SystemProcess,
					&NtosImageName,
					(ULONG64)ImageBaseAddress,
					Nt->OptionalHeader.SizeOfImage);
	}
}

//
// Notify routine that is called when processes are created.  We log the created
// process name to the appropriate object log file.
//
VOID
PsCreateProcessNotifyRoutine(
	__in HANDLE ParentId,
	__in HANDLE ProcessId,
	__in BOOLEAN Create)
{
	POBJECT_LOG ObjectLog = NULL;
	KAPC_STATE ApcState;
	PEPROCESS ProcessObject = NULL;
	NTSTATUS Status;
	BOOLEAN Attached = FALSE;

	do
	{
		//
		// We don't care when processes are terminated
		//
		if (!Create)
			break;

		//
		// Get the process object associated with the process identifier so that we
		// can stack attach to it and query our process name
		//
		Status = PsLookupProcessByProcessId(
				ProcessId,
				&ProcessObject);

		if (!NT_SUCCESS(Status))
		{
			DebugPrint(("PsCreateProcessNotifyRoutine: PsLookupProcessByProcessId(%u) failed, %.8x",
					ProcessId, Status));
			break;
		}

		//
		// Only attach if necessary
		//
		if (PsGetCurrentProcess() != ProcessObject)
		{
			KeStackAttachProcess(
					ProcessObject,
					&ApcState);

			Attached = TRUE;
		}

		//
		// Queue this process object's name metadata.  We don't queue create data
		// since we should have already seen the create happen within our hook.
		//
		QueueObjectAccessAndNameMetadata(
				ProcessObject,
				TRUE,
				FALSE,
				FALSE);

	} while (0);

	if (Attached)
		KeUnstackDetachProcess(
				&ApcState);

	if (ProcessObject)
		ObDereferenceObject(
				ProcessObject);
}

//
// Notify routine that is called when images are loaded.  We log the image base
// address association with a given image file path so that we can associate
// call stacks with them.
//
VOID
PsLoadImageNotifyRoutine(
	__in PUNICODE_STRING FullImageName,
	__in HANDLE ProcessId,
	__in PIMAGE_INFO ImageInfo)
{
	PEPROCESS ProcessObject = NULL;
	NTSTATUS Status;

	do
	{
		//
		// Lookup the process object that the image is being loaded into as this
		// will be our object context for the log
		//
		Status = PsLookupProcessByProcessId(
				ProcessId,
				&ProcessObject);
	
		if (!NT_SUCCESS(Status))
		{
			//
			// If the supplied process identifier is zero and we have a valid
			// system process object, we'll use that
			//
			if ((ProcessId == 0) && (SystemProcessObject))
				ProcessObject = SystemProcessObject;
			else
			{
				DebugPrint(("PsLoadImageNotifyRoutine: PsLookupProcessByProcessId(%u) failed on image %wZ, %.8x",
						ProcessId, FullImageName, Status));
				break;
			}
		}

		//
		// Queue information about this image
		//
		Status = QueueImageInfoMetadata(
				ProcessObject,
				FullImageName,
				(ULONG64)ImageInfo->ImageBase,
				(ULONG)ImageInfo->ImageSize);

	} while (0);

	//
	// Lose our reference that we got from PsLookupProcessByProcessId
	//
	if ((ProcessId != 0) && (ProcessObject))
		ObDereferenceObject(
				ProcessObject);
}

#pragma code_seg("INIT")

//
// Instruments a function by hooking its prologue and redirecting its execution
// to another function
//
NTSTATUS 
InstrumentFunction(
	__in PINSTRUMENTATION_INFORMATION Information,
	__in BOOLEAN Enable)
{
	UNICODE_STRING SymbolNameString;
	ULONG_PTR SymbolAddress;
	NTSTATUS Status = STATUS_SUCCESS;
	PVOID LockedAddress;
	PMDL Mdl = NULL;

	do
	{
		//
		// First, we resolve the absolute address of the symbol and determine
		// whether or not we can safely hook it
		//
		RtlInitUnicodeString(
				&SymbolNameString,
				Information->SymbolName);

		SymbolAddress = (ULONG_PTR)MmGetSystemRoutineAddress(
				&SymbolNameString);

		//
		// If we couldn't resolve the address of the function, we can't proceed
		// with our hook
		//
		if (SymbolAddress == 0)
		{
			DebugPrint(("InstrumentFunction: %wZ could not be found",
					&SymbolNameString));

			Status = STATUS_UNSUCCESSFUL;
			break;
		}
		// 
		// If things did match, offset the real address by the size of the
		// overwritten preamble to account for the copied instructions that we
		// will execute beforehand
		//
		else
			Information->RestorePointer = 
				SymbolAddress + Information->PreambleSize;

		KeMemoryBarrier();

		//
		// Now, prepare the trampoline and copy it over the preamble of
		// the function to be hooked
		//
		Mdl = IoAllocateMdl(
				(PVOID)SymbolAddress,
				Information->TrampolineSize,
				FALSE,
				FALSE,
				NULL);

		if (!Mdl)
		{
			Status = STATUS_INSUFFICIENT_RESOURCES;
			break;
		}

		MmBuildMdlForNonPagedPool(
				Mdl);

		__try
		{
			//
			// Probe and lock
			//
			MmProbeAndLockPages(
					Mdl,
					KernelMode,
					IoModifyAccess);

			//
			// Map the page so that we can modify it safely
			//
			LockedAddress = MmMapLockedPagesSpecifyCache(
					Mdl,
					KernelMode,
					MmWriteCombined,
					NULL,
					FALSE,
					NormalPagePriority);

			if (!LockedAddress)
			{
				Status = STATUS_INSUFFICIENT_RESOURCES;
				break;
			}

		} __except(EXCEPTION_EXECUTE_HANDLER)
		{
			break;
		}

		//
		// If we are enabling instrumentation, overwrite the preamble with our
		// trampoline code
		//
		if (Enable)
		{
			//
			// Finalize the trampoline and copy it over the preamble of the routine
			// to be hooked
			//
			*(PULONG_PTR)(Information->Trampoline + 2) = 
				(ULONG_PTR)Information->HookRoutine;

			RtlCopyMemory(
					LockedAddress,
					Information->Trampoline,
					Information->TrampolineSize);
		}
		//
		// If we aren't enabling, restore the preamble
		//
		else
			RtlCopyMemory(
					LockedAddress,
					Information->OrigRoutine,
					Information->TrampolineSize);

	} while (0);

	//
	// If we ended up using the MDL, then free it now
	//
	if (Mdl)
	{
		if (LockedAddress)
			MmUnmapLockedPages(
					LockedAddress,
					Mdl);

		MmUnlockPages(
				Mdl);

		IoFreeMdl(Mdl);
	}

	return Status;
}


//
// Initializes the object log subsystem, registers process create and load image
// notify routines, and begins instrumenting functions
//
NTSTATUS 
DriverEntry(
	__in PDRIVER_OBJECT DriverObject,
	__in PUNICODE_STRING RegistryPath)
{
	UNICODE_STRING EventName;
	NTSTATUS Status;

	do
	{
		//
		// If a debugger is not attached then we fail as without the debugger
		// PatchGuard will catch our ntoskrnl modifications
		//
		if (KD_DEBUGGER_NOT_PRESENT == TRUE)
		{
			DebugPrint(("DriverEntry: A debugger MUST be attached for dynba to function due to PatchGuard limitations."));

			Status = STATUS_NOT_SUPPORTED;
			break;
		}

		//
		// Cache the system process object
		//
		SystemProcessObject = PsGetCurrentProcess();

		//
		// Initialize the object log subsystem
		//
		Status = InitializeObjectLog();

		if (!NT_SUCCESS(Status))
		{
			DebugPrint(("DriverEntry: InitializeObjectLog failed, %.8x",
					Status));
			break;
		}

		//
		// Queue a symbolic metadata entry for the creation of the system process
		// since our driver will not have seen it
		//
		QueueObjectAccessAndNameMetadata(
				SystemProcessObject,
				TRUE,
				TRUE,
				TRUE);

		//
		// Queue image information for ntoskrnl so that we can resolve kernel
		// stacks that contain the mapping.  We must do this now because ntos will
		// always be mapped before our driver is loaded and thus we will never get
		// to describe its image information.
		//
		QueueNtosImageInfo(
				SystemProcessObject);

		//
		// Register our create process & load image notify routines so that we can
		// log metadata about processes that are created & images that are loaded
		// in the context of processes
		//
		Status = PsSetCreateProcessNotifyRoutine(
				PsCreateProcessNotifyRoutine,
				FALSE);

		if (!NT_SUCCESS(Status))
		{
			DebugPrint(("DriverEntry: PsSetCreateProcessNotifyRoutine failed, %.8x",
					Status));
			break;
		}

		Status = PsSetLoadImageNotifyRoutine(
				PsLoadImageNotifyRoutine);

		if (!NT_SUCCESS(Status))
		{
			DebugPrint(("DriverEntry: PsSetLoadImageNotifyRoutine failed, %.8x",
					Status));
			break;
		}

		//
		// Instrument ObCreateObject now that we have our state ready to begin
		// handling events
		//
		Status = InstrumentFunction(
				&ObCreateObjectInfo,
				TRUE);

		if (!NT_SUCCESS(Status))
		{
			DebugPrint(("DriverEntry: InstrumentFunction(ObCreateObject) failed, %.8x", 
					Status));
			break;
		}

	} while (0);

	return Status;
}

#pragma code_seg()
