//
// Dynamic trust boundary metadata collection driver
//
// mmiller@hick.org 
// 3/2008
//
#include "precomp.h"

//
// The handle of the thread responsible for processing the object log queue
//
HANDLE ObjectLogWorkerThreadHandle = NULL;

//
// Lookaside list to use when allocating work items
//
PNPAGED_LOOKASIDE_LIST ObjectLogTaskWorkItems;

//
// Generic object log table that is used to store object log contexts
//
RTL_GENERIC_TABLE ObjectLogTable;

//
// Mutex to synchronize the object log table
//
FAST_MUTEX ObjectLogTableMutex;

//
// A handle to the directory where object type logs are created
//
HANDLE ObjectLogDirectory = NULL;

//
// Mutex that guards the work item queue
//
KSPIN_LOCK ObjectLogQueueSpinLock;

//
// Work item queue for work items that are queued of the
// object log directory
//
LIST_ENTRY ObjectLogQueue;

//
// The pool that object log sequence numbers come from
//
ULONG64 ObjectLogSequencePool = 1;

//
// Notification event that is used to indicate that there are entries in the
// queue
//
KEVENT ObjectLogQueueEvent;

//
// The thread responsible for processing the object log queue
//
PETHREAD ObjectLogWorkerThread = NULL;

extern
OB_PREOP_CALLBACK_STATUS
ObPreCallback(
	__in POBJECT_LOG ObjectLog,
	__in POB_PRE_OPERATION_INFORMATION Information);
extern
VOID
ObPostCallback(
	__in POBJECT_LOG ObjectLog,
	__in POB_POST_OPERATION_INFORMATION Information);

extern 
NTSTATUS 
ObjectTypeSecurityProcedure(
	__in PVOID Object,
	__in SECURITY_OPERATION Operation,
	__in PSECURITY_INFORMATION Information,
	__in PSECURITY_DESCRIPTOR SecurityDescriptor,
	__inout PULONG CapturedLength,
	__inout PSECURITY_DESCRIPTOR *ObjectSecurityDescriptor,
	__in POOL_TYPE PoolType,
	__in PGENERIC_MAPPING GenericMapping);

//
// Gets the object log associated with the provided object type.  Each object
// type has its own distinct logging context.
//
NTSTATUS 
LookupObjectLog(
	__in PVOID Object,
	__out POBJECT_LOG *OutObjectLog)
{
	volatile POBJECT_LOG *TargetNode;
	POBJECT_HEADER ObjectHeader;
	POBJECT_LOG ObjectLog;
	OBJECT_LOG PrototypeObjectLog = { 0 };
	ULONG_PTR ObjectTypeNameKey;
	NTSTATUS Status = STATUS_SUCCESS;
	BOOLEAN NewElement;

	//
	// Get the object type's name key using the object type in the object's
	// header
	//
	ObjectHeader = GET_OBJECT_HEADER(Object);

	//
	// Populate the prototype object log that we will use to query the generic
	// table with
	//
	PrototypeObjectLog.Key        = (ULONG_PTR)ObjectHeader->Type->Name.Buffer;
	PrototypeObjectLog.ObjectType = ObjectHeader->Type;

	//
	// Search the object log table for the context associated with this type.  If
	// we don't find one, create a new context and return that to the caller.
	//
	ExAcquireFastMutex(
			&ObjectLogTableMutex);

	ObjectLog = (POBJECT_LOG)RtlInsertElementGenericTable(
			&ObjectLogTable,
			&PrototypeObjectLog,
			sizeof(OBJECT_LOG),
			&NewElement);

	//
	// If we inserted a new element into the table, queue a work item to
	// initialize this object log entry.  We do this while we hold the object log
	// table mutex to make sure that our work item is queued before any other
	// potential work items that would be tied to this object log.
	//
	if (NewElement)
	{
		POBJECT_LOG_WORKITEM WorkItem;

		WorkItem = AllocateObjectLogWorkItem(
				ObjectLogTaskInit);

		if (WorkItem)
		{
			WorkItem->Init.ObjectLog = ObjectLog;

			QueueObjectLogWorkItem(
					WorkItem);
		}
		else
		{
			DebugPrint(("LookupObjectLog: AllocateObjectLogWorkItem(Init) failed, cannot initialize object log."));
			ASSERT(0);
		}
	}

	ExReleaseFastMutex(
			&ObjectLogTableMutex);

	//
	// Provide the caller with the object log pointer regardless of whether or
	// not we succeeded
	//
	*OutObjectLog = ObjectLog;

	return Status;
}

//
// Generic compare routine for the object log table
//
RTL_GENERIC_COMPARE_RESULTS 
ObjectLogCompareRoutine(
	__in PRTL_GENERIC_TABLE Table,
	__in POBJECT_LOG FirstStruct,
	__in POBJECT_LOG SecondStruct)
{
	if (FirstStruct->Key < SecondStruct->Key)
		return GenericLessThan;
	else if (FirstStruct->Key > SecondStruct->Key)
		return GenericGreaterThan;
	else
		return GenericEqual;
}

//
// Generic allocation routine for the object log table
//
PVOID
ObjectLogAllocateRoutine(
	__in PRTL_GENERIC_TABLE Table,
	__in CLONG ByteSize)
{
	return ExAllocatePoolWithTag(
			NonPagedPool,
			ByteSize,
			ALLOC_TAG);
}

//
// Generic free routine for the object log table
//
VOID
ObjectLogFreeRoutine(
	__in PRTL_GENERIC_TABLE Table,
	__in PVOID Buffer)
{
	ExFreePool(
			Buffer);
}

//
// Writes data to an object log
//
NTSTATUS
WriteObjectLog(
	__in POBJECT_LOG ObjectLog,
	__in POBJECT_METADATA_HEADER Metadata)
{
	IO_STATUS_BLOCK IoStatus;

	return ZwWriteFile(
			ObjectLog->FileHandle,
			NULL,
			NULL,
			NULL,
			&IoStatus,
			Metadata,
			Metadata->Length,
			NULL,
			NULL);
}

//
// Allocates an object log work item for a given task 
//
POBJECT_LOG_WORKITEM 
AllocateObjectLogWorkItem(
	__in OBJECT_LOG_TASK Task)
{
	POBJECT_LOG_WORKITEM WorkItem;

	WorkItem = (POBJECT_LOG_WORKITEM)ExAllocateFromNPagedLookasideList(
			ObjectLogTaskWorkItems);

	if (WorkItem)
	{
		RtlZeroMemory(
				WorkItem,
				sizeof(OBJECT_LOG_WORKITEM));

		WorkItem->Task = Task;
	}

	return WorkItem;
}

//
// Queues an object log work item to be processed in a worker thread
//
VOID
QueueObjectLogWorkItem(
	__in POBJECT_LOG_WORKITEM WorkItem)
{
	ExInterlockedInsertTailList(
			&ObjectLogQueue,
			&WorkItem->QueueEntry,
			&ObjectLogQueueSpinLock);

	//
	// Signal the event to indicate that we have a work item ready to process
	//
	KeSetEvent(
			&ObjectLogQueueEvent,
			0,
			FALSE);
}

//
// Frees a work item
//
VOID
FreeObjectLogWorkItem(
	__in POBJECT_LOG_WORKITEM WorkItem)
{
	if (WorkItem->Task == ObjectLogTaskWrite)
	{
		if (WorkItem->Write.Object)
			ObDereferenceObject(
					WorkItem->Write.Object);

		if (WorkItem->Write.ImageInfoMetadata)
			ExFreePool(
					WorkItem->Write.ImageInfoMetadata);
		if (WorkItem->Write.CallStackMetadata)
			ExFreePool(
					WorkItem->Write.CallStackMetadata);
		if (WorkItem->Write.NameMetadata)
			ExFreePool(
					WorkItem->Write.NameMetadata);
		if (WorkItem->Write.SecurityDescriptorMetadata)
			ExFreePool(
					WorkItem->Write.SecurityDescriptorMetadata);
		if (WorkItem->Write.SecurityContextMetadata)
			ExFreePool(
					WorkItem->Write.SecurityContextMetadata);
		if (WorkItem->Write.AccessMetadata)
			ExFreePool(
					WorkItem->Write.AccessMetadata);
	}
	else if (WorkItem->Task == ObjectLogTaskQueryName)
	{
		if (WorkItem->QueryName.Object)
			ObDereferenceObject(
					WorkItem->QueryName.Object);
	}

	ExFreeToNPagedLookasideList(
			ObjectLogTaskWorkItems,
			WorkItem);
}

//
// Increments the object log sequence pool and returns the next unique sequence
// number
//
ULONG64
GetNextObjectLogSequence()
{
	return InterlockedIncrement64(
			&ObjectLogSequencePool);
}

//
// Create/open the directory where we will track object type logs
//
NTSTATUS
OpenObjectLogDirectory()
{
	OBJECT_ATTRIBUTES Attributes;
	IO_STATUS_BLOCK IoStatus;
	UNICODE_STRING LogDirectoryName;

	RtlInitUnicodeString(
			&LogDirectoryName,
			L"\\SystemRoot\\dynba");

	InitializeObjectAttributes(
			&Attributes,
			&LogDirectoryName,
			OBJ_OPENIF | OBJ_CASE_INSENSITIVE | OBJ_KERNEL_HANDLE,
			NULL,
			NULL);

	return ZwCreateFile(
			&ObjectLogDirectory,
			GENERIC_WRITE,
			&Attributes,
			&IoStatus,
			NULL,
			FILE_ATTRIBUTE_NORMAL,
			FILE_SHARE_READ|FILE_SHARE_WRITE,
			FILE_OPEN_IF,
			FILE_DIRECTORY_FILE,
			NULL,
			0);
}

//
// Opens the file associated with an object log
//
NTSTATUS
OpenObjectLogFile(
	__in POBJECT_LOG ObjectLog)
{
	OBJECT_ATTRIBUTES Attributes;
	IO_STATUS_BLOCK IoStatus;
	NTSTATUS Status;

	ASSERT(ObjectLog->FileHandle == NULL);

	//
	// Open the instrumentation log file for this object type
	//
	InitializeObjectAttributes(
			&Attributes,
			&ObjectLog->ObjectType->Name,
			OBJ_CASE_INSENSITIVE | OBJ_KERNEL_HANDLE,
			ObjectLogDirectory,
			NULL);

	Status = ZwCreateFile(
			&ObjectLog->FileHandle,
			GENERIC_WRITE,
			&Attributes,
			&IoStatus,
			NULL,
			FILE_ATTRIBUTE_NORMAL,
			FILE_SHARE_READ|FILE_SHARE_WRITE,
			FILE_OPEN_IF,
			FILE_SEQUENTIAL_ONLY|FILE_SYNCHRONOUS_IO_NONALERT,
			NULL,
			0);

#if DBG
	//
	// If we successfully open the log file, complete our initialization
	//
	if (!NT_SUCCESS(Status))
	{
		DebugPrint(("OpenObjectLogFile: ZwCreateFile('%wZ') failed -- will not be able to log for this object type, %.8x",
				&ObjectLog->ObjectType->Name,
				Status));
	}
#endif

	return Status;
}

//
// Processes an object log initialization work item
//
VOID 
ProcessObjectLogInit(
	__in POBJECT_LOG ObjectLog)
{
	OB_OPERATION_REGISTRATION OpRegistration;
	OB_CALLBACK_REGISTRATION CbRegistration;
	NTSTATUS Status;

	//
	// Hook the object type's security procedure so that we can be notified
	// when a security descriptor is assigned to the object
	//
	ObjectLog->SecurityProcedure = 
		ObjectLog->ObjectType->TypeInfo.SecurityProcedure;

	ObjectLog->ObjectType->TypeInfo.SecurityProcedure =
		ObjectTypeSecurityProcedure;

	//
	// Force this object type to support callbacks even if it normally does
	// not
	//
	ObjectLog->ObjectType->TypeInfo.SupportsObjectCallbacks = TRUE;

	//
	// Register our post callback with the object manager and flag this object
	// type as supporting object filtering as most object types will not by
	// default.
	//
	CbRegistration.Version                    = OB_FLT_REGISTRATION_VERSION;
	CbRegistration.RegistrationContext        = ObjectLog;
	CbRegistration.OperationRegistrationCount = 1;
	CbRegistration.OperationRegistration      = &OpRegistration;

	OpRegistration.ObjectType    = &ObjectLog->ObjectType;
	OpRegistration.Operations    = OB_OPERATION_HANDLE_CREATE;
	OpRegistration.PreOperation  = ObPreCallback;
	OpRegistration.PostOperation = ObPostCallback;

	RtlInitUnicodeString(
		&CbRegistration.Altitude,
		L"40100");

	Status = ObRegisterCallbacks(
			&CbRegistration,
			&ObjectLog->ObRegistrationHandle);

	if (!NT_SUCCESS(Status))
	{
		DebugPrint(("LookupObjectLog: ObRegisterCallbacks for %wZ failed, %.8x.",
				&ObjectLog->ObjectType->Name,
				Status));
		ASSERT(0);
	}
}

//
// Processes an object log write work item
//
VOID
ProcessObjectLogWrite(
	__in POBJECT_LOG_WORKITEM WorkItem)
{
	IO_STATUS_BLOCK IoStatus;
	POBJECT_LOG ObjectLog;
	NTSTATUS Status;

	if (WorkItem->Write.ObjectLog)
		ObjectLog = WorkItem->Write.ObjectLog;
	//
	// If we weren't passed an object log, try to find it.  If the object
	// log has not yet been opened it will be immediately opened.  This is
	// safe because the queue is not processed until after the log
	// directory is opened.
	//
	else
		Status = LookupObjectLog(
				WorkItem->Write.Object,
				&ObjectLog);

	//
	// No object log?  That's bad news.
	//
	if (!ObjectLog)
	{
		DebugPrint(("ObjectLogWorker: LookupObjectLog failed, %.8x",
				Status));
		return;
	}
	// 
	// If the object log has not yet had its log file opened, do that now
	//
	else if (!ObjectLog->FileHandle)
	{
		Status = OpenObjectLogFile(
				ObjectLog);

		if (!NT_SUCCESS(Status))
			return;
	}
	
	//
	// If we successfully open the object log, write whatever metadata we
	// were provided to it
	//
	if (WorkItem->Write.AccessMetadata)
		WriteObjectLog(
				ObjectLog,
				&WorkItem->Write.AccessMetadata->Header);

	if (WorkItem->Write.SecurityContextMetadata)
		WriteObjectLog(
				ObjectLog,
				&WorkItem->Write.SecurityContextMetadata->Header);

	if (WorkItem->Write.SecurityDescriptorMetadata)
		WriteObjectLog(
				ObjectLog,
				&WorkItem->Write.SecurityDescriptorMetadata->Header);

	if (WorkItem->Write.NameMetadata)
		WriteObjectLog(
				ObjectLog,
				&WorkItem->Write.NameMetadata->Header);

	if (WorkItem->Write.CallStackMetadata)
		WriteObjectLog(
				ObjectLog,
				&WorkItem->Write.CallStackMetadata->Header);

	if (WorkItem->Write.ImageInfoMetadata)
		WriteObjectLog(
				ObjectLog,
				&WorkItem->Write.ImageInfoMetadata->Header);
}

//
// Processes an object log query name work item.  Returns TRUE if the work item
// should be re-queued.
//
VOID
ProcessObjectLogQueryName(
	__in POBJECT_LOG_WORKITEM WorkItem)
{
	POBJECT_NAME_METADATA NameMetadata = NULL;
	NTSTATUS Status;

	//
	// Query the object's name metadata
	//
	Status = GetObjectNameMetadata(
			WorkItem->QueryName.Object,
			&NameMetadata);

	if (NT_SUCCESS(Status))
	{
		PVOID TempObject;

		//
		// Close your eyes, this is a bit hacky.  We repurpose our current work
		// item to a write task and call the process write work item directly.
		//
		WorkItem->Task = ObjectLogTaskWrite;

		//
		// This shouldn't be necessary, but we're all about being correct.  We
		// just carry our reference directly over to the write work item.
		//
		TempObject = WorkItem->QueryName.Object;
		WorkItem->Write.Object = TempObject;
		WorkItem->Write.ObjectLog = NULL;
		WorkItem->Write.NameMetadata = NameMetadata;

#if 0
		{
			UNICODE_STRING Dbg;

			Dbg.Length = NameMetadata->NameLength;
			Dbg.MaximumLength = NameMetadata->NameLength;
			Dbg.Buffer = (PWCHAR)((PUCHAR)NameMetadata + sizeof(OBJECT_METADATA_NAME));

			DebugPrint(("Writing name metadata %wZ",
					&Dbg));
		}
#endif

		//
		// Processes the write work item
		//
		ProcessObjectLogWrite(
				WorkItem);
	}
}

//
// Processes an individual object log work item
//
VOID
ProcessObjectLogWorkItem(
	__in POBJECT_LOG_WORKITEM WorkItem)
{
	NTSTATUS Status;

	//
	// Process the work item based on the task requested
	//
	switch (WorkItem->Task)
	{
		//
		// Initializes an object log so that it can be used for subsequent
		// operations
		//
		case ObjectLogTaskInit:
			ProcessObjectLogInit(
					WorkItem->Init.ObjectLog);
			break;

		//
		// Write metadata to an object log
		//
		case ObjectLogTaskWrite:
			ProcessObjectLogWrite(
					WorkItem);
			break;


		//
		// Queries the name of a given object and writes it to disk
		//
		case ObjectLogTaskQueryName:
			ProcessObjectLogQueryName(
					WorkItem);
			break;

		default:
			ASSERT(0);
			break;
	}

	//
	// Once the work item has been processed, we free it
	//
	FreeObjectLogWorkItem(
			WorkItem);
}

//
// Writes object metadata to an object log in the context of a system worker
// thread
//
VOID
ObjectLogWorkerThreadRoutine(
	__in PVOID Unused)
{
	PLIST_ENTRY QueueEntry;
	NTSTATUS Status;

	UNREFERENCED_PARAMETER(Unused);

	//
	// Set the global so that we know this thread will be processing the queue
	// and we can ignore any events that occur from within it
	//
	ObjectLogWorkerThread = PsGetCurrentThread();

	while (TRUE)
	{
		//
		// Wait for the queue event to become signaled
		//
		Status = KeWaitForSingleObject(
				&ObjectLogQueueEvent,
				Executive,
				KernelMode,
				FALSE,
				NULL);

		if (!NT_SUCCESS(Status))
		{
			DebugPrint(("ObjectLogWorker: KeWaitForSingleObject failed, %.8x",
					Status));
			break;
		}

		//
		// If we've yet to open the object log directory, attempt to do so now.
		// We cannot actually process the queue until we are able to open the log
		// directory as we won't be able to log metadata.
		//
		if (ObjectLogDirectory == NULL)
		{
			Status = OpenObjectLogDirectory();

			//
			// Begin waiting on the event again if we fail
			//
			if (!NT_SUCCESS(Status))
				continue;
		}

		//
		// Reset the event prior to processing the queue.  This may cause an
		// extraneous call to KeWaitForSingleObject but it nevertheless prevents
		// us from missing a notification about an entry being inserted into the
		// queue
		//
		KeResetEvent(
				&ObjectLogQueueEvent);

		//
		// Flush the object log queue by processing each of the queued work items
		//
		while ((QueueEntry = ExInterlockedRemoveHeadList(
				&ObjectLogQueue,
				&ObjectLogQueueSpinLock)))
		{
			POBJECT_LOG_WORKITEM WorkItem =
				CONTAINING_RECORD(QueueEntry, OBJECT_LOG_WORKITEM, QueueEntry);

			ProcessObjectLogWorkItem(
					WorkItem);
		}
	}
}


//
// Initializes resources needed by the object log subsystem
//
NTSTATUS
InitializeObjectLog()
{
	//
	// Initialize the paged lookaside list that we'll use to allocate task
	// work items.  The lookaside list must be backed against the non-paged
	// pool.
	//
	ObjectLogTaskWorkItems = (PNPAGED_LOOKASIDE_LIST)ExAllocatePoolWithTag(
			NonPagedPool,
			sizeof(NPAGED_LOOKASIDE_LIST),
			ALLOC_TAG);

	if (!ObjectLogTaskWorkItems)
	{
		DebugPrint(("InitializeObjectLog: Failed to allocate object log task work item lookaside list"));
		
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	ExInitializeNPagedLookasideList(
			ObjectLogTaskWorkItems,
			NULL,
			NULL,
			0,
			sizeof(OBJECT_LOG_WORKITEM),
			ALLOC_TAG,
			0);

	//
	// Initialize the spin lock & list that will be used to temporarily queue work
	// items that are processed before we are able to open the log directory.
	//
	KeInitializeSpinLock(
			&ObjectLogQueueSpinLock);

	InitializeListHead(
			&ObjectLogQueue);

	KeInitializeEvent(
			&ObjectLogQueueEvent,
			NotificationEvent,
			FALSE);

	//
	// Initialize the generic table that we will use to contain object log
	// contexts
	//
	RtlInitializeGenericTable(
			&ObjectLogTable,
			ObjectLogCompareRoutine,
			ObjectLogAllocateRoutine,
			ObjectLogFreeRoutine,
			NULL);

	ExInitializeFastMutex(
			&ObjectLogTableMutex);

	//
	// Finally, create thread responsible for processing the object log queue
	//
	return PsCreateSystemThread(
			&ObjectLogWorkerThreadHandle,
			0,
			NULL,
			NULL,
			NULL,
			ObjectLogWorkerThreadRoutine,
			NULL);
}
