//
// Dynamic trust boundary metadata collection driver
//
// mmiller@hick.org 
// 3/2008
//
#include "precomp.h"

//
// Implementation of RtlLookupFunctionEntry that walks the loaded module list
// stored in the PEB of the invoking process and then walks the exception
// directory to extract the corresponding runtime function entry for the
// supplied RIP value.  This function is inherently unsafe and should never be
// used in non-research software.  You have been warned.
//
// The performance of this function can be improved.  The exception directory is
// a sorted array and thus we can perform a simple binary search rather than a
// linear search.
//
PRUNTIME_FUNCTION 
RtlLookupFunctionEntryUserMode(
	__in ULONG64 Rip,
	__out PULONG64 ImageBase,
	__out PVOID UnsupportedHistoryTable)
{
	PROCESS_BASIC_INFORMATION BasicInformation;
	PRUNTIME_FUNCTION RuntimeFunction = NULL;
	PLIST_ENTRY ListEntry;
	NTSTATUS Status;
	PPEB Peb;

	__try
	{
		//
		// Query basic information for the calling process so that we can obtain
		// the base address of the PEB
		//
		Status = ZwQueryInformationProcess(
				NtCurrentProcess(),
				ProcessBasicInformation,
				&BasicInformation,
				sizeof(PROCESS_BASIC_INFORMATION),
				NULL);

		if (!NT_SUCCESS(Status))
		{
			DebugPrint(("RtlLookupFunctionEntryUserMode: ZwQueryInformationProcess failed, %.8x",
					Status));
			__leave;
		}

		//
		// Grab the PEB base address and verify that the LDR data is sane enough
		// for us to consider at this point
		//
		Peb = BasicInformation.PebBaseAddress;

		if ((!Peb) ||
		    (Peb->Ldr == NULL) ||
			 (!Peb->Ldr->Initialized))
		{
			__leave;
		}

		//
		// Walk the in-memory order list to find the module that contains the
		// supplied RIP.  This is dangerous -- we don't hold the ldr lock at this
		// point so the LDR could tear things out from under us.
		//
		for (ListEntry = Peb->Ldr->InMemoryOrderModuleList.Flink;
		     ListEntry != &Peb->Ldr->InMemoryOrderModuleList;
		     ListEntry = ListEntry->Flink)
		{
			PLDR_DATA_TABLE_ENTRY LdrEntry;
			PIMAGE_NT_HEADERS NtHeader;
			PRUNTIME_FUNCTION Current;
			ULONG ExceptionSize;
			ULONG Rva;
			
			LdrEntry = CONTAINING_RECORD(
					ListEntry, LDR_DATA_TABLE_ENTRY, InMemoryOrderLinks);

			//
			// If RIP is out of range of this module, proceed to the next entry
			//
			if ((Rip < LdrEntry->DllBase) ||
			    (Rip >= LdrEntry->DllBase + LdrEntry->SizeOfImage))
				continue;

			Rva = (ULONG)(Rip - LdrEntry->DllBase);

			//
			// We've found our module. Get the exception directory from it.
			//
			NtHeader = RtlImageNtHeader(
					(PVOID)LdrEntry->DllBase);

			if (!NtHeader)
				continue;

			Current = (PRUNTIME_FUNCTION)RtlImageDirectoryEntryToData(
					(PVOID)LdrEntry->DllBase,
					TRUE,
					IMAGE_DIRECTORY_ENTRY_EXCEPTION,
					&ExceptionSize);

			if (!Current)
				continue;

			//
			// Walk the exception directory's array of runtime functions until we
			// reach the end or find our matching runtime function entry.
			//
			ExceptionSize /= sizeof(RUNTIME_FUNCTION);

			while (ExceptionSize--)
			{
				//
				// Is this the entry we're searching for?
				//
				if ((Rva >= Current->BeginAddress) &&
				    (Rva < Current->EndAddress))
				{
					RuntimeFunction = Current;

					//
					// If this is an implicitly chained entry, walk until we find the
					// root of the chain.  Apparently RtlVirtualUnwind in kernel-mode
					// does not support this, or it assumes that
					// RtlLookupFunctionEntry will do it for it.
					//
					while (RuntimeFunction->UnwindData & 1)
					{
						RuntimeFunction = 
							(PRUNTIME_FUNCTION)(LdrEntry->DllBase + (RuntimeFunction->UnwindData - 1));
					}

					if (ImageBase)
						*ImageBase = LdrEntry->DllBase;

					break;
				}

				Current++;
			}
		}

	} __except(EXCEPTION_EXECUTE_HANDLER)
	{
	}

	return RuntimeFunction;
}

//
// Collects the current call stack information to the object log as being
// associated with a provided object.  This routine can be called at dispatch in
// the path of the object type OpenProcedure.
//
NTSTATUS
GetObjectCallStackMetadata(
	__in PVOID Object,
	__out POBJECT_CALLSTACK_METADATA *OutCallStackMetadata)
{
	POBJECT_CALLSTACK_METADATA CallStackMetadata;
	PRUNTIME_FUNCTION RuntimeFunction;
	NTSTATUS Status = STATUS_SUCCESS;
	CONTEXT UnwindContext;
	BOOLEAN IsUserModeRip;
	ULONG64 ImageBase;
	ULONG64 EstablisherFrame;
	ULONG FrameCount = 0;
	PVOID HandlerData;

	//
	// Allocate storage for our metadata.  We must allocate from non-paged pool
	// because this routine can, unfortunately, be called at dispatch.
	//
	CallStackMetadata = (POBJECT_CALLSTACK_METADATA)ExAllocatePoolWithTag(
			NonPagedPool,
			sizeof(OBJECT_CALLSTACK_METADATA) + 
				(sizeof(OBJECT_CALLSTACK_METADATA_FRAME) * MAX_FRAMES),
			ALLOC_TAG);

	if (!CallStackMetadata)
		return STATUS_INSUFFICIENT_RESOURCES;

	//
	// Initialize the call stack metadata structure -- we wait to write it until
	// we've successfully unwound our first frame
	//
	InitializeObjectMetadata(
			CallStackMetadata,
			OBJECT_METADATA_TAG_CALL_STACK,
			sizeof(OBJECT_CALLSTACK_METADATA),
			Object);

	CallStackMetadata->Flags = 0;

	//
	// If we're running at dispatch, we cannot proceed.  We return status success
	// even though we didn't actually pull down a call stack.
	//
	if (KeGetCurrentIrql() > PASSIVE_LEVEL)
	{
		CallStackMetadata->Flags |= OBJECT_CALLSTACK_METADATA_FLAG_AT_DIRQL;

		return STATUS_SUCCESS;
	}

	//
	// Capture our current context which we'll use to unwind and keep looping
	// until we encounter RIP=0 or an error occurs
	//
	RtlCaptureContext(
			&UnwindContext);

	do
	{
		__try
		{
			//
			// Get the runtime function entry associated with the instruction pointer
			// at the current unwind context.  If RIP is below the system range start,
			// we use a custom RtlLookupFunctionEntry implementation that is able to
			// manually walk the PEB for the calling process to find an image that
			// matches
			//
			IsUserModeRip = UnwindContext.Rip < (ULONG64)MM_SYSTEM_RANGE_START;

			if (IsUserModeRip)
				RuntimeFunction = RtlLookupFunctionEntryUserMode(
						UnwindContext.Rip,
						&ImageBase,
						NULL);
			else
				RuntimeFunction = RtlLookupFunctionEntry(
						UnwindContext.Rip,
						&ImageBase,
						NULL);
			
			//
			// If we successfully find a runtime function entry, perform a virtual
			// unwind
			//
			if (RuntimeFunction)
			{
				RtlVirtualUnwind(
						UNW_FLAG_NHANDLER,
						ImageBase,
						UnwindContext.Rip,
						RuntimeFunction,
						&UnwindContext,
						&HandlerData,
						&EstablisherFrame,
						0);
			}
			//
			// If we didn't find a runtime function entry, we may have encountered a
			// leaf function.  For example, when we traverse into user-mode, the
			// system call routines are leaf functions.  In these cases, we make the
			// bold assumption that the return address is at the current RSP.
			//
			else if (
				(!IoIs32bitProcess(NULL)) && 
				(IsUserModeRip) &&
				(UnwindContext.Rsp < (ULONG64)MM_SYSTEM_RANGE_START))
			{
				UnwindContext.Rip  = *(PULONG64)UnwindContext.Rsp;
				UnwindContext.Rsp += sizeof(ULONG_PTR);

				//
				// If the unwound RIP is not a valid user-mode address, we assume the
				// unwind failed.  Bail out at this point.
				//
				if (UnwindContext.Rip > (ULONG64)MM_SYSTEM_RANGE_START)
					break;
			}
			//
			// Otherwise, if we failed to find a runtime function entry then we refuse
			// to progress further.
			//
			else
				break;

		} 
		// 
		// If we catch an exception while pulling down the stack, swallow it
		// 
		__except(EXCEPTION_EXECUTE_HANDLER)
		{
			break;
		}

		//
		// Populate & write our current callstack frame
		//
		if (UnwindContext.Rip != 0)
		{
			POBJECT_CALLSTACK_METADATA_FRAME Frames;

			Frames = (POBJECT_CALLSTACK_METADATA_FRAME)(
				(PUCHAR)CallStackMetadata + sizeof(OBJECT_CALLSTACK_METADATA));

			Frames[FrameCount].Rsp       = UnwindContext.Rsp;
			Frames[FrameCount].Rip       = UnwindContext.Rip;
			Frames[FrameCount].ImageBase = ImageBase;
			FrameCount++;
		}

	} while ((UnwindContext.Rip != 0) && (FrameCount < MAX_FRAMES));

	//
	// Update the length of this metadata element based on the number of frames
	// we found
	//
	if (CallStackMetadata)
	{
		CallStackMetadata->Header.Length += 
			FrameCount * sizeof(OBJECT_CALLSTACK_METADATA_FRAME);

		CallStackMetadata->NumberOfFrames = FrameCount;
	}

	//
	// Populate our output pointer
	//
	*OutCallStackMetadata = CallStackMetadata;

	return Status;
}

//
// Gets a process object's name using ZwQueryInformationProcess(ImageFileName)
//
NTSTATUS
GetProcessNameMetadata(
	__in PVOID Object,
	__out POBJECT_NAME_METADATA *OutNameMetadata)
{
	POBJECT_NAME_METADATA NameMetadata = NULL;
	PUNICODE_STRING ImageFileName = NULL;
	KAPC_STATE ApcState;
	NTSTATUS Status;
	BOOLEAN Attached = FALSE;
	PUCHAR NameBuffer;
	ULONG ImageFileNameSize = 512;
	ULONG Tried = 0;

	do
	{
		// 
		// Allocate storage for the name information
		//
		if (!(NameBuffer = (PUCHAR)ExAllocatePoolWithTag(
				NonPagedPool,
				ImageFileNameSize + sizeof(OBJECT_NAME_METADATA),
				ALLOC_TAG)))
		{
			DebugPrint(("GetProcessName: ExAllocatePoolWithTag failed."));

			Status = STATUS_INSUFFICIENT_RESOURCES;
			break;	
		}

		//
		// Initialize the pointers that we'll be working with to the appropriate
		// offsets in the allocated buffers
		//
		NameMetadata  = (POBJECT_NAME_METADATA)NameBuffer;
		ImageFileName = (PUNICODE_STRING)(NameBuffer + sizeof(OBJECT_NAME_METADATA));

		//
		// If we are not running as the process we are about to query, we must
		// stack attach
		//
		if (Object != PsGetCurrentProcess())
		{
			ASSERT(GET_OBJECT_HEADER(Object)->Type == *PsProcessType);

			KeStackAttachProcess(
					Object,
					&ApcState);

			Attached = TRUE;
		}

		//
		// Query the file object's name.   We assume that the caller is running in
		// the context of the process that we are to get the image file name of.
		//
		if (!NT_SUCCESS(Status = ZwQueryInformationProcess(
				NtCurrentProcess(),
				ProcessImageFileName,
				ImageFileName,
				ImageFileNameSize,
				&ImageFileNameSize)))
		{
			if (Status != STATUS_BUFFER_TOO_SMALL)
			{
				DebugPrint(("GetProcessName: ZwQueryInformationProcess failed, %.8x.",
						Status));
				break;
			}

			//
			// Free the buffer so that it can be allocated again
			//
			ExFreePool(
					NameBuffer);

			NameBuffer = NULL;
		}

	} while ((!NT_SUCCESS(Status)) && (++Tried <= 2));

	//
	// Detach if we were previously attached
	//
	if (Attached)
		KeUnstackDetachProcess(
				&ApcState);

	//
	// If we succeeded, initialize the output object
	//
	if (NT_SUCCESS(Status))
	{
		InitializeObjectMetadata(
				NameMetadata,
				OBJECT_METADATA_TAG_NAME,
				sizeof(OBJECT_NAME_METADATA) + ImageFileName->Length,
				Object);

		NameMetadata->NameLength = ImageFileName->Length;

		RtlMoveMemory(
				(PUCHAR)NameMetadata + sizeof(OBJECT_NAME_METADATA),
				ImageFileName->Buffer,
				ImageFileName->Length);

		*OutNameMetadata = NameMetadata;
	}

	return Status;
}


//
// Gets an object's name generically using ObQueryNameString
//
NTSTATUS
GetObjectNameCommon(
	__in PVOID Object,
	__out POBJECT_NAME_METADATA *OutNameMetadata)
{
	POBJECT_NAME_INFORMATION NameInformation = NULL;
	POBJECT_NAME_METADATA NameMetadata = NULL;
	POBJECT_HEADER ObjectHeader = GET_OBJECT_HEADER(Object);
	NTSTATUS Status;
	PUCHAR NameBuffer;
	ULONG NameInformationSize = 512;
	ULONG Tried = 0;

	do
	{
		// 
		// Allocate storage for the name information
		//
		if (!(NameBuffer = (PUCHAR)ExAllocatePoolWithTag(
				NonPagedPool,
				NameInformationSize + sizeof(OBJECT_NAME_METADATA),
				ALLOC_TAG)))
		{
			DebugPrint(("GetObjectName: ExAllocatePoolWithTag failed."));

			Status = STATUS_INSUFFICIENT_RESOURCES;
			break;	
		}

		//
		// Initialize the pointers that we'll be working with to the appropriate
		// offsets in the allocated buffers
		//
		NameMetadata    = (POBJECT_NAME_METADATA)NameBuffer;
		NameInformation = (POBJECT_NAME_INFORMATION)(NameBuffer + sizeof(OBJECT_NAME_METADATA));

		//
		// Query the file object's name
		//
		if (!NT_SUCCESS(Status = ObQueryNameString(
				Object,
				NameInformation,
				NameInformationSize,
				&NameInformationSize)))
		{
			if (Status != STATUS_BUFFER_TOO_SMALL)
			{
				DebugPrint(("GetObjectName: ObQueryNameString failed, %.8x.",
						Status));
				break;
			}

			//
			// Free the buffer so that it can be allocated again
			//
			ExFreePool(
					NameBuffer);

			NameBuffer = NULL;
		}

	} while ((!NT_SUCCESS(Status)) && (++Tried <= 2));

	//
	// If we succeeded and the name has a non-zero length,
	// initialize the output object
	//
	if (NT_SUCCESS(Status) && NameInformation->Name.Length)
	{
		InitializeObjectMetadata(
				NameMetadata,
				OBJECT_METADATA_TAG_NAME,
				sizeof(OBJECT_NAME_METADATA) + NameInformation->Name.Length,
				Object);

		NameMetadata->NameLength = NameInformation->Name.Length;

		RtlMoveMemory(
				(PUCHAR)NameMetadata + sizeof(OBJECT_NAME_METADATA),
				NameInformation->Name.Buffer,
				NameInformation->Name.Length);
	}
	// 
	// Otherwise, free whatever resources we may have allocated
	//
	else if (NameMetadata)
	{
		ExFreePool(
				NameMetadata);

		NameMetadata = NULL;
	}

	*OutNameMetadata = NameMetadata;

	return Status;
}

//
// Gets the object name for a file object
//
NTSTATUS
GetObjectNameFileObject(
	__in PVOID Object,
	__out POBJECT_NAME_METADATA *OutNameMetadata)
{
	POBJECT_NAME_INFORMATION NameInformation = NULL;
	PFILE_OBJECT FileObject;
	NTSTATUS Status;
	ULONG Count = 0;

	//
	// Sanity check that the file object is in a valid state -- this is disabled
	// as it does not work well
	//
	FileObject = (PFILE_OBJECT)Object;

	if (FileObject->DeviceObject == NULL)
	{
		DebugPrint(("GetObjectNameFileObject: Object %p has a null DeviceObject, invalid (count=%d).",
				FileObject, Count));
		return STATUS_DEVICE_NOT_READY;
	}

	//
	// Query the file's DOS device name.
	//
	Status = IoQueryFileDosDeviceName(
			(PFILE_OBJECT)Object,
			&NameInformation);

	//
	// On success, translate this name information into our metadata structure
	//
	if (NT_SUCCESS(Status))
	{
		POBJECT_NAME_METADATA NameMetadata;
		ULONG NameMetadataLength;

		NameMetadataLength = 
			sizeof(OBJECT_NAME_METADATA) + NameInformation->Name.Length;

		NameMetadata = (POBJECT_NAME_METADATA)ExAllocatePoolWithTag(
				PagedPool,
				NameMetadataLength,
				ALLOC_TAG);

		if (NameMetadata)
		{
			InitializeObjectMetadata(
					NameMetadata,
					OBJECT_METADATA_TAG_NAME,
					NameMetadataLength,
					Object);
	
			NameMetadata->NameLength = NameInformation->Name.Length;
	
			RtlCopyMemory(
					(PUCHAR)NameMetadata + sizeof(OBJECT_NAME_METADATA),
					NameInformation->Name.Buffer,
					NameInformation->Name.Length);
	
			*OutNameMetadata = NameMetadata;
		}
		else
			Status = STATUS_INSUFFICIENT_RESOURCES;

		//
		// Free the name information allocated by IoQueryFileDosDeviceName
		//
		ExFreePool(
				NameInformation);
	}

	return Status;
#if 0
	PFILE_NAME_INFORMATION NameInformation = NULL;
	POBJECT_NAME_METADATA NameMetadata = NULL;
	PFILE_OBJECT FileObject;
	NTSTATUS Status;
	PUCHAR NameBuffer;
	ULONG NameInformationSize = 512;
	ULONG Tried = 0;

	do
	{
		//
		// Sanity check that the file object is in a valid state
		//
		FileObject = (PFILE_OBJECT)Object;

		if (FileObject->DeviceObject == NULL)
		{
			DebugPrint(("GetObjectNameFileObject: Object %p has a null DeviceObject, invalid.",
					FileObject));

			Status = STATUS_DEVICE_NOT_READY;
			break;
		}

		// 
		// Allocate storage for the name information
		//
		if (!(NameBuffer = (PUCHAR)ExAllocatePoolWithTag(
				NonPagedPool,
				NameInformationSize + sizeof(OBJECT_METADATA_HEADER),
				ALLOC_TAG)))
		{
			DebugPrint(("GetObjectNameFileObject: ExAllocatePoolWithTag failed."));

			Status = STATUS_INSUFFICIENT_RESOURCES;
			break;	
		}

		//
		// Initialize the pointers that we'll be working with to the appropriate
		// offsets in the allocated buffers
		//
		NameMetadata    = (POBJECT_NAME_METADATA)NameBuffer;
		NameInformation = (PFILE_NAME_INFORMATION)(NameBuffer + sizeof(OBJECT_METADATA_HEADER));

		//
		// Query the file object's name
		//
		if (!NT_SUCCESS(Status = IoQueryFileInformation(
				Object,
				FileNameInformation,
				NameInformationSize,
				NameInformation,
				&NameInformationSize)))
		{
			if (Status != STATUS_BUFFER_TOO_SMALL)
			{
				DebugPrint(("GetObjectNameFileObject: IoQueryFileInformation failed, %.8x.",
						Status));
				break;
			}

			//
			// Free the buffer so that it can be allocated again
			//
			ExFreePool(
					NameBuffer);

			NameBuffer = NULL;
		}

	} while ((!NT_SUCCESS(Status)) && (++Tried <= 2));

	//
	// If we succeeded and the name has a non-zero length,
	// initialize the output object
	//
	if (NT_SUCCESS(Status) && NameMetadata->NameLength)
	{
		InitializeObjectMetadata(
				NameMetadata,
				OBJECT_METADATA_TAG_NAME,
				NameInformationSize + sizeof(OBJECT_METADATA_HEADER),
				Object);
	}
	// 
	// Otherwise, free whatever resources we may have allocated
	//
	else if (NameMetadata)
	{
		ExFreePool(
				NameMetadata);

		NameMetadata = NULL;
	}

	*OutNameMetadata = NameMetadata;

	return Status;
#endif
}

//
// Collects object name information to the object log
//
NTSTATUS
GetObjectNameMetadata(
	__in PVOID Object,
	__out POBJECT_NAME_METADATA *OutNameMetadata)
{
	POBJECT_HEADER Header;
	POBJECT_TYPE ObjectType;
	NTSTATUS Status;

	//
	// Get the object type of the object we are working with
	//
	Header     = GET_OBJECT_HEADER(Object);
	ObjectType = Header->Type;

	//
	// If the object is a process object, call the routine that is able to query
	// the process image file name
	//
	if (ObjectType == *PsProcessType)
		Status = GetProcessNameMetadata(
				Object,
				OutNameMetadata);
	//
	// If this object is a file object, we use a different method to obtain the
	// name information
	//
	else if (ObjectType == *IoFileObjectType)
		Status = GetObjectNameFileObject(
				Object,
				OutNameMetadata);
	//
	// Otherwise, use the common routine which queries object names using
	// ObQueryNameString
	//
	else
		Status = GetObjectNameCommon(
				Object,
				OutNameMetadata);

	return Status;
}

//
// Gets information about the security context that an object is being created
// within
//
NTSTATUS
GetObjectSecurityContextMetadata(
	__in PVOID Object,
	__out POBJECT_SECURITY_CONTEXT_METADATA *OutSecurityContextMetadata)
{
	POBJECT_SECURITY_CONTEXT_METADATA SecurityContextMetadata = NULL;
	SECURITY_SUBJECT_CONTEXT SubjectContext = { 0 };
	PTOKEN_PRIMARY_GROUP PtPrimaryGroup = NULL;
	PTOKEN_PRIMARY_GROUP CtPrimaryGroup = NULL;
	PTOKEN_OWNER PtOwner = NULL;
	PTOKEN_OWNER CtOwner = NULL;
	NTSTATUS Status = STATUS_SUCCESS;
	ULONG PtOwnerLength;
	ULONG CtOwnerLength;
	ULONG PtPrimaryGroupLength;
	ULONG CtPrimaryGroupLength;
	ULONG MetadataLength;
	ULONG Offset;

	do
	{
		//
		// Capture the caller's subject context
		//
		SeCaptureSubjectContext(
				&SubjectContext);

		//
		// Query owner and primary group SID for the primary token, if any
		//
		if (SubjectContext.PrimaryToken)
		{
			SeQueryInformationToken(
					SubjectContext.PrimaryToken,
					TokenOwner,
					&PtOwner);

			SeQueryInformationToken(
					SubjectContext.PrimaryToken,
					TokenPrimaryGroup,
					&PtPrimaryGroup);
		}

		//
		// Query owner and primary group SID for the client token, if any
		//
		if (SubjectContext.ClientToken)
		{
			SeQueryInformationToken(
					SubjectContext.ClientToken,
					TokenOwner,
					&CtOwner);

			SeQueryInformationToken(
					SubjectContext.ClientToken,
					TokenPrimaryGroup,
					&CtPrimaryGroup);
		}

		//
		// Build the security metadata structure.  SIDs are trusted, right? :)
		//
		PtOwnerLength =
			((PtOwner && PtOwner->Owner) ? RtlLengthSid(PtOwner->Owner) : 0);
		CtOwnerLength =
			((CtOwner && CtOwner->Owner) ? RtlLengthSid(CtOwner->Owner) : 0);
		PtPrimaryGroupLength =
			((PtPrimaryGroup && PtPrimaryGroup->PrimaryGroup) ? RtlLengthSid(PtPrimaryGroup->PrimaryGroup) : 0);
		CtPrimaryGroupLength = 
			((CtPrimaryGroup && CtPrimaryGroup->PrimaryGroup) ? RtlLengthSid(CtPrimaryGroup->PrimaryGroup) : 0);

		MetadataLength = 
			sizeof(OBJECT_SECURITY_CONTEXT_METADATA) +
			PtOwnerLength + 
			CtOwnerLength +
			PtPrimaryGroupLength +
			CtPrimaryGroupLength;

		SecurityContextMetadata = (POBJECT_SECURITY_CONTEXT_METADATA)ExAllocatePoolWithTag(
				PagedPool,
				MetadataLength,
				ALLOC_TAG);

		if (!SecurityContextMetadata)
		{
			Status = STATUS_INSUFFICIENT_RESOURCES;
			break;
		}

		//
		// Populate security metadata header information
		//
		InitializeObjectMetadata(
				SecurityContextMetadata,
				OBJECT_METADATA_TAG_SECURITY_CONTEXT,
				MetadataLength,
				Object);

		SecurityContextMetadata->ImpersonationLevel = SubjectContext.ImpersonationLevel;
		SecurityContextMetadata->ClientTokenOwnerSidLength = CtOwnerLength;
		SecurityContextMetadata->ClientTokenPrimaryGroupSidLength = CtPrimaryGroupLength;
		SecurityContextMetadata->PrimaryTokenOwnerSidLength = PtOwnerLength;
		SecurityContextMetadata->PrimaryTokenPrimaryGroupSidLength = PtPrimaryGroupLength;
		Offset = sizeof(OBJECT_SECURITY_CONTEXT_METADATA);

		//
		// Populate dynamic SID fields
		//
		if (CtOwnerLength)
		{
			RtlCopyMemory(
					(PUCHAR)SecurityContextMetadata + Offset,
					CtOwner->Owner,
					CtOwnerLength);
			Offset += CtOwnerLength;
		}

		if (CtPrimaryGroupLength)
		{
			RtlCopyMemory(
					(PUCHAR)SecurityContextMetadata + Offset,
					CtPrimaryGroup->PrimaryGroup,
					CtPrimaryGroupLength);
			Offset += CtPrimaryGroupLength;
		}

		if (PtOwnerLength)
		{
			RtlCopyMemory(
					(PUCHAR)SecurityContextMetadata + Offset,
					PtOwner->Owner,
					PtOwnerLength);
			Offset += PtOwnerLength;
		}

		if (PtPrimaryGroupLength)
		{
			RtlCopyMemory(
					(PUCHAR)SecurityContextMetadata + Offset,
					PtPrimaryGroup->PrimaryGroup,
					PtPrimaryGroupLength);
			Offset += PtPrimaryGroupLength;
		}

	} while (0);

	SeReleaseSubjectContext(
			&SubjectContext);

	*OutSecurityContextMetadata = SecurityContextMetadata;

	return Status;
}

//
// Gets the security descriptor metadata associated with an object
//
NTSTATUS
GetObjectSecurityDescriptorMetadata(
	__in PVOID Object,
	__in BOOLEAN UseObGetObjectSecurity,
	__out POBJECT_SECURITY_DESCRIPTOR_METADATA *OutSecurityDescriptorMetadata)
{
	POBJECT_SECURITY_DESCRIPTOR_METADATA SecurityDescriptorMetadata = NULL;
	PISECURITY_DESCRIPTOR ObjectSd = NULL;
	POBJECT_HEADER ObjectHeader;
	NTSTATUS Status = STATUS_SUCCESS;
	BOOLEAN MemoryAllocated = FALSE;
	ULONG SecurityDescriptorMetadataLength;
	ULONG ObjectSdLength;

	do
	{
		//
		// If the caller requested, use ObGetObjectSecurity to get the object's
		// security descriptor.  It is only safe to call this in the
		// ObPostCallback as the object may not be fully initialized in the
		// context of other hooks such as ObCreateObject.
		//
		if (UseObGetObjectSecurity)
		{
			Status = ObGetObjectSecurity(
					Object,
					&ObjectSd,
					&MemoryAllocated);

			if (!NT_SUCCESS(Status))
			{
				DebugPrint(("GetObjectSecurityDescriptorMetadata: ObGetObjectSecurity(%p) failed, %.8x",
						Object, Status));
				break;
			}
		}
		//
		// If we aren't using ObGetObjectSecurity, we directly reference the
		// security descriptor as found in the object header.  We must mask off
		// the fastref count in the low bits of the security descriptor to get the
		// actual address.
		//
		else
		{
			ObjectHeader = GET_OBJECT_HEADER(Object);

			ObjectSd = (PISECURITY_DESCRIPTOR)((ULONG_PTR)ObjectHeader->SecurityDescriptor & ~0xf);
		}

		//
		// If we have a valid security descriptor, let's log it.  There is a race
		// condition here that may need to be addressed.
		//
		if (!ObjectSd)
		{
			Status = STATUS_UNSUCCESSFUL;
			break;
		}

		//
		// For now I'm assuming that all security descriptors are
		// self-relative.  We'll see if this holds.
		//
		ASSERT(ObjectSd->Control & SE_SELF_RELATIVE);

		//
		// Get the length of the security descriptor
		//
		ObjectSdLength = RtlLengthSecurityDescriptor(
				ObjectSd);

		SecurityDescriptorMetadataLength = 
			sizeof(OBJECT_SECURITY_DESCRIPTOR_METADATA) + ObjectSdLength;

		//
		// Allocate and initialize the metadata we're going to log
		//
		SecurityDescriptorMetadata = (POBJECT_SECURITY_DESCRIPTOR_METADATA)ExAllocatePoolWithTag(
				PagedPool,
				SecurityDescriptorMetadataLength,
				ALLOC_TAG);

		if (!SecurityDescriptorMetadata)
		{
			Status = STATUS_INSUFFICIENT_RESOURCES;
			break;
		}

		InitializeObjectMetadata(
				SecurityDescriptorMetadata,
				OBJECT_METADATA_TAG_SECURITY_DESC,
				SecurityDescriptorMetadataLength,
				Object);

		SecurityDescriptorMetadata->SdLength = ObjectSdLength;

		RtlCopyMemory(
				(PUCHAR)SecurityDescriptorMetadata + sizeof(OBJECT_SECURITY_DESCRIPTOR_METADATA),
				ObjectSd,
				ObjectSdLength);

	} while (0);

	if (!NT_SUCCESS(Status))
	{
		if (SecurityDescriptorMetadata)
			ExFreePool(
					SecurityDescriptorMetadata);

		SecurityDescriptorMetadata = NULL;
	}

	//
	// If we used ObGetObjectSecurity we must release our reference to the
	// security descriptor
	//
	if ((ObjectSd) && (UseObGetObjectSecurity))
		ObReleaseObjectSecurity(
				ObjectSd,
				MemoryAllocated);

	*OutSecurityDescriptorMetadata = SecurityDescriptorMetadata;

	return Status;
}

//
// Gets metadata that describes an attempt to access an object.  An access can
// include defining the object or using the object.
//
NTSTATUS
GetObjectAccessMetadata(
	__in PVOID Object,
	__in OBJECT_ACCESS_METADATA_FLAGS AccessFlags,
	__in ULONG AccessMask,
	__out POBJECT_ACCESS_METADATA *OutAccessMetadata)
{
	POBJECT_SECURITY_CONTEXT_METADATA SecurityContextMetadata = NULL;
	POBJECT_SECURITY_DESCRIPTOR_METADATA SecurityDescriptorMetadata = NULL;
	POBJECT_CALLSTACK_METADATA CallStackMetadata = NULL;
	POBJECT_ACCESS_METADATA AccessMetadata;
	NTSTATUS Status = STATUS_SUCCESS;
	ULONG SecurityContextMetadataLength;
	ULONG SecurityDescriptorMetadataLength;
	ULONG CallStackMetadataLength;
	ULONG AccessMetadataLength = 0;
	ULONG Offset;

	//
	// Query each of the metadata elements that we'll encapsulate into the access
	// metadata.  It's OK if these fail.
	//
	GetObjectSecurityContextMetadata(
			Object,
			&SecurityContextMetadata);

	GetObjectCallStackMetadata(
			Object,
			&CallStackMetadata);

	//
	// We only care about the security descriptor if this is a definition of the
	// object.
	//
	if (AccessFlags.Definition)
		GetObjectSecurityDescriptorMetadata(
				Object,
				FALSE,
				&SecurityDescriptorMetadata);

	//
	// Calculate the size of the access metadata structure and allocate storage
	// for it
	//
	SecurityContextMetadataLength = 
		(SecurityContextMetadata) ? SecurityContextMetadata->Header.Length : 0;
	SecurityDescriptorMetadataLength = 
		(SecurityDescriptorMetadata) ? SecurityDescriptorMetadata->Header.Length : 0;
	CallStackMetadataLength = 
		(CallStackMetadata) ? CallStackMetadata->Header.Length : 0;

	AccessMetadataLength = 
		sizeof(OBJECT_ACCESS_METADATA) +
		SecurityContextMetadataLength +
		SecurityDescriptorMetadataLength +
		CallStackMetadataLength;

	AccessMetadata = (POBJECT_ACCESS_METADATA)ExAllocatePoolWithTag(
			NonPagedPool,
			AccessMetadataLength,
			ALLOC_TAG);

	if (AccessMetadata)
	{
		//
		// Initialize the static fields within the access metadata
		//
		InitializeObjectMetadata(
				AccessMetadata,
				OBJECT_METADATA_TAG_ACCESS,
				AccessMetadataLength,
				Object);

		AccessMetadata->ProcessObject = (ULONG64)PsGetCurrentProcess();
		AccessMetadata->Flags         = AccessFlags.AsUlong;
		AccessMetadata->AccessMask    = AccessMask;

		AccessMetadata->SecurityContextLength    = SecurityContextMetadataLength;
		AccessMetadata->SecurityDescriptorLength = SecurityDescriptorMetadataLength;
		AccessMetadata->CallStackLength          = CallStackMetadataLength;

		//
		// Conditionally copy the sub-metadata elements at the appropriate offsets
		//
		Offset = sizeof(OBJECT_ACCESS_METADATA);

		if (SecurityContextMetadataLength)
		{
			RtlCopyMemory(
					(PUCHAR)AccessMetadata + Offset,
					SecurityContextMetadata,
					SecurityContextMetadataLength);
			Offset += SecurityContextMetadataLength;
		}

		if (SecurityDescriptorMetadataLength)
		{
			RtlCopyMemory(
					(PUCHAR)AccessMetadata + Offset,
					SecurityDescriptorMetadata,
					SecurityDescriptorMetadataLength);
			Offset += SecurityDescriptorMetadataLength;
		}

		if (CallStackMetadataLength)
		{
			RtlCopyMemory(
					(PUCHAR)AccessMetadata + Offset,
					CallStackMetadata,
					CallStackMetadataLength);
			Offset += CallStackMetadataLength;
		}
	}
	else
		Status = STATUS_INSUFFICIENT_RESOURCES;

	//
	// Before returning, we free the resources we used to get the sub-metadata
	// elements that were then copied into a larger allocation
	//
	if (SecurityContextMetadata)
		ExFreePool(
				SecurityContextMetadata);
	if (SecurityDescriptorMetadata)
		ExFreePool(
				SecurityDescriptorMetadata);
	if (CallStackMetadata)
		ExFreePool(
				CallStackMetadata);

	//
	// Give the caller what they want
	//
	*OutAccessMetadata = AccessMetadata;

	return Status;
}
