//
// Dynamic trust boundary metadata collection driver
//
// mmiller@hick.org 
// 3/2008
//
#ifndef _DYNBA_METADATA_H
#define _DYNBA_METADATA_H

#pragma pack(push)
#pragma pack(1)

//
// Tags that identify each metadata structure
//
#define OBJECT_METADATA_TAG_CREATE           0x1
#define OBJECT_METADATA_TAG_SECURITY_CONTEXT 0x2
#define OBJECT_METADATA_TAG_SECURITY_DESC    0x3
#define OBJECT_METADATA_TAG_NAME             0x4
#define OBJECT_METADATA_TAG_CALL_STACK       0x5
#define OBJECT_METADATA_TAG_IMAGE_INFO       0x6
#define OBJECT_METADATA_TAG_ACCESS           0x7

//
// Metadata header initialization
//
#define InitializeObjectMetadata(Header, T, L, O)   \
	((POBJECT_METADATA_HEADER)Header)->Tag = (GetNextObjectLogSequence() << 4) | (T); \
	((POBJECT_METADATA_HEADER)Header)->Length = L; \
	((POBJECT_METADATA_HEADER)Header)->Object = (ULONG64)(O)

//
// The header shared by all metadata structures
//
typedef struct _OBJECT_METADATA_HEADER
{
	//
	// The metadata's tag (type and sequence)
	//
	ULONG64 Tag;

	//
	// The length of of the metadata
	//
	ULONG Length;

	//
	// The object the metadata is attached to
	//
	ULONG64 Object;

} OBJECT_METADATA_HEADER, *POBJECT_METADATA_HEADER;

//
// Object security context metadata
//
typedef struct _OBJECT_SECURITY_CONTEXT_METADATA
{
	OBJECT_METADATA_HEADER Header;

	SECURITY_IMPERSONATION_LEVEL ImpersonationLevel;
	ULONG ClientTokenOwnerSidLength;
	ULONG ClientTokenPrimaryGroupSidLength;
	ULONG PrimaryTokenOwnerSidLength;
	ULONG PrimaryTokenPrimaryGroupSidLength;

	//
	// Dynamic buffers follows in order:
	//
	// Client token owner SID
	// Client token primary group SID
	// Primary token owner SID
	// Primary token primary group SID
	//

} OBJECT_SECURITY_CONTEXT_METADATA, *POBJECT_SECURITY_CONTEXT_METADATA;

//
// Object metadata that describes the security descriptor of the object in terms
// of owner, group, and DACLs
//
typedef struct _OBJECT_SECURITY_DESCRIPTOR_METADATA
{
	//
	// Metadata header
	//
	OBJECT_METADATA_HEADER Header;

	//
	// The length of the object's security descriptor
	//
	ULONG SdLength;

	//
	// Dynamic buffers follow in this order:
	//
	// Sd
	//

} OBJECT_SECURITY_DESCRIPTOR_METADATA, *POBJECT_SECURITY_DESCRIPTOR_METADATA;

//
// Object name metadata
//
typedef struct _OBJECT_NAME_METADATA
{
	//
	// Metadata header
	//
	OBJECT_METADATA_HEADER Header;

	//
	// The length of the name that follows, in bytes
	//
	USHORT NameLength;

	//
	// The name of the object
	//
	//WCHAR Name[0];

} OBJECT_NAME_METADATA, *POBJECT_NAME_METADATA;

typedef struct _OBJECT_CALLSTACK_METADATA_FRAME
{
	//
	// RSP for the frame
	//
	ULONG64 Rsp;
	//
	// RIP for the frame
	//
	ULONG64 Rip;
	//
	// The image base of the module that RIP is a part of
	//
	ULONG64 ImageBase;

} OBJECT_CALLSTACK_METADATA_FRAME, *POBJECT_CALLSTACK_METADATA_FRAME;

//
// The call stack is from a use of the object (such as when a handle is created
// to the object)
//
#define OBJECT_CALLSTACK_METADATA_FLAG_USE      0x1
//
// The call stack could not be obtained because the thread was running at DIRQL
//
#define OBJECT_CALLSTACK_METADATA_FLAG_AT_DIRQL 0x2

//
// Variable-sized metadata that describes a callstack
//
typedef struct _OBJECT_CALLSTACK_METADATA
{
	//
	// Metadata header
	//
	OBJECT_METADATA_HEADER Header;

	//
	// The flags associated with this callstack.  In the future, 1 will indicate
	// that this callstack occurred during a use of the object.
	//
	ULONG Flags;

	//
	// The number of frames described
	//
	ULONG NumberOfFrames;

	//
	// Array of callstack frames terminated by a NULL frame
	//
	//OBJECT_CALLSTACK_METADATA_FRAME Frames[0];

} OBJECT_CALLSTACK_METADATA, *POBJECT_CALLSTACK_METADATA;

//
// Metadata that describes information about an image that has been loaded into
// a process
//
typedef struct _OBJECT_IMAGEINFO_METADATA
{
	//
	// Metadata header
	//
	OBJECT_METADATA_HEADER Header;

	//
	// The base address the image has been loaded at
	//
	ULONG64 BaseAddress;

	//
	// The size of the image
	//
	ULONG64 ImageSize;

	//
	// The length of the image name that has been loaded
	//
	USHORT ImageNameLength;

} OBJECT_IMAGEINFO_METADATA, *POBJECT_IMAGEINFO_METADATA;

//
// Access flags that are significant in the Flags field of the access metadata
//
typedef union _OBJECT_ACCESS_METADATA_FLAGS
{
	struct
	{
		ULONG Definition :  1;
		ULONG Reserved   : 31;
	};
	ULONG AsUlong;
} OBJECT_ACCESS_METADATA_FLAGS, *POBJECT_ACCESS_METADATA_FLAGS;

//
// Metadata that describes a particular access to an object whether it be a
// definition of an object or a use of an object
//
typedef struct _OBJECT_ACCESS_METADATA
{
	//
	// Metadata header
	//
	OBJECT_METADATA_HEADER Header;

	//
	// The process context the access occurs within
	//
	ULONG64 ProcessObject;

	//
	// Flags about the type of access (definition, use, etc)
	//
	ULONG Flags;

	//
	// The access mask that was granted -- only applicable when an object is
	// used.
	//
	ULONG AccessMask;

	//
	// Length of the security context metadata for this access
	//
	ULONG SecurityContextLength;

	//
	// Length of the security descriptor metadata for this access
	//
	ULONG SecurityDescriptorLength;

	//
	// Length of the call stack metadata for this access
	//
	ULONG CallStackLength;

	//
	// Dynamic fields follow in this order:
	//
	// OBJECT_SECURITY_CONTEXT_METADATA
	// OBJECT_SECURITY_DESCRIPTOR_METADATA
	// OBJECT_METADATA_CALL_STACK
	//
	
} OBJECT_ACCESS_METADATA, *POBJECT_ACCESS_METADATA;

#pragma pack(pop)

//
// Super hack -- set the low bit of the InitialStack in the provided thread
// object as a flag for indicating whether we may have recursed.
//
#define SET_RECURSE_FLAG(Thread) \
	InterlockedOr( \
			(PLONG)((PUCHAR)Thread+0x1ec), \
			0x01)
#define CLEAR_RECURSE_FLAG(Thread) \
	InterlockedAnd( \
			(PLONG)((PUCHAR)Thread+0x1ec), \
			0xfffffffe)

#define CHECK_RECURSE_FLAG(Thread) \
	((((PUCHAR)Thread)[0x1ec] & 1) != 0)

//
// Public interface
//

#ifndef USER_MODE
NTSTATUS
GetObjectCallStackMetadata(
	__in PVOID Object,
	__out POBJECT_CALLSTACK_METADATA *OutCallStackMetadata);

NTSTATUS
GetProcessNameMetadata(
	__in PVOID Object,
	__out POBJECT_NAME_METADATA *OutNameMetadata);

NTSTATUS
GetObjectNameMetadata(
	__in PVOID Object,
	__out POBJECT_NAME_METADATA *OutNameMetadata);

NTSTATUS
GetObjectSecurityContextMetadata(
	__in PVOID Object,
	__out POBJECT_SECURITY_CONTEXT_METADATA *OutSecurityContextMetadata);

NTSTATUS
GetObjectSecurityDescriptorMetadata(
	__in PVOID Object,
	__in BOOLEAN UseObGetObjectSecurity,
	__out POBJECT_SECURITY_DESCRIPTOR_METADATA *OutSecurityDescriptorMetadata);

NTSTATUS
GetObjectAccessMetadata(
	__in PVOID Object,
	__in OBJECT_ACCESS_METADATA_FLAGS AccessFlags,
	__in ULONG AccessMask,
	__out POBJECT_ACCESS_METADATA *AccessMetadata);

#endif

#endif
